unit SetList;
{$I Common.inc}
//************************************************************************
//                            *
//                                                                       *
//  :                                                     *
// TSetList -- ,                            *
//************************************************************************

interface

uses Root, Items;

type

  //      

  PListRec = ^TListRec;
  TListRec = record
    mItem : TItem;
    mNext : PListRec;
  end;

  // ,  

  TSetList = class (TSet)
    protected
      mHead: PListRec; //  
      procedure Clr(arg: boolean); override;
    private
      mCurrent : PListRec;  //    
      mStack : TBuffer;   //      mCurrent
      procedure Del(var p, q : PListRec);
      procedure Ins(p: PListRec; var q: PListRec; arg: TItem);
    public
      destructor Destroy;  override;
      function Exist(arg : TItem): boolean; override;
      function Insert(arg: TItem): boolean; override;
      procedure Delete(arg: TItem); override;
      function GetFirst: TItem; override;
      function GetNext: TItem; override;
      procedure PositionPush;  override;
      procedure PositionPop;   override;
      procedure Add(arg : TSet); override;
      procedure Sub(arg : TSet); override;
      procedure Mul(arg : TSet); override;
      procedure ExOr(arg : TSet); override;
  end;

//////////////////////////////////////////////////////////////////////////
implementation
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// TSetList -- ,  
//////////////////////////////////////////////////////////////////////////

//    

function TSetList.GetFirst: TItem;
begin
  mCurrent:= mHead;
  if Assigned(mCurrent)
    then Result:= mCurrent^.mItem
    else Result:= nil;
end;

//    

function TSetList.GetNext: TItem;
begin
  if Assigned(mCurrent)
    then mCurrent:= mCurrent^.mNext;
  if Assigned(mCurrent)
    then Result:= mCurrent^.mItem
    else Result:= nil;
end;

//      

procedure TSetList.PositionPop;
begin
  mCurrent:= PListRec(mStack.Pop);
end;

//      

procedure TSetList.PositionPush;
begin
  if not Assigned(mStack)
    then mStack:= TBuffer.Create;
  mStack.Push(TItem(mCurrent));
end;

//     

procedure TSetList.Clr(arg: boolean);
var p : PListRec;
begin
  while Assigned(mHead) do begin
    p:= mHead;
    mHead:= mHead^.mNext;
    if arg and Assigned(p^.mItem)
       then p^.mItem.Free;
    Dispose(p);
  end;
  mCount:= 0;
end;

//  

destructor TSetList.Destroy;
begin
  mStack.Free;
  inherited;
end;

//     

function TSetList.Exist(arg: TItem): boolean;
var p : PListRec;
begin
  Result:= false;
  if not Assigned(arg) then Exit;
  p:= mHead;
  while Assigned(p) do begin
    if p.mItem = arg then begin Result:= true; Break end;
    case p^.mItem.Compare(arg) of
      cmpLess,
      cmpIncomp: p:=p^.mNext;  //   
      cmpEq    : begin Result:= true; Break end;
      cmpGreate: Break;  // 
    end;
  end;
end;

//   

procedure TSetList.Ins(p: PListRec; var q: PListRec; arg: TItem);
var r : PListRec;
begin
  Inc(mCount);
  New(r);  r^.mItem:= arg; r^.mNext:= p;
  if Assigned(q)
    then q^.mNext:= r
    else mHead:= r;
  q:= r;
end;

//    
//  FALSE,
//         

function TSetList.Insert(arg: TItem): boolean;
var p, q : PListRec;
    cmp : TCompare;
begin
  Result:= false;
  if not Assigned(arg) then Exit;
  cmp:= cmpEq;
  p:= mHead; q:=nil;
  while Assigned(p) do begin
    cmp:= p^.mItem.Compare(arg);
    if cmp in [cmpEq, cmpGreate] then Break;
    q:= p;  p:= p^.mNext
  end;
  if not Assigned(p) or (cmp = cmpGreate) then begin
    Ins(p,q,arg);
    Result:= true;
  end;
end;

//   

procedure TSetList.Del(var p, q: PListRec);
begin
  if Assigned(q)
    then q^.mNext:= p^.mNext
    else mHead:= p^.mNext;
  Dispose(p);
  if Assigned(q)
    then p:= q^.mNext
    else p:= mHead;
  Dec(mCount);
end;

//     

procedure TSetList.Delete(arg: TItem);
var p, q : PListRec;
begin
  if not Assigned(arg) then Exit;
  p:= mHead; q:= nil;
  while Assigned(p) do begin
    case p^.mItem.Compare(arg) of
      cmpLess: begin q:= p;  p:= p^.mNext;  end;
      cmpEq:   begin Del(p,q); Break; end;
      else Break;
    end;
  end;
end;

//  :     

procedure TSetList.Add(arg: TSet);
var p, q : PListRec;
    t : TItem;
    cmp : TCompare;
begin
  if not Assigned(arg) then Exit;
  if not (arg is TSet) then Exit;
  p:= mHead; q:= nil;
  arg.PositionPush;
  t:= arg.GetFirst;
  while Assigned(t) do begin
    cmp:= cmpEq;
    while Assigned(p) do begin
      cmp:= p^.mItem.Compare(t);
      case cmp of
        cmpLess  : begin q:= p; p:= p^.mNext end;
        cmpEq    : begin t:= arg.GetNext; break end;
        else  Break;
      end;
    end;
    if not Assigned(p) or (cmp = cmpGreate) then begin
      Ins(p,q,t);
      if Assigned(t) then t:= arg.GetNext;
    end;
  end;
  arg.PositionPop;
end;

//  :     

procedure TSetList.Sub(arg: TSet);
var p, q : PListRec;
    t : TItem;
    cmp : TCompare;
begin
  if not Assigned(arg) then Exit;
  if not arg.InheritsFrom(TSet) then Exit;
  p:= mHead; q:= nil;
  arg.PositionPush;
  t:= arg.GetFirst;
  while Assigned(t) do begin
    cmp:= cmpLess;
    while Assigned(p) do begin
      cmp:= p^.mItem.Compare(t);
      if cmp in [cmpEq, cmpGreate] then Break;
      q:= p; p:= p^.mNext;
    end;
    if cmp = cmpEq then Del(p,q);
    if not Assigned(p) then Break;
    t:= arg.GetNext;
  end;
  arg.PositionPop;
end;

//  :     

procedure TSetList.Mul(arg: TSet);
var p, q : PListRec;
    t : TItem;
    cmp : TCompare;
begin
  if not Assigned(arg) then Exit;
  p:= mHead; q:= nil;
  arg.PositionPush;
  t:= arg.GetFirst;
  while Assigned(t) do begin
    cmp:= cmpEq;
    while Assigned(p) do begin
      cmp:= p^.mItem.Compare(t);
      if cmp <> cmpEq then Break;
      q:= p;
      p:= p^.mNext;
    end;
    if cmp = cmpLess then Del(p,q);
    if not Assigned(p) then Break;
    if cmp = cmpGreate then t:= arg.GetNext;
  end;
  while Assigned(p) do Del(p,q);
  arg.PositionPop;
end;

//   :    

procedure TSetList.ExOr(arg: TSet);
var p, q : PListRec;
    t : TItem;
    cmp : TCompare;
begin
  if not Assigned(arg) then Exit;
  p:= mHead; q:= nil;
  arg.PositionPush;
  t:= arg.GetFirst;
  while Assigned(t) do begin
    while Assigned(p) do begin
      cmp:= p^.mItem.Compare(t);
      case cmp of
        cmpLess:
          begin  q:= p; p:= p^.mNext end;
        cmpEq:
          begin
            Del(p, q);
            t:= arg.GetNext;
            if not Assigned(t) then break;
          end;
        cmpGreate:
          begin
            Ins(p, q, t);
            t:= arg.GetNext;
            if not Assigned(t) then break;
          end;
      end;
    end;
    while Assigned(t) do begin
      Ins(p, q, t);
      t:= arg.GetNext
    end;
  end;
  arg.PositionPop;
end;

//////////////////////////////////////////////////////////////////////////

end.
