unit Graph;
{$I Common.inc}
//************************************************************************
//                                                            *
//                                                                       *
//  :                                                     *
//                                                                       *
// TNode --   ()                                       *
// TNodeInt -- ,                                  *
// TCondensNode --                                      *
// TLink --                                                   *
// TFarLink --                                       *
// TGraph --                                                  *
// THugeGraph --    -                       *
// TCostNodes --                                  *
// TSortedNum --                                   *
// TGamma --                                              *
// TAreal --                                       *
// TTreeLink --                            *
// TPair --                              *
// TPairsBlock --                           *
// THamNode --            *
// THamLink --               *
// THamGraph --                    *
// TClosed --                     *
//************************************************************************

interface

uses Root, Items;

type

  TReal = Single;  //   

  //         :

  TCenter = ( Undefined,  //  
              InCenter,   //    
              OutCenter,  //    
              InOutCenter // -   
            );

  //   :

  TPairs = (pMaxN,  //  
            pMaxW,  //  
            pMinW   //  
           );

  //  :

  TNode = class;     // 
  TLink = class;     // 
  TGraph = class;    // 
  TFarLink = class;  //  

  //  

  TNode = class (TItem)      //  () 
    private
      mLnkOut : TSet;        //   
      mLnkIn  : TSet;        //   
      mFarLinks: TSet;       //     
      function GetNear(aFar: TNode): TNode;
      function GetFarLink(aFar: TNode): TFarLink;
      procedure SortFarLinks;
      procedure PrintFarLinks(var aFile: TextFile);
      procedure ExpoFarLinks;
      procedure ExpoData;
    protected
      mOwner  : TGraph;   //     
      mColor  : integer;  // 
      mDist   : integer;  //  ()
                          //   ( )
                          //      ()
      mPred   : TNode;    //   ()
      mLink   : TLink;    //    (, )
      mRoot   : TItem;    //  ( ,  )
      mFlow   : integer;  //   ()
                          //   ( )
      mLimit  : integer;  //   ( )
      mPower  : integer;  //    ( )
      //--------------------
      //  
      mBlossom: TNode;    //    (TBlossom)
      mPair   : TLink;    //        
      mRight  : TNode;    //      
      mLeft   : TNode;    //      
      mRLnk   : TLink;    //      
      mOut    : boolean;  //    :
                          // false -- , true -- 
      mOwnDest : boolean; //     :
                          // false -- , true --  
      mProfit : integer;  //  ()  
                          //      
      //--------------------

      function  MakeLink(aDest: TNode; aVal : integer): TLink;
      function  GetLink(aDest: TNode): TLink;
      procedure RemoveLink(aDest: TNode);
      procedure AddTreeLinks(aMaxTree: boolean; aLinks: TSet);
      procedure SetLinksLimit(aLimit: integer);
      function  IsLeft: boolean;
      function  IsRight: boolean;
      procedure ResetNode;
      //    
      function GetBlossom: TNode; //    
      function GetPair: TNode;    //    
    public
      mValue : integer;      //  
      constructor Create(aVal: integer; aOwner: TGraph);
      destructor Destroy;  override;
      function  Compare(arg: TItem): TCompare; override;
      function  GetName: string;  virtual;
      procedure PrintLinks(var aFile: TextFile);
      procedure Print(var aFile: TextFile); override;
      function  GenStrongArea: TSet;
      //       
      function  OutGetCnt: integer;
      function  OutLinkFirst: TLink;
      function  OutLinkNext: TLink;
      procedure OutPosPush;
      procedure OutPosPop;
      function  OutGetLink(aIndex: integer): TLink;
      procedure OutGammaAdd(aRes : TSet);
      procedure OutGammaGet(aRes : TSet);
      function  CalcCost: integer;
      function  GenGammaOut: TSet;
      //       
      function  InGetCnt: integer;
      function  InLinkFirst: TLink;
      function  InLinkNext: TLink;
      procedure InPosPush;
      procedure InPosPop;
      function  InGetLink(aIndex: integer): TLink;
      procedure InGammaAdd(aRes : TSet);
      function  GenGammaIn: TSet;
  end;

  // ,  

  TNodeInt = class (TNode)
    public
      mNumber  : integer;  //  
      constructor Create(aNumber, aVal: integer; aOwner: TGraph);
      function Compare(arg: TItem): TCompare; override;
      function Copy: TItem; override;
      function GetName: string;  override;
      function GetLinkByNumber(aNumber : integer): TLink;
  end;

  //  --   

  TLink = class (TItem)
    protected
      mLoadLinks : boolean;  //   
      mLimit  : integer;     //   ()
    public
      mValue: integer;  //  (, ) 
      mOwner: TNode;    //   ()
      mDest : TNode;    //  
      mColor: integer;  //   ()
      //    :
      mLow   : integer; //     mLow <= mHigh
      mHigh  : integer; //   
      mFlow  : integer; //  
      mTemp  : integer; //    mHigh
      mDirect: integer; //   (+1 -- , -1 -- )
      constructor Create(aOwner, aDest: TNode; aVal: integer;
                         aLoadLinks: boolean);
      function  Compare(arg: TItem): TCompare; override;
      function  Copy: TItem; override;
      function  GetReverse: TLink; //   
      procedure Print(var aFile: TextFile); override;
      procedure Print2(var aFile: TextFile);
  end;

  //  

  TGraph = class (TItem)
    private
      mInitMap : boolean;   //     
      //   p-:
      mMapDirect : TCenter; //     
    protected
      mNodes : TSet;        //  
    public
      mName : String;       //  
      mDirect : boolean;    //  
      mLoadNodes : boolean; //   
      mLoadLinks : boolean; //   
    protected
    //  
      procedure PosPush;
      procedure PosPop;
      procedure ResetNodes;                    //  
      function  GetNode(aNode: TNode): TNode;  //     
      procedure SetLink(aSource, aDest : TNode; aVal : integer);
      procedure SetLinksLimit(aLimit: integer);
      procedure Compress; //     
    public
    //  
      constructor Create(const aName: string; aDir, aLNodes, aLLinks : boolean);
      destructor Destroy;  override;
      function  InsertNode(aNode: TNode): boolean;
      procedure RemoveNode(aNode: TNode);  //    
      function  Compare(arg: TItem): TCompare; override;
      procedure Print(var aFile: TextFile); override;
      procedure Save(const aName: String); virtual; abstract;
      function  Nodes: integer;
      procedure ExpoFarLinks;
      procedure ExpoLinksData;
    //  
      function  NodeFirst: TNode;
      function  NodeNext: TNode;
    //    
      function  CopyGr(aRevers: boolean): TGraph;
      function  Copy: TItem; override;
      function  CopyByNodes(aNodes: TSet): TGraph;
      function  CopyByLinks(aLinks: TSet): TGraph;
    // 
      function  TestLink(arg1, arg2 : TNode): boolean;
      function  CalcSteps(arg1, arg2 : TNode): integer;
      function  GenPath(aSource, aDest : TNode): TBuffer;
      function  GenDijkstra(aSource, aDest : TNode): TBuffer;
      function  GenQuickPath(aSource, aDest : TNode): TBuffer;
      function  GetQuickPathStr(aSource, aDest : TNode): String;
    private
      procedure InitMap_Floyd; //   
    public
    //   (  )
      function IsLinked: boolean;
    //     
      function  CalcAreas: integer;  //  
      function  GenAreas: TSet;      //   
    //      
      function  GenStrongAreas: TSet;
      function  GenCondens: TGraph;
    //  , ,  
      function  GenUndepend: TSet;
      function  GenClique: TSet;
      function  GenDominating: TCostSet;
    // 
      function  GenPaints(aQuick: boolean): TSet;
    // 
      function  GenCenter(var aLambda: integer; aDirect: TCenter): TSet;
      function  GenPCentLambda(aLambda: integer; aDirect: TCenter): TCostSet;
      function  GenPCenters(aPoly: integer;
                            var aLambda: integer;
                            aCenter: TCenter): TCostSet;
    private
      procedure InitMap(aDirect: TCenter);
      procedure DoneMap;
      function  GenIntersections(aGams: TSet): TBuffer;
      function  GenLimitGams(aLambda: integer): TSet;
      function  GenLimitGamma(aNode: TNode; aLambda: integer): TSet;
    // P-
    public
      //   
      function GenPMedian_A (aPoly: integer;   //  (P)
                           aDirect: TCenter;   //   . 
                           var aCost: integer  //  ()
                          ): TSet;
      //   
      function GenPMedian (aPoly: integer;     //  (P)
                           aDirect: TCenter;   //   . 
                           var aCost: integer  //  ()
                          ): TSet;
    private
      function MedianCost: integer;     //   
      procedure MedianCopy(aRes: TSet); //    
    //  
    public
      function GenCoverTree(aMaxTree: boolean; var aCost: integer): TSet;
      function GenCoverDir(var aRoot: TNode; var aCost: integer): TSet;
      procedure ExpoAllCovers;
    // 
    public
      function CalcMaxFlow0(aSource, aDest : TNode): integer;
      function CalcMinFlow (aSource, aDest : TNode): integer;
      function CalcMaxFlow (aSource, aDest : TNode): integer;
      function CalcMinCostFlow (aSource, aDest : TNode;
                                aFlow: integer ): integer;
    // 
    public
      function GenPairs_Dissect(aMode: TPairs): TCostSet;
      function GenPairs_Edmonds(aMode: TPairs): TCostSet;
      function MarkMinPairsDicoty(var aPairs: integer): integer;
      function GenMinPairsDicoty: TCostSet;
      function GenMinPairsDicotyFlow: TCostSet;
    //  
    protected
      function GenPostPathUndir(var aCost: integer): TBuffer;
      function GenPostPathDir(var aCost: integer): TBuffer;
    public
      function GenPostPath(var aCost: integer): TBuffer;
    public
    //   ( )
      function GenHamPath(aStart, aFin: TNode; var aCost: integer): TBuffer;
      function GenHamPathDir(aStart, aFin: TNode; var aCost: integer): TBuffer;
      function GenHamPathStrip(aStart, aFin, aStrip: TNode;
                               var aCost: integer): TBuffer;
      function GenHamPath_Full(aStart, aFin: TNode; var aCost: integer): TBuffer;
      function GenHamPath_Greed(aStart, aFin: TNode;
                                var aCost: integer;
                                aGreed: boolean): TBuffer;

    //     ( )
      function GenHamilton_Full(var aCost: integer): TBuffer;
      function GenHamilton_Greed(var aCost: integer;
                                 aGreed: boolean): TBuffer;
      function GenHamilton(var aCost: integer): TBuffer;
      function GenHamilton_Quick(var aCost: integer): TBuffer;
    //     
      procedure MakeFull(aValue: integer);
      //    ()
      function GenTravelPath(aStart, aFin: TNode; var aCost: integer): TBuffer;
      //    ()
      function GenTravelCycle(var aCost: integer): TBuffer;
  end;

  //  ,    

  THugeGraph = class (TGraph)
    private
      function  GetLinkByNumber(aSource, aDest : integer): TLink;
      procedure MakeLink(aSource, aDest, aVal : integer);
    public
      constructor GenRandom(aDir: boolean;          // 
                            aLoadNodes, aLoadLinks, // 
                            aNodes,                 //  
                            aLinks                  //  ,%
                            : integer);
      constructor GenFull(aDir: boolean;            // 
                          aLoadNodes, aLoadLinks,   // 
                          aNodes: integer);         //  
      constructor GenDicoty(aLoadLinks,             //  
                            aPairs: integer);       //   
      function  GetNode(aNumber: integer): TNodeInt;
  end;

  //      

  TFarLink = class(TItem)   //  
    mNodeFar : TNode;       //  
    mNodeNear: TNode;       //      
    mDist    : integer;     //     
    mStep    : integer;     // ,    
    constructor Create(aNear, aFar: TNode; aDist: integer);
    function Compare(arg: TItem): TCompare; override;
    procedure Print(var aFile: TextFile);  override;
  end;

  //     ( )

  TCondensNode = class (TNode)
    public
      mName  : string;//  =      
      mNodes : TSet;  //  
      mOut   : TSet;  //      ()
      constructor Create(aNodes: TSet; aOwner: TGraph);
      destructor Destroy;  override;
      function  Copy: TItem; override;
      function  GetName: string;  override;
  end;

  //     

  TCostNodes = class (TCostSet)
    mNode: TNode;  // -  
    constructor Create(aCost: integer; aSet: TSet; aNode: TNode);
    procedure   Print(var aFile: TextFile); override;
  end;

  //     

function GenHugeGraph (aDir: boolean;          // 
                       aLoadNodes, aLoadLinks, // 
                       aNodes,                 //  
                       aLinks                  //  ,%
                       : integer): TGraph;

//////////////////////////////////////////////////////////////////////////
implementation

uses Assembly, SetList, SetUtils, SysUtils, DateUtils, GrChars;
//////////////////////////////////////////////////////////////////////////

const             // "" :
    CWhite = 0;   //  ( )
    CGray  = 1;   // 
    CBlack = 2;   // 
    CRed   = 3;   // 

    CInfinity = MaxInt;  // 

//////////////////////////////////////////////////////////////////////////
//     
//////////////////////////////////////////////////////////////////////////

function Minimum(a1, a2 : integer): integer;
begin
  if a1 < a2 then Result:= a1 else Result:= a2;
end;

//////////////////////////////////////////////////////////////////////////
//    
//     
//////////////////////////////////////////////////////////////////////////

function NodesToStr(aSet: TSet): string;
var N: TNode;
begin
  Result:= '';
  N:= aSet.GetFirst as TNode;
  while Assigned(N) do begin
    if Length(Result)<>0 then Result:= Result+',';
    Result:= Result + N.GetName;
    N:= aSet.GetNext as TNode;
  end;
end;

//////////////////////////////////////////////////////////////////////////
// TNode --  () 
//////////////////////////////////////////////////////////////////////////

constructor TNode.Create(aVal: integer; aOwner: TGraph);
begin
  inherited Create;             //  
  mOwner:= aOwner;              //   ()
  mBlossom:= Self;              //   
  mLnkOut:= CreateSet;          //    
  if Assigned(mOwner) then begin
    if mOwner.mLoadNodes        //   
      then mValue:= aVal;       //    (  = 0)
    if mOwner.mDirect           //  
      then mLnkIn := CreateSet; //      
  end
end;

destructor TNode.Destroy;
begin
  //   :
  mLnkOut.ClrAndDestroy;   //  
  mLnkOut.Free;            //   
  mLnkIn.Free;             //   
  //  :
  if Assigned(mFarLinks) then begin
    mFarLinks.ClrAndDestroy;
    mFarLinks.Free;
  end;
  inherited;
end;

//  :

function TNode.Compare(arg: TItem): TCompare;
begin
  Result:= cmpEq;
  if arg = Self then Exit;
  if GetName < (arg as TNode).GetName
    then Result:= cmpLess
  else if GetName > (arg as TNode).GetName
    then Result:= cmpGreate
end;

//      ,
//   

function TNode.GetName: string;
var N: integer;
begin
  N:= integer(Self) mod 99;
  Result:='x'+Char(Ord('0')+N div 10)+Char(Ord('0')+N mod 10);
end;

//      ( )

procedure TNode.SetLinksLimit(aLimit: integer);
var L : TLink;
begin
  L:= OutLinkFirst;
  while Assigned(L) do begin
    L.mLimit:= aLimit;
    L:= OutLinkNext;
  end;
end;

//      
//     

function TNode.GetFarLink(aFar: TNode): TFarLink;
var  FL : TFarLink; //   
begin
  Result:= nil;
  if not Assigned(mFarLinks) then Exit;
  mFarLinks.PositionPush;
  FL:= mFarLinks.GetFirst as TFarLink;   //  
  while Assigned(FL) do begin            // 
    if FL.mNodeFar = aFar then begin     //  
      Result:= FL;                       //  
      Break;
    end;
    FL:= mFarLinks.GetNext as TFarLink;  //  
  end;
  mFarLinks.PositionPop;
end;

//        

function TNode.GetNear(aFar: TNode): TNode;
var  FL : TFarLink; //   
begin
  Result:= nil;
  if not Assigned(mFarLinks) then Exit;
  mFarLinks.PositionPush;
  FL:= mFarLinks.GetFirst as TFarLink;   //  
  while Assigned(FL) do begin            // 
    if FL.mNodeFar = aFar then begin     //  
      Result:= FL.mNodeNear;             //  
      Break;
    end;
    FL:= mFarLinks.GetNext as TFarLink;  //  
  end;
  mFarLinks.PositionPop;
end;

//      

procedure TNode.SortFarLinks;
var NewList : TSet;
begin
  if not Assigned(mFarLinks) then Exit;
  NewList:= CreateSet;
  NewList.CopyItems(mFarLinks);
  mFarLinks.Free;
  mFarLinks:= NewList;
end;

//      ,   .

procedure TNode.ResetNode;
begin
  mColor  := 0;       // 
  mPred   := nil;     //    
  mLink   := nil;     //     
  mDist   := MaxInt;  //    
  mFlow   := MaxInt;  //  
  mRoot   := nil;     //  
  mBlossom:= Self;    //   
  mPair   := nil;     //   
  mRight  := nil;     //     
  mLeft   := nil;     //     
  mOut    := false;   // false -- , true --    
end;

//        
// aDest - -
// aVal  - "" 

function TNode.MakeLink(aDest: TNode; aVal: integer): TLink;
begin
  Result:= nil;
  if not Assigned(aDest) then Exit;
  Result:= GetLink(aDest);            //  ?
  if Assigned(Result) then Exit;      // , 
  // ,  
  Result:= TLink.Create(Self, aDest, aVal, mOwner.mLoadLinks);
  //    
  if not mLnkOut.Insert(Result) then begin
    Result.Free;
    Result:= nil;
    Exit;
  end;
  if mOwner.mDirect //  
    //      
    then aDest.mLnkIn.Insert(Result);
end;

//    

procedure TNode.RemoveLink(aDest: TNode);
var  L: TLink;
begin
  if not Assigned(aDest) then Exit;
  OutPosPush;
  L:= OutLinkFirst;
  while Assigned(L) and (L.mDest <> aDest) do  L:= OutLinkNext;
  OutPosPop;
  if not Assigned(L) then Exit;
  mLnkOut.Delete(L);
  if mOwner.mDirect  //  
    //      
    then aDest.mLnkIn.Delete(L);
  L.Free;
end;

//     

function TNode.GetLink(aDest: TNode): TLink;
begin
  OutPosPush;
  Result:= OutLinkFirst;
  while Assigned(Result) and (Result.mDest <> aDest)
    do  Result:= OutLinkNext;
  OutPosPop;
end;

//    
//     
// , . 33

function TNode.GenStrongArea: TSet;
var Temp: TSet;
begin
  Result:= GenGammaOut;              //   
  if not mOwner.mDirect then Exit;   // ,    
  Temp:= GenGammaIn;                 // ,   
  Result.Mul(Temp);                  //  
  Temp.Free;                         // 
end;


//       

//    

function TNode.OutLinkFirst: TLink;
begin
  Result:= mLnkOut.GetFirst as TLink;
end;

//    

function TNode.OutLinkNext: TLink;
begin
  Result:= mLnkOut.GetNext as TLink;
end;

//   

procedure TNode.OutPosPush;
begin
  mLnkOut.PositionPush;
end;

//   

procedure TNode.OutPosPop;
begin
  mLnkOut.PositionPop;
end;

//     

function TNode.OutGetLink(aIndex: integer): TLink;
begin
  Result:= TLink(mLnkOut.GetItem(aIndex));
end;

// -  

function TNode.OutGetCnt: integer;
begin
  Result:= 0;
  if Assigned(mLnkOut) then Result:= mLnkOut.GetCount;
end;

//      aSet

procedure TNode.OutGammaAdd(aRes : TSet);
var L: TLink;
begin
  L:= OutLinkFirst;            //   
  while Assigned(L) do begin   //  
    aRes.Insert(L.mDest);      //     
    L:= OutLinkNext;           //   
  end;
end;

//   

procedure TNode.OutGammaGet(aRes: TSet);
begin
  aRes.Clear;
  OutGammaAdd(aRes);
end;

//   ,   

function TNode.GenGammaOut: TSet;
var Node: TNode;
    Temp: TSet;
begin
  Result:= CreateSet;
  Result.Insert(Self);        //     
  Temp:= CreateSet;
  repeat
    Temp.CopyItems(Result);           //   
    Node:= Temp.GetFirst as TNode;    //   
    while Assigned(Node) do begin
      Node.OutGammaAdd(Result);       //   
      Node:= Temp.GetNext as TNode;
    end;
  until Result.GetCount = Temp.GetCount;  //   
  Temp.Free;
end;

//       

function TNode.CalcCost: integer;
var L: TLink;
begin
  Result:=1;
  if mOwner.mLoadNodes then Result:= mValue;
  if not mOwner.mLoadLinks then Exit;
  //      ()
  L:= OutLinkFirst;
  while Assigned(L) do begin
    Inc(Result, L.mValue);
    L:= OutLinkNext;
  end;
end;

//       

//    

function TNode.InLinkFirst: TLink;
begin
  if mOwner.mDirect
    then Result:= mLnkIn.GetFirst as TLink
    else Result:= OutLinkFirst
end;

//    

function TNode.InLinkNext: TLink;
begin
  if mOwner.mDirect
    then Result:= mLnkIn.GetNext as TLink
    else Result:= OutLinkNext
end;

//   

procedure TNode.InPosPush;
begin
  if mOwner.mDirect
    then mLnkIn.PositionPush
    else OutPosPush
end;

//   

procedure TNode.InPosPop;
begin
  if mOwner.mDirect
    then mLnkIn.PositionPop
    else OutPosPop
end;

//     

function TNode.InGetLink(aIndex: integer): TLink;
begin
  if mOwner.mDirect
    then Result:= TLink(mLnkIn.GetItem(aIndex))
    else Result:= OutGetLink(aIndex);
end;

//      aSet

procedure TNode.InGammaAdd(aRes: TSet);
var L: TLink;
begin
  L:= InLinkFirst;
  while Assigned(L) do begin
    aRes.Insert(L.mOwner);
    L:= InLinkNext;
  end;
end;

//   

function TNode.InGetCnt: integer;
begin
  Result:= 0;
  if mOwner.mDirect
    then begin
      if Assigned(mLnkIn) then Result:= mLnkIn.GetCount
    end
    else Result:= OutGetCnt
end;

//  TRUE,      
//   

function TNode.IsLeft: boolean;
begin
  Result:= mLnkOut.GetCount <> 0;
end;

//  TRUE,      
//   

function TNode.IsRight: boolean;
begin
  Result:= mLnkIn.GetCount <> 0;
end;

//    ( )   

procedure TNode.PrintLinks(var aFile: TextFile);
begin
  if Assigned(mLnkOut) then mLnkOut.Print(aFile);
end;

//    ( ) 

procedure TNode.Print(var aFile: TextFile);
begin
  Write(aFile, GetName+' ');
  if Assigned(mOwner)
    then if mOwner.mLoadNodes then begin
      if mOut
        then Write(aFile,'=', mValue:4 )
        else Write(aFile,'-', mValue:4 );
    end;
  Write(aFile, ' -> ');
  PrintLinks(aFile);
end;

//    ( )   

procedure TNode.PrintFarLinks(var aFile: TextFile);
var FL: TFarLink;
begin
  Print(aFile);
  with mFarLinks do begin
    PositionPush;
    FL:= GetFirst as TFarLink;
    while Assigned(FL) do begin
      FL.Print(aFile);
      FL:= GetNext as TFarLink;
    end;
    PositionPop;
  end;
end;

//    ( )

procedure TNode.ExpoData;
var L: TLink;
    Flag: boolean;
begin
  OutPosPush;
  Flag:= false;
  //  :
  L:= OutLinkFirst;
  while Assigned(L) do begin
    with L do begin
      Write(mOwner.GetName+'->'+mDest.GetName, '  C=', mValue:2);
      //   :
      Writeln('  L=', mLow:2,  '  H=', mHigh:2,
              '  F=', mFlow:2, '  T=', mTemp:2);
    end;
    Flag:= true;
    L:= OutLinkNext;
  end;
  if Flag then Writeln;
  OutPosPop;
end;

//    ( )

procedure TNode.ExpoFarLinks;
begin
  PrintFarLinks(Output);
end;

//   ,    

function TNode.GenGammaIn: TSet;
var Node: TNode;
    Temp: TSet;
begin
  if not mOwner.mDirect then begin
    //       
    Result:= GenGammaOut;
    Exit;
  end;
  Result:= CreateSet;
  Result.Insert(Self);        //     
  Temp:= CreateSet;
  repeat
    Temp.CopyItems(Result);           //   
    Node:= Temp.GetFirst as TNode;    //   
    while Assigned(Node) do begin
      Node.InGammaAdd(Result);        //   
      Node:= Temp.GetNext as TNode;
    end;
  until Result.GetCount = Temp.GetCount;  //   
  Temp.Free;
end;

// #########      #########

//   ()  

function TNode.GetBlossom: TNode;
begin
  //    
  if mBlossom <> Self
    then Result:= mBlossom.GetBlossom
    else Result:= Self;
end;

//      ,
//    ()
//  :
// mPair   : TLink;     --  
// mOwnDest  : boolean; --     :
//                      false -- , true -- 

function TNode.GetPair: TNode;
begin
  Result:= nil;
  if not Assigned(mPair) then Exit;
  if mOwnDest
    then Result:= mPair.mOwner  //  
    else Result:= mPair.mDest   //  
end;

//////////////////////////////////////////////////////////////////////////
// TNodeInt -- ,  
//////////////////////////////////////////////////////////////////////////

constructor TNodeInt.Create(aNumber, aVal: integer; aOwner: TGraph);
begin
  inherited Create(aVal, aOwner);
  mNumber:= aNumber;
end;


function TNodeInt.Compare(arg: TItem): TCompare;
begin
  Result:= cmpEq;
  if arg = Self then Exit;
  Result:= cmpLess;
  if (arg is TNodeInt) and (mNumber > (arg as TNodeInt).mNumber)
    then Result:= cmpGreate
end;


function TNodeInt.Copy: TItem;
begin
  Result:= TNodeInt.Create(mNumber, mValue, mOwner);
end;

//       

function TNodeInt.GetLinkByNumber(aNumber: integer): TLink;
begin
  OutPosPush;
  Result:= OutLinkFirst;
  while Assigned(Result) do begin
    if (Result.mDest as TNodeInt).mNumber = aNumber then Break;
    Result:= OutLinkNext;
  end;
  OutPosPop;
end;

function TNodeInt.GetName: string;
begin
  Result:= IntToStr(mNumber);
  while Length(Result) < 3 do Insert('0', Result, 1);
end;

//////////////////////////////////////////////////////////////////////////
// TLink -    
//////////////////////////////////////////////////////////////////////////

constructor TLink.Create(aOwner, aDest: TNode; aVal: integer;
                         aLoadLinks: boolean);
begin
  inherited Create;         //  
  mOwner:= aOwner;          // -    
  mDest:= aDest;            // - 
  mLoadLinks:= aLoadLinks;  //   
  if aLoadLinks             //  
    then mValue:= aVal      //   ""
    else mValue:= 1;        //     = 1
end;

//   

function TLink.GetReverse: TLink;
begin
  Result:= mDest.GetLink(mOwner);
end;

function TLink.Copy: TItem;
begin
  Result:= TLink.Create(nil, nil, mValue, mLoadLinks);
end;


function TLink.Compare(arg: TItem): TCompare;
begin
  Result:= inherited Compare(arg);
  if Result = cmpEq then Exit;
  //   
  Result:= mOwner.Compare((arg as TLink).mOwner);
  if Result in [cmpLess, cmpGreate] then Exit;
  //   
  Result:= mDest.Compare((arg as TLink).mDest);
end;

//     

procedure TLink.Print(var aFile: TextFile);
begin
  Write(aFile, ' '+mDest.GetName);
  if mLoadLinks then Write(aFile, '=', mValue:2);
end;

//  ,   

procedure TLink.Print2(var aFile: TextFile);
var N1, N2 : TNode;
begin
  N1:=mOwner;  N2:=mDest;
  if not mOwner.mOwner.mDirect and
     (mOwner.Compare(mDest) = cmpGreate) then begin
    N1:=mDest;   N2:=mOwner;
  end;
  Write(aFile, ' ' + N1.GetName + '-' + N2.GetName);
  if mLoadLinks then Write(aFile, '=', mValue:3, '':1);
end;

//////////////////////////////////////////////////////////////////////////
// TGraph  --    
//////////////////////////////////////////////////////////////////////////

constructor TGraph.Create(const aName: string;
                          aDir, aLNodes, aLLinks: boolean);
begin
  inherited Create;         //   
  mName:= aName;            //  
  mDirect := aDir;          //  
  mLoadNodes := aLNodes;    //   
  mLoadLinks := aLLinks;    //   
  mNodes:= CreateSet;       //   
  mMapDirect:= Undefined;   //     
end;


destructor TGraph.Destroy;
begin
  DoneMap;               //    
  mNodes.ClrAndDestroy;  //   
  mNodes.Free;           //  
  inherited;
end;

//  :

//   

function TGraph.Nodes: integer;
begin
  Result:= mNodes.GetCount;
end;

//   

function TGraph.NodeFirst: TNode;
begin
  Result:= mNodes.GetFirst as TNode;
end;

//   

function TGraph.NodeNext: TNode;
begin
  Result:= mNodes.GetNext as TNode;
end;

//   

procedure TGraph.PosPush;
begin
  mNodes.PositionPush;
end;

//   

procedure TGraph.PosPop;
begin
  mNodes.PositionPop;
end;

//      

procedure TGraph.ResetNodes;
var N : TNode;
begin
  PosPush;
  N:= NodeFirst;
  while Assigned(N) do begin
    N.ResetNode;
    N:= NodeNext;
  end;
  PosPop;
end;

//      (nil -   )

function TGraph.GetNode(aNode: TNode): TNode;
begin
  Result:= mNodes.GetObject(aNode) as TNode;
end;

//  

function TGraph.Compare(arg: TItem): TCompare;
begin
  Result:= inherited Compare(arg);
  if Result <> cmpEq then Result:= cmpGreate;
end;

//  :     
//     

procedure TGraph.Compress;
var  N: TNode;
     L: TLink;
    FL: TFarLink;
begin
  // C    :
  InitMap_Floyd;
  //     :
  N:= NodeFirst;
  while Assigned(N) do begin
    //      N:
    FL:= N.mFarLinks.GetFirst as TFarLink;
    while Assigned(FL) do begin
      if (FL.mNodeFar <> N) then begin
        //      N --> mNodeFar
        L:= N.GetLink(FL.mNodeFar);
        //   ,     
        if Assigned(L) and (L.mValue > FL.mDist)
            //   
            then L.mValue:= FL.mDist;
      end;
      FL:= N.mFarLinks.GetNext as TFarLink;
    end;
    N:= NodeNext;
  end;
end;

//    

procedure TGraph.Print(var aFile: TextFile);
begin
  Writeln(aFile, mName);
  mNodes.Print(aFile);
end;

//  ,  FALSE    

function TGraph.InsertNode(aNode: TNode): boolean;
begin
  Result:= mNodes.Insert(aNode)
end;

//      
//  ,   !

procedure TGraph.RemoveNode(aNode: TNode);
var LS, LD: TLink;
    N: TNode;
begin
  if not mNodes.Exist(aNode) then Exit;
  //  :    
  if mDirect then begin //  
    //    ,    
    //     :
    LS:= aNode.OutLinkFirst;
    while Assigned(LS) do begin
      N:= LS.mDest;               //  
      N.mLnkIn.Delete(LS);        //     
      LS.Free;
      LS:= aNode.OutLinkNext;
    end;
    //     :
    LS:= aNode.InLinkFirst;
    while Assigned(LS) do begin
      N:= LS.mOwner;              //  
      N.mLnkOut.Delete(LS);       //     
      LS.Free;                    //   
      LS:= aNode.InLinkNext;
    end;
  end else begin //  
    //    ,    mLnkOut
    //     :
    LS:= aNode.OutLinkFirst;
    while Assigned(LS) do begin
      N:= LS.mDest;               //  
      //         :
      LD:= N.OutLinkFirst;
      while Assigned(LD) and (LD.mDest <> aNode)
        do LD:= N.OutLinkNext;
      N.mLnkOut.Delete(LD);       //   
      LD.Free;                    //   
      LS.Free;                    //    
      LS:= aNode.OutLinkNext;
    end;
  end;  // if
  //   
  with aNode do begin
    mLnkOut.Clear;
    if Assigned(mLnkIn) then mLnkIn.Clear;
  end;
  mNodes.Delete(aNode);    //    
end;

//     
// aSource - -
// aDest   - -
// aVal    - ""   

procedure TGraph.SetLink(aSource, aDest: TNode; aVal: integer);
begin
  if not Assigned(aSource) or not Assigned(aDest) then Exit;
  aSource.MakeLink(aDest, aVal);         // .  
  if not mDirect                         //   ,
    then aDest.MakeLink(aSource, aVal);  //  .  
end;

//      

procedure TGraph.SetLinksLimit(aLimit: integer);
var N: TNode;
begin
  N:= NodeFirst;
  while Assigned(N) do begin
    N.SetLinksLimit(aLimit);
    N:= NodeNext;
  end;
end;

//########################################################################
//                                                                       #
//                                   #
//                                                                       #
//########################################################################

//        

function TGraph.CopyGr(aRevers: boolean): TGraph;
var  N1, N2 : TNode;
         ND : TNode;
          L : TLink;
begin
  Result:= TGraph.Create(mName, mDirect, mLoadNodes, mLoadLinks);
  //  
  N1:= NodeFirst;
  while Assigned(N1) do begin
    N2:= TNode(N1.Copy);
    Result.InsertNode(N2);
    N1:= NodeNext;
  end;
  //  
  N1:= NodeFirst;   //   
  while Assigned(N1) do begin
    N2:= Result.GetNode(N1);     //  
    L:= N1.OutLinkFirst;         //   
    while Assigned(L) do begin
      ND:= Result.GetNode(L.mDest);  //   
      //       
      if aRevers
        then Result.SetLink(ND, N2, L.mValue)
        else Result.SetLink(N2, ND, L.mValue);
       L:= N1.OutLinkNext;  //  
    end;
    N1:= NodeNext;   //   
  end;
end;

//       

function TGraph.Copy: TItem;
begin
  Result:= CopyGr(false);
end;

//     

function TGraph.CopyByNodes(aNodes: TSet): TGraph;
var  N1, N2 : TNode;    //     
         ND : TNode;    //   
          L : TLink;
begin
  Result:= TGraph.Create('CopyByNodes', mDirect, mLoadNodes, mLoadLinks);
  //  :
  N1:= NodeFirst;                   //   
  while Assigned(N1) do begin
    if aNodes.Exist(N1) then begin  //     
      N2:= N1.Copy as TNode;        //   
      Result.InsertNode(N2);        //    
    end;
    N1:= NodeNext;
  end;
  //  :
  N1:= NodeFirst;                   //   
  while Assigned(N1) do begin
    if aNodes.Exist(N1) then begin  //     
      N2:= Result.GetNode(N1);      //  
      L:= N1.OutLinkFirst;          //   
      while Assigned(L) do begin    //   
        ND:= Result.GetNode(L.mDest);     //   
        Result.SetLink(N2, ND, L.mValue); //  
        L:= N1.OutLinkNext;         //  
      end;
    end;
    N1:= NodeNext;   //   
  end;
end;

//      ()

function TGraph.CopyByLinks(aLinks: TSet): TGraph;
var  NS, ND : TNode;
          L : TLink;
begin
  Result:= TGraph.Create('CopyByLinks', mDirect, mLoadNodes, mLoadLinks);

  //    ,   

  L:= TLink(aLinks.GetFirst);      //    
  while Assigned(L) do begin
    NS:= Result.GetNode(L.mOwner); //     
    if not Assigned(NS)            //  
      then begin
        NS:= TNode(L.mOwner.Copy); //  
        Result.InsertNode(NS);     //  
      end;
    ND:= Result.GetNode(L.mDest);  //     
    if not Assigned(ND)            //  
      then begin
        ND:= TNode(L.mDest.Copy);  //  
        Result.InsertNode(ND);     //  
      end;
    Result.SetLink(NS, ND, L.mValue); //  
    L:= TLink(aLinks.GetNext); //    
  end;
end;


//########################################################################
//                                                                       #
//          ,                      #
//                                                                       #
//########################################################################

//  TRUE      arg1   arg2
// (  )

function TGraph.TestLink(arg1, arg2: TNode): boolean;
var SN1 : TSet;    //   
    SN2 : TSet;    //     ()
   node : TNode;   //    SN1
    old : integer; //   
      i : integer;
begin
  SN1:= CreateSet;  SN2:= CreateSet;
  SN1.Insert(arg1); //      
  repeat
    Result:= SN1.Exist(arg2);   //    ?
    if Result then Break;       // ,   
    old:= SN2.GetCount;         //  
    //   SN2 := Gamma(SN1)
    for i:= 1 to SN1.GetCount do begin
      node:= TNode(SN1.GetItem(i)); // node -   
      node.OutGammaAdd(SN2);        // SN2:= SN2 + Gamma(node)
    end;
    if SN2.GetCount = old then Break;  // ,   
    SN1.CopyItems(SN2);   // SN1 := SN2
  until false;
  SN2.Free;  SN1.Free;
end;

//      arg1   arg2
//      
// (  )

function TGraph.CalcSteps(arg1, arg2: TNode): integer;
var SN1 : TSet;     //   
    SN2 : TSet;     //     ()
   node : TNode;    //    SN1
     ok : boolean;  //    
    old : integer;  //    
      i : integer;
begin
  Result:= 0;
  SN1:= CreateSet;  SN2:= CreateSet;
  SN1.Insert(arg1); //      
  repeat
    ok:= SN1.Exist(arg2);  //    
    if ok then Break;
    old:= SN2.GetCount;    //    
    //  :  SN2 := Gamma (SN1)
    for i:= 1 to SN1.GetCount do begin
      node:= TNode(SN1.GetItem(i)); // node -   
      node.OutGammaAdd(SN2);        // SN2:= SN2 + Gamma(node)
    end;
    if SN2.GetCount = old then Break;  // ,    
    Inc(Result);
    SN1.CopyItems(SN2);   // SN1 := SN2
  until false;
  SN2.Free;  SN1.Free;
  if not ok then Result:= -1;   //  
end;

//      
//       (  )
// aSource -  
// aDest -  

function TGraph.GenPath(aSource, aDest: TNode): TBuffer;

var  Que : TBuffer;  //  
    Node : TNode;      //  
    Link : TLink;      //  
      OK : boolean;    //    
begin
  Result:= nil;
  if aSource = aDest then Exit;  //   
  Que:= TBuffer.Create;  //    ()
  ResetNodes;              //     
  aSource.mColor:= CGray;  //  = 
  Que.Put(aSource);        //   
  // ,    :
  repeat
    OK:= false;                   //    
    Node:= TNode(Que.Get);        //   
    Link:= Node.OutLinkFirst;     //   
    while Assigned(Link) do begin //   
      with Link do if mDest.mColor = CWhite then begin
        //     ()
        mDest.mPred:= Node;       //   
        OK:= mDest = aDest;       // ?
        if OK then Break;         //    while
        mDest.mColor:= CGray;     //   
        Que.Put(mDest);           //    
      end;
      Link:= Node.OutLinkNext; //   
    end; // while
    Node.mColor:= CBlack;      //    
    // ,      
  until OK or (Que.GetCount = 0);

  //     Que
  if Assigned(aDest.mPred) then begin //   
    Que.Clear;                        //  
    Que.Put(aDest);                   //    
    Node:= aDest.mPred;               //    
    while Assigned(Node) do begin     //  
      Que.Put(Node);                  //  -->  
      Node:= Node.mPred;              //   
    end;
    Que.Reversion;                    //  
    Result:= Que;
  end else begin
    Que.Free;       //    ,  
  end;
end;

//      
//         
// (  )

function TGraph.GenDijkstra(aSource, aDest: TNode): TBuffer;

var  Que : TBuffer;  //  
    Node : TNode;      //  
    Link : TLink;      //  

begin
  Result:= nil;
  if aSource = aDest then Exit;  //   
  Que:= TBuffer.Create;  //    ()

  ResetNodes;              //     
  aSource.mColor:= CGray;  //  = 
  aSource.mDist:= 0;       //    = 0
  Que.Put(aSource);        //   

  repeat                   // ,    
    Node:= TNode(Que.Get); //   
    Link:= Node.OutLinkFirst; //   
    while Assigned(Link) do with Link do begin  // .  
      if mDest.mDist > Node.mDist + Ord(mLoadNodes)*Node.mValue + mValue
        then begin
        //      ,
        //     
        mDest.mDist:= Node.mDist + Ord(mLoadNodes)*Node.mValue + mValue;
        mDest.mPred:= Node;
        //      ,   ,
        //      (,  )
        if mDest.mColor <> CGray then mDest.mColor:= CWhite;
      end;
      if mDest.mColor = CWhite then begin
        //      
        mDest.mColor:= CGray;   //   
        Que.Put(mDest);         //    
      end;
      Link:= Node.OutLinkNext;  //   
    end; // while
    Node.mColor:= CBlack;   // ,   
  until Que.GetCount = 0;   // ,   

  //     Que
  if Assigned(aDest.mPred) then begin //   
    Que.Clear;                        //  
    Que.Put(aDest);                   //    
    Node:= aDest.mPred;               //    
    while Assigned(Node) do begin     //  
      Que.Put(Node);                  //  -->  
      if Node = aSource then Break;   // ,   
      Node:= Node.mPred;              //   
    end;
    Que.Reversion;                    //  
    Result:= Que;
  end else begin
    Que.Free;   //    ,  
  end;
end;

//    
//     -
//     
// (  )
// , "     ", . 53

// TFarLink -    

constructor TFarLink.Create(aNear, aFar: TNode; aDist: integer);
begin
  inherited Create;
  mNodeNear:= aNear;  //  
  mNodeFar:= aFar;    //  
  mDist:= aDist;      //    
end;

//     

function TFarLink.Compare(arg: TItem): TCompare;
begin
  if mDist > (arg as TFarLink).mDist
    then Result:= cmpGreate
  else if mDist < (arg as TFarLink).mDist
    then Result:= cmpLess
  else Result:= mNodeFar.Compare((arg as TFarLink).mNodeFar)
end;


procedure TFarLink.Print(var aFile: TextFile);
begin
  Write(aFile,'Far  = ');          //  
  if Assigned(mNodeFar)
    then mNodeFar.Print(aFile)
    else Writeln('NIL');
  Write(aFile,'Near = ');          //  
  if Assigned(mNodeNear)
    then mNodeNear.Print(aFile)
    else Writeln('NIL');           //    
  Writeln(aFile,'Dist = ', mDist);
  Writeln(aFile,'- - - - - - -');
end;

//       -

procedure TGraph.InitMap_Floyd;

  //- - - - - - - - - - - - - - - - - - - - -
  //     
  procedure InitFarLinks(aNode: TNode);
  var     Node : TNode;     //  
          Link : TLink;     //  
            FL : TFarLink;  //  
  begin
    PosPush;
    Node:= NodeFirst;
    while Assigned(Node) do begin   //    
      if Node = aNode then begin
        FL:= TFarLink.Create(Node, Node, 0);   //   
      end else with aNode do begin
        FL:= TFarLink.Create(nil {near}, Node {far}, CInfinity);
        Link:= OutLinkFirst;            //   
        while Assigned(Link) do begin   //   
          if Link.mDest = Node then begin
            //     
            FL.mNodeNear:= Node;
            FL.mDist:= Link.mValue; // + Ord(mLoadNodes)*Node.mValue;
            FL.mStep:= 1;      //      
            Break;
          end;
          Link:= OutLinkNext;     //   
        end; // while
      end; // else
      aNode.mFarLinks.Insert(FL);        //   
      Node:= NodeNext;
    end;
    PosPop;
  end;
  //- - - - - - - - - - - - - - - - - - - - -
  //     (  )

  function  Handle(aNode: TNode; aStep: integer): boolean;
  var   FL1 : TFarLink;  //      
        FL2 : TFarLink;  //      

    //      FL2

    function Test_FL2(aNear : TNode): boolean;
    var  Dist : integer;  //  
         FL : TFarLink;   //     aNode
    begin
      Result:= false;
      Dist:= FL1.mDist + FL2.mDist;  //    
      FL:= aNode.GetFarLink(FL2.mNodeFar); //    
      if FL.mDist > Dist then begin    //    
        FL.mNodeNear:= FL1.mNodeNear;  //   
        FL.mDist:= Dist;               //  
        FL.mStep:= aStep+1;  //       
        Result:= true;       //    
      end;
    end;

  begin  // Handle
    Result:= false;
    //    aNode
    FL1:= aNode.mFarLinks.GetFirst as TFarLink;
    while Assigned(FL1) do begin
      //      
      if FL1.mStep=aStep then with FL1.mNodeFar.mFarLinks do begin
        PositionPush;
        FL2:= GetFirst as TFarLink;   //    
        while Assigned(FL2) do begin
          with FL2 do  //      
          if Assigned(mNodeNear)            //  
             and (mDist<>0)                 //     
             and (mNodeFar<>aNode)          //     
            then if Test_FL2(FL2.mNodeNear) //   
              then Result:= true;           //  ,  
          FL2:= GetNext as TFarLink;        //   
        end; // while
        PositionPop;
      end;
      FL1:= aNode.mFarLinks.GetNext as TFarLink;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - -

var  Node : TNode;    //  
     Step : integer;  //   ( )
     Flag : boolean;  //   

begin { InitMap_Floyd }

  if mInitMap then Exit;  //    
  mInitMap:= true;        // ,   

  //   :
  Node:= NodeFirst;
  while Assigned(Node) do begin        //  
    with Node do begin
      if Assigned(mFarLinks)           //   
        then mFarLinks.ClrAndDestroy   // 
        else mFarLinks:= CreateSet;    //   
    end;
    InitFarLinks(Node);   //   
    Node:= NodeNext;
  end;

  //  
  for Step:= 1 to mNodes.GetCount-2  do begin
    Flag:= false;
    Node:= NodeFirst;
    while Assigned(Node) do begin     //  
      if Handle(Node, Step)           //    
        then Flag:= true;             //  
      Node:= NodeNext;
    end;
    if not Flag then Break;  //    ,  
  end;
end;

//      
//   

function TGraph.GenQuickPath(aSource, aDest: TNode): TBuffer;
var Next: TNode;

begin
  Result:= nil;
  InitMap_Floyd; //     ,   

  //    :
  Next:= aSource.GetNear(aDest);
  if not Assigned(Next) then Exit;

  Result:= TBuffer.Create;   // .  
  Result.Put(aSource);       //   
  //   :
  while Next <> aDest do begin
    Result.Put(Next);        //   
    Next:= Next.GetNear(aDest);
  end;
  Result.Put(Next);   //   
end;

//      
//   

function TGraph.GetQuickPathStr(aSource, aDest: TNode): String;
var Next : TNode;
      FL : TFarLink;
       S : String;
begin
  Result:= aSource.GetName;
  InitMap_Floyd;//     ,   

  //    :
  Next:= aSource.GetNear(aDest);
  if not Assigned(Next) then Exit;   //   
  //   :
  repeat
    Result:= Result + ' -> ' + Next.GetName;
    Next:= Next.GetNear(aDest);
  until Next = aDest;
  FL:= aSource.GetFarLink(aDest); //    
  Str(FL.mDist, S);               //   
  Result:= Result + ' -> ' + Next.GetName + ' : ' + S;
end;

//########################################################################
//                                                                       #
//                                                       #
//              ,  2 . 29                             #
//                                                                       #
//########################################################################

// TCondensNode --  
//     
// (   )

constructor TCondensNode.Create(aNodes: TSet; aOwner: TGraph);
var Name : string;
     Val : integer;
      N  : TNode;
begin
  Name:='';  Val:= 0;
  //      ()
  N:= TNode(aNodes.GetFirst);
  while Assigned(N) do begin
    Name:= Name+ N.GetName;      //  
    Val:= Val + N.mValue;        //   
    N:= TNode(aNodes.GetNext);
  end;
  inherited Create(Val, aOwner);
  mName:= Name;
  //   
  mNodes:= TSet(aNodes.Copy);
  //  mOut -      ()
  mOut:= CreateSet;
  N:= TNode(mNodes.GetFirst);
  while Assigned(N) do begin
    N.OutGammaAdd(mOut);
    N:= TNode(mNodes.GetNext);
  end;
  mOut.Sub(mNodes);  //     
end;

destructor TCondensNode.Destroy;
begin
  mNodes.Free;
  mOut.Free;
  inherited;
end;

function TCondensNode.GetName: string;
begin
  Result:= mName;
end;

function TCondensNode.Copy: TItem;
begin
  Result:= nil;
end;

//   (  )

function TGraph.IsLinked: boolean;
var Node: TNode;
    S : TSet;
    NIn, NOut : integer;
begin
  Result:= false;
  //   :
  Node:= NodeFirst;
  if not Assigned(Node) then Exit;
  //  ,     Node:
  S:= Node.GenGammaIn;
  NIn:= S.GetCount;
  S.Free;
  //  ,     Node:
  S:= Node.GenGammaOut;
  NOut:= S.GetCount;
  S.Free;
  //     ,   
  //    :
  Result:= (mNodes.GetCount = NIn) and (mNodes.GetCount = NOut);
end;

//   ()  

function TGraph.CalcAreas: integer;
var Copy: TSet;    //     
    Node: TNode;   //  
    Strong: TSet;  //     
begin
  Result:= 0;
  //  -  
  Copy:= CreateSet;  Copy.CopyItems(mNodes);
  //  :
  repeat
    Inc(Result);                    //  
    Node:= Copy.GetFirst as TNode;  //    
    Strong:= Node.GenStrongArea;    //   
    Copy.Sub(Strong);               //   
    Strong.Free                     //  
  until Copy.GetCount = 0;
  Copy.Free;
end;

//   ()   ()

function TGraph.GenAreas: TSet;
var Copy: TSet;    //     
    Node: TNode;   //  
    Strong: TSet;  //     
    Gr: TGraph;    // ,     Strong
begin
  Result:= CreateSet;
  //  -  
  Copy:= CreateSet;  Copy.CopyItems(mNodes);
  //  :
  repeat
    Node:= Copy.GetFirst as TNode;  //    
    Strong:= Node.GenStrongArea;    //   
    Gr:= CopyByNodes(Strong);       //   
    Result.Insert(Gr);              //   
    Copy.Sub(Strong);               //   
    Strong.Free                     //  
  until Copy.GetCount = 0;          //    
  Copy.Free;
end;

//  -
// (  )
// , . 33

function TGraph.GenStrongAreas: TSet;
var Copy: TSet;    //     
    N : TNode;
    Strong: TSet;  //  
begin
  Result:= CreateSet;
  //  -  
  Copy:= CreateSet;  Copy.CopyItems(mNodes);
  //    :
  while Copy.GetCount <> 0 do begin
    N:= Copy.GetFirst as TNode;      //    
    Strong:= N.GenStrongArea;        //   
    Result.Insert(Strong);           //    
    Copy.Sub(Strong);                //    
  end;
  Copy.Free;  //  
end;

//  -
// , . 33

function TGraph.GenCondens: TGraph;
var Strongs : TSet;         //   
          S : TSet;         //  
         CN : TCondensNode; //  -
     Target : TCondensNode; //  -
begin
  Result:= TGraph.Create('Condens', true, false, false);  //  
  Strongs:= GenStrongAreas;               //  -
  S:= Strongs.GetFirst as TSet;           // .    
  while Assigned(S) do begin              //   
    CN:= TCondensNode.Create(S, Result);  //   
    Result.InsertNode(CN);                //    -
    S:= Strongs.GetNext as TSet;          // .    
  end;
  //     :
  S:= CreateSet;  // .    
  CN:= Result.NodeFirst as TCondensNode;        //   
  while Assigned(CN) do begin
    if CN.mOut.GetCount <> 0 then begin         //    <> 0
      Result.PosPush;                           // .  
      Target:= Result.NodeFirst as TCondensNode;//  
      while Assigned(Target) do begin           // .  
        if CN <> Target then begin              //    
          S.CopyItems(CN.mOut);                 //   
          S.Mul(Target.mNodes);                 //   
          if S.GetCount<>0                      //     . 
            then Result.SetLink(CN, Target, 1); //  .   
        end;
        Target:= Result.NodeNext as TCondensNode; // .  
      end;
      Result.PosPop;                           // .  
    end;
    CN:= Result.NodeNext as TCondensNode;      // .  
  end;
  S.Free;
  Strongs.ClrAndDestroy;  //  -
  Strongs.Free;
end;

//########################################################################
//                                                                       #
//                   #
//                   -2.3,  . 46                           #
//              ___ (Bron-Kerbosh)                  #
//                                                   #
//             -                 #
//                                                                       #
//########################################################################

function TGraph.GenUndepend: TSet;
var
  Res: TSet;    //    
  Gamma: TSet;  //     
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //  TRUE,     (aTested)
  //   ,       (aCand)

  function  Test(aCand, aTested: TSet): boolean;
  var Node: TNode;
  begin
    Result:= false;
    Node:= TNode(aTested.GetFirst);   //   
    while Assigned(Node) do begin
      Node.OutGammaGet(Gamma); // Gamma =  
      Gamma.Mul(aCand);        // Gamma =    
      if Gamma.GetCount=0 then begin //   ,
        Result:= true;       //       
        Break;
      end;
      Node:= TNode(aTested.GetNext);  //   
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  // aCand, aTested -    

  procedure Extend(aCand, aTested : TSet);
  var
    newCand : TSet;  //  -
    newTested: TSet; //   
    Node: TNode;     //  
  begin
    //    
    //      
    while (aCand.GetCount<>0) and not Test(aCand, aTested) do begin

      Node:= TNode(aCand.GetFirst);     //   
      Res.Insert(Node);                 //   
      Node.OutGammaGet(Gamma);          // Gamma=  
      //      
      newCand:= TSet(aCand.Copy);
      newTested:= TSet(aTested.Copy);

      newCand.Delete(Node);   //     
      newCand.Sub(Gamma);     //    
      newTested.Sub(Gamma);   //     

      if (newCand.GetCount=0) and (newTested.GetCount=0)
        then //        ,
             //    .  
             Result.Insert(Res.Copy)
        else if newCand.GetCount<>0 then
               //   ,   
               //      
               Extend(newCand, newTested);

      // 
      newTested.Free;       //   
      newCand.Free;         //   
      Res.Delete(Node);     //    
      aCand.Delete(Node);   //    
      aTested.Insert(Node); //    
    end; // while
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var
  Cand : TSet;   // -
  Tested: TSet;  //  

begin  { TGraph.GenUndepend}

  // 
  Result:= CreateSet;      //   -
  Res:= CreateSet;         //  .  
  Cand := CreateSet;       // 
  Cand.CopyItems(mNodes);  // Cand =   
  Tested:= CreateSet;      //  
  Gamma:= CreateSet;       //  

  Extend(Cand, Tested); // . .  . 
  //  
  Gamma.Free;
  Tested.Free;
  Cand.Free;
  Res.Free;
end;

//########################################################################
//                                                                       #
//                                  #
//                   -2.3,  . 46                           #
//               ___ ( Bron-Kerbos                  #
//                                                   #
//                                                                       #
//########################################################################

function TGraph.GenClique: TSet;
var
  Res: TSet;     //  
  Gamma : TSet;  //  
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //  TRUE,      aTested
  //   ,      aCand

  function  Test(aCand, aTested: TSet): boolean;
  var Node: TNode;
  begin
    Result:= false;
    Node:= TNode(aTested.GetFirst);
    while Assigned(Node) do begin
      Node.OutGammaGet(Gamma); // Gamma =  Node
      Gamma.Mul(aCand);        // Gamma =  * 
      if Gamma.GetCount = aCand.GetCount then begin
        //     
        Result:= true;
        Break;
      end;
      Node:= TNode(aTested.GetNext);
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    

  procedure Extend(aCand, aTested : TSet);
  var
    newCand : TSet;  // -
    newTested: TSet; //  
    Node: TNode;     //  
  begin
    //       aTested  ,
    //       aCand
    while (aCand.GetCount<>0) and not Test(aCand, aTested) do begin

      Node:= TNode(aCand.GetFirst); //  
      Res.Insert(Node);             //   
      Node.OutGammaGet(Gamma);      // Gamma:=  Node

      newCand:= TSet(aCand.Copy);   //  
      newCand.Delete(Node);         //   
      newCand.Mul(Gamma);           //    

      newTested:= TSet(aTested.Copy); //  
      newTested.Mul(Gamma);           //   

      if (newCand.GetCount=0) and (newTested.GetCount=0)
        then //        ,
             //    
             Result.Insert(Res.Copy)

        else if newCand.GetCount<>0 then
             //   ,   
             //      
             Extend(newCand, newTested);

      // 
      newTested.Free;       //   
      newCand.Free;         //   
      Res.Delete(Node);     //    
      aCand.Delete(Node);   //    
      aTested.Insert(Node); //    
    end
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var
  Cand : TSet;   // -
  Tested: TSet;  //  

begin   { TGraph.GenClique }

  // 
  Result:= CreateSet;      //   -
  Res:= CreateSet;         //  
  Cand := CreateSet;       // 
  Cand.CopyItems(mNodes);  // Cand =   
  Tested:= CreateSet;      //  
  Gamma:= CreateSet;       //  

  Extend(Cand, Tested);  //  
  //  
  Gamma.Free;
  Tested.Free;
  Cand.Free;
  Res.Free;

end;

//########################################################################
//                                                                       #
//                                     #
//                 -3,  . 50                               #
//                       +                                     #
//                                                                       #
//########################################################################

// TCostNodes -     

constructor TCostNodes.Create(aCost: integer; aSet: TSet; aNode: TNode);
begin
  inherited Create(aCost, aSet, true);
  mNode:= aNode;
end;


procedure TCostNodes.Print(var aFile: TextFile);
var Node: TNode;
begin
  Write(aFile, mNode.GetName, ' = ', mCost, ' -> { ');
  Node:= TNode(mSet.GetFirst);
  while Assigned(Node) do begin
    Write(aFile, Node.GetName + ' ');
    Node:= TNode(mSet.GetNext);
  end;
  Writeln(aFile, '}');
end;

//    

function TGraph.GenDominating: TCostSet;
var  Buf: TBuffer;   //  
     Node: TNode;      //  
     Gamma: TSet;      //   
     CS: TCostNodes;   //  

begin
  Buf:= TBuffer.Create;
  //      ,
  //       
  Node:= NodeFirst;
  while Assigned(Node) do begin
    Gamma:= CreateSet;        //   
    Node.OutGammaGet(Gamma);  // Gamma=  
    Gamma.Insert(Node);       // + 
    // Node.CalcCost    :=   +  
    CS:= TCostNodes.Create(Node.CalcCost, Gamma, Node);
    Buf.Put(CS);         //   
    Node:= NodeNext;
  end;
  //   
  Result:= CollectMinCover(mNodes, Buf);  // unit Assembly
  //     {unit SetUtils}
  RemoveItemsFromBuf(Buf, Result.mSet);
  //     
  Buf.ClrAndDestroy;
  Buf.Free;
end;

//########################################################################
//                                                                       #
//                                                              #
//                   -4,  . 86                             #
//                                                                       #
//########################################################################

//  
// aQuick -  :
// false = ,  
// true  = ,  

function TGraph.GenPaints(aQuick: boolean): TSet;
var  Undep : TSet;       //   
       Buf : TBuffer;  //  
         S : TSet;       //   
        CS : TCostSet;   //   
     Cover : TCostSet;   //  
begin
  Undep:= GenUndepend;   //     
  //       
  //     
  Buf:= TBuffer.Create;      //   
  S:= Undep.GetFirst as TSet;  //   
  while Assigned(S) do begin   //   
    //    ( Cost=1)
    //    :
    CS:= TCostSet.Create(1, S.Copy as TSet, true);
    Buf.Put(CS);
    S:= Undep.GetNext as TSet;
  end;
  //     
  Undep.ClrAndDestroy;
  Undep.Free;
  //     (unit Assembly):
  if aQuick
    then Cover:= CollectGradCover(mNodes, Buf)  // 
    else Cover:= CollectMinCover(mNodes, Buf);  // 
  RemoveItemsFromBuf(Buf, Cover.mSet);  //    
  Buf.ClrAndDestroy;                    //   
  Buf.Free;                             //  
  //  :
  Result:= Cover.mSet.Copy as TSet;  //  -
  Cover.Free;                        //  
  Result.CoverToDissect;             //    
end;

//########################################################################
//                                                                       #
//                          P-                              #
//                      -5,  . 111                         #
//                                                                       #
//########################################################################

//      
//    p-

type  TGamma = class(TItem)
        mRoot  : TNode;  //  
        mGamma : TSet;   // ,    mRoot
        constructor Create(aRoot: TNode; aGamma: TSet);
        destructor Destroy;  override;
        function  Compare(arg: TItem): TCompare; override;
        procedure Print(var aFile: TextFile); override;
      end;

constructor TGamma.Create(aRoot: TNode; aGamma: TSet);
begin
  inherited Create;
  mRoot:= aRoot;       //  
  mGamma:= aGamma;     //   
end;

destructor TGamma.Destroy;
begin
  mGamma.Free;
  inherited;
end;

function TGamma.Compare(arg: TItem): TCompare;
begin
  Result:= mRoot.Compare((arg as TGamma).mRoot)
end;

procedure TGamma.Print(var aFile: TextFile);
var S: string;
begin
  S:= mRoot.GetName + ' -> ' + NodesToStr(mGamma);
  Writeln(aFile, S);
end;

// TSortedNum -    
//   

type  TSortedNum = class(TItem)
        mNumber : integer;
        constructor Create(aNumber : integer);
        function  Compare(arg: TItem): TCompare; override;
        procedure Print(var aFile: TextFile); override;
      end;

constructor TSortedNum.Create(aNumber: integer);
begin
  inherited Create;
  mNumber:= aNumber;
end;

function TSortedNum.Compare(arg: TItem): TCompare;
begin
  Result:= inherited Compare(arg);
  if Result= cmpEq then Exit;
  if mNumber < (arg as TSortedNum).mNumber
    then Result:= cmpLess
  else if mNumber > (arg as TSortedNum).mNumber
    then Result:= cmpGreate  else Result:= cmpEq  ;
end;

procedure TSortedNum.Print(var aFile: TextFile);
begin
  Writeln(aFile,'Num= ',mNumber);
end;


// TAreal
//    
//    p-
// aAreal -   ( mSet)
// aGamma - ,     ( mGamma)

type  TAreal = class(TCostSet)
        //  mSet -  ,   TCostSet
        mGamma : TSet;  // ,      mSet
        constructor Create(aAreal, aGamma: TSet);
        destructor Destroy;  override;
        procedure Print(var aFile: TextFile); override;
      end;

constructor TAreal.Create(aAreal {},
                          aGamma {} : TSet);
begin
  inherited Create(1, aAreal, true);
  mGamma:= aGamma;  // ,      mSet
end;

destructor TAreal.Destroy;
begin
  mGamma.Free;
  inherited;
end;

procedure TAreal.Print(var aFile: TextFile);
var S: string;
begin
  S:= NodesToStr(mGamma) + ' ==> ' + NodesToStr(mSet);
  Writeln(aFile, S);
end;


//    .
//        aLambda
// aDirect = InCenter | OutCenter | InOutCenter

function TGraph.GenCenter(var aLambda: integer; aDirect: TCenter): TSet;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //       aNode.
  //       mFarLinks

  function GetMaxDist(aNode: TNode): integer;
  var FL: TFarLink;
      Value: integer;
  begin
    Result:= 0;
    with aNode.mFarLinks do begin
      FL:= GetFirst as TFarLink;
      while Assigned(FL) do begin
        if mLoadNodes                     //    ...
          then Value:= FL.mNodeFar.mValue //    . 
          else Value:= 1;                 //    
        if Result < FL.mDist * Value
          then Result:= FL.mDist * Value;
        FL:= GetNext as TFarLink;
      end;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var Dist, MinDist: integer;  //    . 
    Node: TNode;

begin  { TGraph.GenCenter }

  //     ,
  //         .
  DoneMap;           //   ,   

  InitMap(aDirect);  //     

  //      
  MinDist:= MaxInt;
  Node:= NodeFirst;
  while Assigned(Node) do begin
    Dist:= GetMaxDist(Node);
    if MinDist > Dist then MinDist:= Dist;
    Node:= NodeNext;
  end;
  Result:= CreateSet;

  //  -
  Node:= NodeFirst;
  while Assigned(Node) do begin
    Dist:= GetMaxDist(Node);    //    
    if Dist = MinDist then Result.Insert(Node);
    Node:= NodeNext;
  end;
  aLambda:= MinDist;  //    
end;


//     
// aDirect = InCenter | OutCenter | InOutCenter

procedure TGraph.InitMap(aDirect: TCenter);

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  procedure InitFarLinks(aNode: TNode; aDirect : TCenter);
  var     Node : TNode;     //  
          Link : TLink;     //  
            FL : TFarLink;  //  
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //        
    procedure LocalIn;
    begin
      with Node do begin
        Link:= OutLinkFirst;            //   
        while Assigned(Link) do begin   //   
          if Link.mDest = aNode then begin  //    
            //     
            FL.mNodeNear:= Node;
            FL.mDist:= Link.mValue;
            FL.mStep:= 1;   //      
            Break;
          end;
          Link:= OutLinkNext;     //   
        end // while
      end // with
    end;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //     
    procedure LocalOut;
    begin
      with Node do begin
        Link:= InLinkFirst;             //   
        while Assigned(Link) do begin   //   
          if Link.mOwner = aNode then begin //    
            //     
            FL.mNodeNear:= Node;
            FL.mDist:= Link.mValue;
            FL.mStep:= 1;   //      
            Break;
          end;
          Link:= InLinkNext;     //   
        end; // while
      end // with
    end;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  begin   { InitFarLinks }
    PosPush;  //    
    Node:= NodeFirst;                //    
    while Assigned(Node) do begin    //    
      if Node = aNode then begin
        FL:= TFarLink.Create(Node, Node, 0);   //    
      end else with Node do begin
        //     
        FL:= TFarLink.Create(nil {near}, Node {far}, CInfinity);
        if (aDirect=InCenter) or not mDirect
          then LocalIn   //      
          else LocalOut; //   
      end; // else
      aNode.mFarLinks.Insert(FL); //   
      Node:= NodeNext;  //    
    end;
    PosPop;  //    
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     (  )

  function  Handle(aNode: TNode; aStep: integer): boolean;
  var   FL1 : TFarLink;  //      
        FL2 : TFarLink;  //      

    //      FL2

    function Test_FL2(aNear : TNode): boolean;
    var  Dist : integer;  //  
         FL : TFarLink;   //     aNode
    begin
      Result:= false;
      Dist:= FL1.mDist + FL2.mDist;  //    
      FL:= aNode.GetFarLink(FL2.mNodeFar); //    
      if FL.mDist > Dist then begin    //    
        FL.mNodeNear:= FL1.mNodeNear;  //   
        FL.mDist:= Dist;               //  
        FL.mStep:= aStep+1;  //       
        Result:= true;       //    
      end;
    end;

  begin  { Handle }
    Result:= false;
    FL1:= aNode.mFarLinks.GetFirst as TFarLink;  //    aNode
    while Assigned(FL1) do begin
      //      
      if FL1.mStep=aStep then with FL1.mNodeFar.mFarLinks do begin
        PositionPush;
        FL2:= GetFirst as TFarLink;   //    
        while Assigned(FL2) do begin
          with FL2 do  //      
          if Assigned(mNodeNear)            //  
             and (mDist<>0)                 //     
             and (mNodeFar<>aNode)          //     
            then if Test_FL2(FL2.mNodeNear) //   
                   then Result:= true;      //  ,  
          FL2:= GetNext as TFarLink;        //   
        end; // while
        PositionPop;
      end;
      FL1:= aNode.mFarLinks.GetNext as TFarLink;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  procedure Create(aDirect : TCenter);
  var  Node : TNode;    //  
       Step : integer;  //   ( )
       Flag : boolean;  //   
  begin
    //   :
    Node:= NodeFirst;
    while Assigned(Node) do begin         //  
      with Node do if Assigned(mFarLinks) //   
        then mFarLinks.ClrAndDestroy      // 
        else mFarLinks:= CreateSet;       //   
      InitFarLinks(Node, aDirect);        //   
      Node:= NodeNext;
    end;
    //  
    for Step:= 1 to mNodes.GetCount-2  do begin
      Flag:= false;
      Node:= NodeFirst;
      while Assigned(Node) do begin     //  
        if Handle(Node, Step)           //    ,
          then Flag:= true;             //     
        Node:= NodeNext;
      end;
      //     ,   
      if not Flag then Break;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var  Node : TNode;      //  
      Buf : TBuffer;  //      
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  procedure SaveInBuf;
  begin
    Node:= NodeFirst;
    //    :
    while Assigned(Node) do begin
      Buf.Put(Node.mFarLinks);
      Node.mFarLinks:= nil;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      
  procedure AddFromBuf;
  var     S : TSet;       //   
        FL1 : TFarLink;   //    S
        FL2 : TFarLink;   //    Node
  begin
    Node:= NodeFirst;
    //  :
    while Assigned(Node) do begin
      S:= Buf.Get as TSet;
      FL1:= S.GetFirst as TFarLink;
      while Assigned(FL1) do begin
        FL2:= Node.GetFarLink(FL1.mNodeFar);
        if Assigned(FL2) then FL2.mDist:= FL2.mDist + FL1.mDist;
        FL1:= S.GetNext as TFarLink;
      end;
      S.ClrAndDestroy;
      S.Free;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

begin  { TGraph.InitMap }

  if mInitMap then Exit;  //    
  mInitMap:= true;        // ,   

  mMapDirect:= aDirect;   //   

  case aDirect of
    InCenter,   //    
    OutCenter:  //    
      Create(aDirect);
    InOutCenter: begin
      // -   
      //     :
      Create(InCenter);
      //         :
      Buf:= TBuffer.Create;
      SaveInBuf;
      //     :
      Create(OutCenter);
      //  ,   ,
      //   :
      AddFromBuf;
      Buf.ClrAndDestroy;
      Buf.Free;
    end;
  end;

  //       
  // (   )

  Node:= NodeFirst;
  while Assigned(Node) do begin
    Node.SortFarLinks;
    Node:= NodeNext;
  end;

end;

//    ,
//       

procedure TGraph.DoneMap;
var  Node : TNode;    //  
begin
  //  ,  :
  if not mInitMap then Exit;
  mInitMap:= false;       //   
  mMapDirect:= Undefined; //   

  //     :
  Node:= NodeFirst;
  while Assigned(Node) do begin
    with Node do begin
      if Assigned(mFarLinks) then begin
        mFarLinks.ClrAndDestroy;   //  
        mFarLinks.Free;            //  
        mFarLinks:= nil;
      end;
    end;
    Node:= NodeNext;
  end;
end;

//   

procedure TGraph.ExpoFarLinks;
var Node: TNode;
begin
  PosPush;
  Node:= NodeFirst;
  while Assigned(Node) do begin
    Node.ExpoFarLinks;
    Node:= NodeNext;
  end;
  PosPop;
end;

//     

procedure TGraph.ExpoLinksData;
var Node: TNode;
begin
  PosPush;
  Node:= NodeFirst;
  while Assigned(Node) do begin
    Node.ExpoData;
    Node:= NodeNext;
  end;
  PosPop;
end;

//   ,    aNode
//      aLambda
// (  )
//      .

function TGraph.GenLimitGamma(aNode: TNode; aLambda: integer): TSet;
var FL : TFarLink;    //   
begin
  Result:= CreateSet;
  with aNode.mFarLinks do begin  //    
    FL:= GetFirst as TFarLink;
    while Assigned(FL) do begin
      if aLambda >= (FL.mDist * aNode.mValue) //    
        then Result.Insert(FL.mNodeFar);      //    
      FL:= GetNext as TFarLink;
    end;
  end;
end;

//     TGamma

function TGraph.GenLimitGams(aLambda: integer): TSet;
var  Node : TNode;
     GammaSet: TSet;
     GammaItem: TGamma;
begin
  Result:= CreateSet;
  Node:= NodeFirst;
  while Assigned(Node) do begin
    //    
    GammaSet:= GenLimitGamma(Node, aLambda);
    //   +
    GammaItem:= TGamma.Create(Node, GammaSet);
    Result.Insert(GammaItem);
    Node:= NodeNext;
  end;
end;

//  ,   
// aGams -    ,
//     

function TGraph.GenIntersections(aGams: TSet): TBuffer;
var
    GammaItem : TGamma;   //    
        Nodes : TSet;     //   (TAreal.mSet)
     GammaSet : TSet;     //   (TAreal.mGamma)
         Temp : TSet;     //    
      Existed : TSet;     //     
        Areal : TAreal;   //   ""
         Flag : boolean;  //   
          Cnt : integer;  //     . .
            i : integer;  // 
begin
  Result:= TBuffer.Create;    //  
  Temp:= CreateSet;           //    
  Existed:= CreateSet;        //    
  Nodes:= CreateSet;          //   (TAreal.mSet)
  GammaSet:= CreateSet;       //   (TAreal.mGamma)
  GammaSet.CopyItems(mNodes); //    
  //      - TAreal , :
  // mNodes -    
  // mGamma -  ,   ,   
  Result.Put(TAreal.Create(Nodes, GammaSet));
  Flag:= false;
  repeat
    Cnt:= Result.GetCount;  //     
    for i:=1 to Cnt do begin
      //     
      Flag:= false;                 //    
      Areal:= Result.Get as TAreal; //    
      //     
      GammaItem:= aGams.GetFirst as TGamma;
      while Assigned(GammaItem) do begin
        if not Areal.mSet.Exist(GammaItem.mRoot) then begin
          //        
          //     
          Temp.CopyItems(Areal.mGamma);     //    
          Temp.Mul(GammaItem.mGamma);       //    
          if Temp.GetCount <> 0 then begin
            //    ,
            //        
            Nodes:= CreateSet;               //  (TAreal.mSet)
            Nodes.CopyItems(Areal.mSet);     //   
            Nodes.Insert(GammaItem.mRoot);   //    
            if Existed.Exist(Nodes) then begin
              //        
              Nodes.Free;    //   
              //Nodes:= nil; //    
            end else begin
              //      
              Existed.Insert(Nodes);     //    
              //       
              GammaSet:= CreateSet;      //   (TAreal.mGamma)
              GammaSet.CopyItems(Temp);  //   GammaSet= Temp
              Result.Put(TAreal.Create(Nodes, GammaSet));
            end; // if
            Flag:= true;  //    
          end; // if
        end; // if
        GammaItem:= aGams.GetNext as TGamma;  //   
      end; // while
      //      :
      //          ,
      //        
      //   ,     
      if Flag then begin
        Existed.Delete(Areal.mSet);//    
        Areal.Free;                //    
      end else begin
        Result.Put(Areal);         //      
      end;
    end; // for
    // ,    
  until not Flag and (Cnt=Result.GetCount);
  Temp.Free;
  Existed.Free;
end;


//   
//     aLambda ()
// aDirect = InCenter | OutCenter | InOutCenter

function TGraph.GenPCentLambda(aLambda: integer; aDirect: TCenter): TCostSet;
var   Gams : TSet;
       Buf : TBuffer;
begin
  //     (  ):
  case aDirect of
    InCenter:    InitMap(OutCenter);
    OutCenter:   InitMap(InCenter);
    InOutCenter: InitMap(InOutCenter);
  end;
  //   
  //     () aLambda
  Gams:= GenLimitGams(aLambda);
  //    
  Buf:= GenIntersections(Gams);
  //      ()
  Result:= CollectMinCover(mNodes, Buf);
  //    
  RemoveItemsFromBuf(Buf, Result.mSet);
  //      
  Gams.ClrAndDestroy;
  Gams.Free;
  Buf.ClrAndDestroy;
  Buf.Free;
end;


//    P-
//     aLambda
// ( ).
// aPoly -   
// aDirect = InCenter | OutCenter | InOutCenter

function TGraph.GenPCenters(aPoly: integer;
                            var aLambda: integer;
                            aCenter: TCenter): TCostSet;

var Lambdes : TSet;       // - (  )
       LRec : TSortedNum; //  -
     Lambda : integer;    //   

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     ()
  procedure LambdasCreate;

    procedure AddLambdas(aNode: TNode);
    var FL : TFarLink;    //   
        SN : TSortedNum;  //    
        Lambda: integer;  //     
    begin
      with aNode.mFarLinks do begin
        FL:= GetFirst as TFarLink;
        while Assigned(FL) do begin
          Lambda:= FL.mDist * aNode.mValue;
          SN:= TSortedNum.Create(Lambda);
          if not Lambdes.Insert(SN) then SN.Free;
          FL:= GetNext as TFarLink;
        end;
      end;
    end;

  var  Node : TNode;    //  

  begin  { LambdasCreate }
    Lambdes:= CreateSet;
    Node:= NodeFirst;
    while Assigned(Node) do begin   //  
      AddLambdas(Node);             //  
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

begin { TGraph.GenPCenters }
  //    :
  case aCenter of
    InCenter   : InitMap(OutCenter);
    OutCenter  : InitMap(InCenter);
    InOutCenter: InitMap(InOutCenter);
  end;
  //  - (  )
  LambdasCreate;
  //   ()   
  LRec:= Lambdes.GetFirst as TSortedNum;
  repeat
    Lambda:= LRec.mNumber;            //   
    Result:= GenPCentLambda(Lambda, aCenter);  //  
    //    ,  :
    if Result.mSet.GetCount <= aPoly then Break;
    //          
    Result.ClrAndDestroy;
    Result.Free;
    Result:= nil;
    //      ( )
    LRec:= Lambdes.GetNext as TSortedNum;
  until not Assigned(LRec);
  //   :
  aLambda:= Lambda;
  //     
  Lambdes.ClrAndDestroy;
  Lambdes.Free;
  //    :
  DoneMap;
end;

//########################################################################
//                                                                       #
//                        P-                                      #
//                  ,  . 135                               #
//                                                                       #
//########################################################################

//   "" ( )

procedure TGraph.MedianCopy(aRes: TSet);
var N: TNode;
begin
  aRes.Clear;
  N:= NodeFirst;
  while Assigned(N) do begin
    if N.mColor = CBlack then aRes.Insert(N);
    N:= NodeNext;
  end;
end;

//   p-
// (     )
//   :
// CWhite --    
// CBlack --     
// CGray  --         

function TGraph.MedianCost: integer;
var Node: TNode;       //  
    Dist: integer;     //    
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      
  function FindMinDist: integer;
  var FL: TFarLink;     //   
  begin
    Result:= CInfinity;
    with Node.mFarLinks do begin
      FL:= GetFirst as TFarLink;
      while Assigned(FL) do begin
        if FL.mNodeFar <> Node then begin
          if FL.mNodeFar.mColor = CBlack then begin
            Result:= FL.mDist;
            Break;
          end;
        end;
        FL:= GetNext as TFarLink;
      end;
    end; // with
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
begin  { MedianCost }
  //    ( ),
  // (        )
  Result:= 0;
  Node:= NodeFirst;
  while Assigned(Node) do begin
    if Node.mColor <> CBlack then begin
      //       
      Dist:= FindMinDist;  //    Node
      Inc(Result, Node.mValue * Dist);   //  
    end;
    Node:= NodeNext;
  end;
end;

//   " " 
//     ( )
// aDirect = InCenter | OutCenter | InOutCenter

function TGraph.GenPMedian_A (
                     aPoly: integer;     //  (P)
                     aDirect: TCenter;   // , ., -.
                     var aCost: integer  //  ()
                    ): TSet;
var
    BestCost: integer;  //  
    Level : integer;    //     
    NN: integer;        //  ,    
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  //   
  // (   )

  procedure Searching(aIndex: integer {   } );
  var i: integer;
      N: TNode;        //  
      Cost: integer;   //   p-
  begin
    //   
    for i:= aIndex to NN do begin
      N:= mNodes.GetItem(i + Level) as TNode;  //  
      N.mColor:= CBlack;
      //  p- ?
      if Level = aPoly-1 then begin
        //     
        Cost:= MedianCost;
        if Cost < BestCost then begin
          //    ,   
          BestCost:= Cost;
          MedianCopy(Result);  //   
        end;
      end else begin
        //   p-  ,
        //    
        Inc(Level);       // Level+1
        Searching(i);     //    
        Dec(Level);       // Level-1
      end;
      N.mColor:= CGray;   //     
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

begin { TGraph.GenPMedian_A }
  //      mColor, mPred:
  ResetNodes;
  //    :
  case aDirect of
    InCenter   : InitMap(OutCenter);
    OutCenter  : InitMap(InCenter);
    InOutCenter: InitMap(InOutCenter);
  end;
  //   :
  BestCost:= MaxInt;
  Result:= CreateSet;
  Level:=0;                     //     
  NN:= mNodes.GetCount+1-aPoly; //  , .   
  Searching(1);        //     
  DoneMap;             //    
  aCost:= BestCost;    //  
end;


//     p- ( )
// aDirect = InCenter | OutCenter | InOutCenter

function TGraph.GenPMedian(aPoly: integer;    // 
                          aDirect: TCenter;   // , ., -.
                          var aCost: integer  //  ()
                          ): TSet;
var
    BestCost: integer;  //  
    Level : integer;    //     
    NN: integer;        //  ,    
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      
  //   :
  // CWhite --    
  // CBlack --     
  // CGray  --        

  function EvalBest(aBest: integer): boolean;
  var Node: TNode;       //  
        FL: TFarLink;    //   
       Sum: integer;     //  
      Dist: integer;     //      
         S: TSet;        //    FL
      WhiteCnt: integer; //   

    //     
    // Node --    
    function FindMinDist: integer;
    var FL: TFarLink;     //   
    begin
      Result:= CInfinity;
      with Node.mFarLinks do begin
        FL:= GetFirst as TFarLink;
        while Assigned(FL) do begin
          if (FL.mNodeFar <> Node) and
             (FL.mNodeFar.mColor <> CGray) then begin
            //      
            Result:= FL.mDist;
            Break;
          end;
          FL:= GetNext as TFarLink;
        end;
      end; // with
    end;

  begin  { Eval }
    //    ( ),
    //       ,
    //       ()
    // (        )

    Sum:= 0;  //  
    //      
    WhiteCnt:= mNodes.GetCount - aPoly;
    S:= CreateSet;
    Node:= NodeFirst;
    while Assigned(Node) do begin
      if Node.mColor <> CBlack then begin
        //      
        Dist:= FindMinDist;  //    Node
        case Node.mColor of
          CWhite: begin
                    //     
                    //    
                    FL:= TFarLink.Create(nil, Node, Node.mValue*Dist);
                    S.Insert(FL);
                  end;
          CGray:  begin
                    //      
                    Inc(Sum, Node.mValue*Dist); //  
                    Dec(WhiteCnt);     //    -1
                    if Sum>=aBest then Break;
                  end;
        end;
      end;
      Node:= NodeNext;
    end;
    //     
    //  WhiteCnt      
    FL:= S.GetFirst as TFarLink;
    while (WhiteCnt > 0) and (Sum < aBest) do begin
      Inc(Sum, FL.mDist);           //  
      FL:= S.GetNext as TFarLink;   //   
      Dec(WhiteCnt)
    end;
    //     :
    S.ClrAndDestroy;
    S.Free;
    Result:= Sum < aBest;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //       
  // (   )

  procedure Searching(aIndex: integer {   } );
  var i: integer;
      N: TNode;        //  
      Cost: integer;   //   p-
  begin
    //   
    for i:= aIndex to NN do begin
      N:= mNodes.GetItem(i + Level) as TNode;  //  
      N.mColor:= CBlack;
      //  p- ?
      if Level = aPoly-1 then begin
        //     
        Cost:= MedianCost;
        if Cost < BestCost then begin
          //    ,   
          BestCost:= Cost;
          MedianCopy(Result);  //   
        end;
      end else begin
        //   p-  .
        //       .
        //    ,  ""   
        if  EvalBest(BestCost) then begin
          //    
          Inc(Level);    // Level+1
          Searching(i);  //    
          Dec(Level);    // Level-1
        end;
      end;
      N.mColor:= CGray;  //     
    end;
    //        
    //     
    for i:= aIndex to NN do begin
      N:= mNodes.GetItem(i + Level) as TNode;  //  
      N.mColor:= CWhite;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

begin  { TGraph.GenPMedian }

  //      mColor, mPred:
  ResetNodes;
  //    :
  case aDirect of
    InCenter   : InitMap(OutCenter);
    OutCenter  : InitMap(InCenter);
    InOutCenter: InitMap(InOutCenter);
  end;
  //   :
  BestCost:= MaxInt;
  Result:= CreateSet;
  Level:=0;            //     
  NN:= mNodes.GetCount+1-aPoly; // - , .   
  Searching(1);        //     
  DoneMap;             //    
  aCost:= BestCost;    //  
end;


//########################################################################
//                                                                       #
//                                                     #
//                                                                       #
//########################################################################

type //      ()

  TTreeLink = class (TLink)
    mMaxTree: boolean;  // false -- ., true -- . 
    constructor Create(aLink: TLink; aMaxTree: boolean);
    function  Compare(arg: TItem): TCompare; override;
    procedure Print(var aFile: TextFile); override;
  end;

//     

constructor TTreeLink.Create(aLink: TLink; aMaxTree: boolean);
begin
  inherited Create(aLink.mOwner,  //  
                   aLink.mDest,   //  
                   aLink.mValue,  //  
                   true           //  
                   );
  mMaxTree:= aMaxTree;
end;

function TTreeLink.Compare(arg: TItem): TCompare;
begin
  Result:= inherited Compare(arg);
  if Result = cmpEq then Exit;
  //        
  //    
  if not mOwner.mOwner.mDirect and  //   
    (mOwner=(arg as TTreeLink).mDest) and (mDest=(arg as TTreeLink).mOwner)
    then begin
      //    ,  :
      Result:= cmpEq;
      Exit;
    end;
  //    ,      ()
  Result:= cmpLess;
  if mValue > (arg as TTreeLink).mValue then Result:= cmpGreate;
  //       
  if mMaxTree
    then if Result=cmpLess then Result:= cmpGreate else Result:= cmpLess;
end;

//      ,
//    -  

procedure TTreeLink.Print(var aFile: TextFile);
var N1, N2 : TNode;
begin
  N1:=mOwner;  N2:=mDest;
  if not mOwner.mOwner.mDirect and
     (mOwner.Compare(mDest) = cmpGreate) then begin
    N1:=mDest;   N2:=mOwner;
  end;
  Write(aFile, ' ' + N1.GetName + N2.GetName);
  if mLoadLinks then Write(aFile, '= ', mValue);
end;

//  ,     
// aLinks -  () 
//     (aMaxTree=false)
//    (aMaxTree=true)
//     

procedure TNode.AddTreeLinks(aMaxTree: boolean; aLinks: TSet);
var  L : TLink;
    TL : TTreeLink;
begin
  //   
  L:= OutLinkFirst;
  while Assigned(L) do begin
    //   ,    
    if (L.mDest.mColor = CWhite) then begin
      TL:= TTreeLink.Create(L, aMaxTree);
      if not aLinks.Insert(TL) then TL.Free;
    end;
    L:= OutLinkNext;
  end;
end;

//   (, . 162)
//   ()  
//    TTreeLink  

function TGraph.GenCoverTree(aMaxTree: boolean;  // false= min, true = max
                             var aCost: integer  // . 
                            ): TSet;             //  TTreeLink

var Node: TNode;      //  
      TL: TTreeLink;  //    
   Links: TSet;       //    

begin
  aCost:=0;              //  
  Result:= CreateSet;    //    
  Links:= CreateSet;     //    
  ResetNodes;            //     mColor = CWhite
  Node:= NodeFirst;      //  
  //     (   )
  while Assigned(Node) and
       (Result.GetCount < mNodes.GetCount-1) do  begin
    Node.mColor:= CBlack;     //     
    //    Links     :
    Node.AddTreeLinks(aMaxTree, Links);
    //   Links     
    TL:= Links.GetFirst as TTreeLink;   //   = 
    Node:= nil;                         //    
    while Assigned(TL) do begin
      Links.Delete(TL);                 //     Links
      //      ,
      //    :
      if TL.mOwner.mColor = CWhite
        then Node:= TL.mOwner   //  
      else if TL.mDest.mColor = CWhite
        then Node:= TL.mDest;   //  
      if Assigned(Node) then begin
        //     :
        Result.Insert(TL);      //    
        Inc(aCost, TL.mValue);  //  
        Break;
      end;
      //      :
      TL.Free;                         //   
      TL:= Links.GetNext as TTreeLink; //   
    end;
  end;
  //    ,    
  if Result.GetCount < mNodes.GetCount-1 then begin
    Result.ClrAndDestroy;
    aCost:=0;
  end;
  //      :
  Links.ClrAndDestroy;
  Links.Free;
end;


//    
//        
//    ,     MaxInt

function TGraph.GenCoverDir(var aRoot: TNode;   // 
                            var aCost: integer  //  () 
                           ): TSet;             //  
var Tree: TSet;     //  
    Links: TSet;    //   
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    ,    aRoot
  //       ().
  //    

  function RootCover(aRoot: TNode): integer;
  var Node: TNode;       //  
      TL : TTreeLink;    //  
  begin
    Result:= 0;          //  
    ResetNodes;          //    mColor = CWhite
    Node:= aRoot;        //   -- 
    //      (   )
    while Assigned(Node) and
         (Tree.GetCount < mNodes.GetCount-1) do  begin
      Node.mColor:= CBlack;     //    
      //      :
      Node.AddTreeLinks(false, Links);
      //   Links     
      TL:= Links.GetFirst as TTreeLink;   //   = 
      Node:= nil;                         //    
      while Assigned(TL) do begin         //   
        Links.Delete(TL);                 //     Links
        //       :
        if TL.mDest.mColor = CWhite then begin
          //     :
          Node:= TL.mDest;        //     
          Tree.Insert(TL);        //    
          Inc(Result, TL.mValue); //  
          Break;
        end;
        //      :
        TL.Free;                         //   
        TL:= Links.GetNext as TTreeLink; //   
      end;
    end;
    Links.ClrAndDestroy; //     
    //    ,   
    if Tree.GetCount < mNodes.GetCount-1 then begin
      Tree.ClrAndDestroy;  //  
      Result:= MaxInt;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var Root: TNode;    //   
    Cost: integer;  //  () 

begin { TGraph.GenCoverDir }
  Result:=nil;  aRoot:= nil;
  aCost:= MaxInt;
  Links:= CreateSet;  //   
  Tree:= CreateSet;   //   
  //   :
  Root:= NodeFirst;
  while Assigned(Root) do begin
    Cost:= RootCover(Root);
    //   :
    if Tree.GetCount > 0 then begin
      //    
      if Cost<aCost then begin
        //   
        if Assigned(Result)
          then Result.ClrAndDestroy //   
          else Result:=CreateSet;   //   
        Result.CopyItems(Tree);     //   ()
        Tree.Clear;                 //   
        aRoot:= Root;               //  
        aCost:= Cost;               //   ()
      end else begin
        Tree.ClrAndDestroy;  //  ,    
      end;
    end;
    Root:= NodeNext;
  end;
  //   :
  Tree.Free;
  Links.Free;
end;

//     (, . 149)
//   

procedure TGraph.ExpoAllCovers;
var
  Links: TSet;       //    
   Tree: TSet;       //   
  Level: integer;    //  
     NN: integer;    //  ,   
    Cnt: integer;    //   
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  procedure Local_Init;
  var Node: TNode;   //  
  begin
    ResetNodes;          //    mColor=0, mPred=nil
    Tree:= CreateSet;    //    
    Links:= CreateSet;   //    
    Node:= NodeFirst;
    while Assigned(Node) do begin
      with Node do begin
        mRoot:= Node;         //  
        AddTreeLinks(false, Links);  //   
      end;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  //     

  function Cost(aTree: TSet): integer;
  var TL: TTreeLink;  //    
  begin
    Result:= 0;
    TL:= aTree.GetFirst as TTreeLink;
    while Assigned(TL) do begin
      Inc(Result, TL.mValue);
      TL:= aTree.GetNext as TTreeLink;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   
  // aTL.mLink.mOwner --   T1
  // aTL.mLink.mDest  --   T2
  function Union(aTL: TTreeLink): boolean;
  var   N2: TNode;   //   T2
        p2: TNode;   //  
        tmp: TNode;  //  
  begin
    with aTL do begin
      //      ,  
      Result:= mOwner.mRoot <> mDest.mRoot;
      if not Result then Exit;
      //      mPred
      N2:= mDest;        //     T2
      p2:= N2.mPred;     //    T2
      N2.mPred:=mOwner;  //     T1
      //   :
      while Assigned(p2) do begin
        tmp:= p2.mPred;  //     T2
        p2.mPred:= N2;   //   
        N2:= p2;         //      2- 
        p2:= tmp;        //    T2
      end;
      //  2-   1-:
      // root1 = mOwner.mRoot.Expo
      // root2 = mDest.mRoot.Expo
      p2:= mDest.mRoot as TNode;    {###}
      N2:= NodeFirst;
      while Assigned(N2) do begin
        if N2.mRoot = p2 then N2.mRoot:= mOwner.mRoot;
        N2:= NodeNext;
      end;
    end;  // with
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  procedure Remove(aTL: TTreeLink);
  var Root: TNode;  //    2- 
      S1 : TSet;    //   
      S2 : TSet;    //   
      Node: TNode;
  begin
    S1:= CreateSet;    S2:= CreateSet;
    //     2- 
    with aTL do if mOwner.mPred = mDest
      then Root:= mOwner    //  
      else Root:= mDest;    //  
    Root.mRoot:= Root;      //    
    Root.mPred:= nil;       //     
    S1.Insert(Root);        //      
    repeat
      //   :
      Node:= NodeFirst;
      while Assigned(Node) do begin
        //     ,    
        if S1.Exist(Node.mPred) then begin
          Node.mRoot:= Root;
          S2.Insert(Node);
        end;
        Node:= NodeNext;
      end;
      // ,    
      if S2.GetCount=0 then Break;
      //        
      S1.CopyItems(S2);
      S2.Clear;
    until false;
    S1.Free; S2.Free;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    
  // ()  L  N-1 
  // L -      = Links.GetCount
  // N -    = mNodes.GetCount

  procedure Searching(aIndex: integer {   } );
  var  i: integer;
      TL: TTreeLink;  //    
  begin
    //   
    for i:= aIndex to NN do begin
      TL:= Links.GetItem(i + Level) as TTreeLink;  //  
      if Union(TL) then begin
        //     ,   
        Tree.Insert(TL);
        if Level = mNodes.GetCount-2 then begin
          //  
          Inc(Cnt);      //   
          Write(Cnt:4, '':2, Cost(Tree):3, '':2);
          Tree.Expo;
        end else begin
          //     ,   ,
          //    
          Inc(Level);    // Level+1
          Searching(i);  //    
          Dec(Level);    // Level-1
        end;
        //    
        Tree.Delete(TL);  //    
        Remove(TL);       //   
      end; // if
    end; // for
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
begin { TGraph.ExpoAllCovers }
  Local_Init;
  Cnt:= 0;     //   
  Level:=0;    //  
  //  ,    :
  // (mNodes.GetCount - 1) =   
  NN:= Links.GetCount + 1 - (mNodes.GetCount - 1);
  Searching(1);
  Links.ClrAndDestroy;
  Links.Free;
end;

//########################################################################
//                                                                       #
//                                                     #
//                                                                       #
//########################################################################

//   
//      (   mLow=0)
//  .      , 4.2 (. 91)

function TGraph.CalcMaxFlow0(aSource, aDest: TNode): integer;
var Delta: integer;   //   
    Que: TBuffer;   //  
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   
  //  ,       (mLow>0)

  function CalcStartFlow: integer;
  var Link: TLink;    //  
  begin
    Result:= 0;
    Link:= aSource.OutLinkFirst;
    while Assigned(Link) do begin
      Inc(Result, Link.mFlow);
      Link:= aSource.OutLinkNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   
  //     

  function CalcDelta: integer;
  var Node: TNode;
      Link: TLink;    //    
  begin
    Result:=0;
    //  mColor=0, mPred=nil, mDist= MaxInt, mFlow= MaxInt
    ResetNodes;
    Que.Clear;   //   
    //     
    aSource.mColor:= CGray;
    Que.Put(aSource);
    while Que.GetCount>0 do begin
      Node:= Que.Get as TNode;
      //   ,   
      if Node=aDest then Break;
      //     :
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        //    ,    
        if (Link.mDest.mColor=CWhite)            //    
           and (Link.mFlow < Link.mHigh) then begin  //    
          Link.mDirect:= +1;  //   
          //        :
          with Link.mDest do begin
            mColor:= CGray;    //    
            mPred:= Node;      //  
            mLink:= Link;      //    
            mFlow:= Minimum(Node.mFlow, Link.mHigh - Link.mFlow); // 
          end;
          Que.Put(Link.mDest); //    
        end;
        Link:= Node.OutLinkNext;
      end;
      //     :
      Link:= Node.InLinkFirst;
      while Assigned(Link) do begin
        //    ,     
        if (Link.mOwner.mColor=CWhite)          //    
           and (Link.mFlow > Link.mLow) then begin //    
          Link.mDirect:= -1;  //   
          //   ( )     :
          with Link.mOwner do begin
            mColor:= CGray;    //    
            mPred:= Node;      //  
            mLink:= Link;      //    
            mFlow:= Minimum(Node.mFlow, Link.mFlow); //  
          end;
          Que.Put(Link.mOwner); //    
        end;
        Link:= Node.InLinkNext;
      end;  // while
    end;  // while
    //    ,     
    //     
    if aDest.mColor = CWhite then Exit;
    //     
    Result:= aDest.mFlow; //   
    //         
    Node:= aDest;  // 
    while Node <> aSource do begin
      with Node.mLink do Inc(mFlow, mDirect * Result); // mDirect = +1 / -1
      Node:= Node.mPred;  //  
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

begin  { TGraph.CalcMaxFlow0 }
  Result:= 0;
  if not Assigned(aSource) or not Assigned(aDest) then Exit;
  Result:= CalcStartFlow; //  
  Que:= TBuffer.Create;   //   
  //    ,   
  repeat
    Delta:= CalcDelta;    //   
    Inc(Result, Delta);   //  
  until Delta=0;
  Que.Free;
end;

//   
// (     )
//  .  .    , .327 (7.8)

function TGraph.CalcMinFlow(aSource, aDest: TNode): integer;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    
  //      
  // arg = false --    
  // arg = true  --   

  procedure RestoreFlowData(arg: boolean);
  var Node: TNode;
      Link: TLink;      //  
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      with Node do begin
        //    :
        Link:= OutLinkFirst;
        while Assigned(Link) do begin
          with Link do begin
            mLow:= mTemp;           // 
            mHigh:= mHigh + mLow;   // 
            //   :
            if arg
              then mFlow:= mFlow + mLow  //   
              else mFlow:= 0;            //    
          end; // with
          Link:= OutLinkNext;
        end; // while
      end; // with
      Node:= NodeNext;
    end; // while
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var Node: TNode;      //  
    Link: TLink;      //  
    Reverse: TLink;   //   T->S   
    Flow: integer;    //  
    SumLow: integer;  //   
    Sx: TNode;        //  
    Tx: TNode;        //  
    OldLink: TLink;   //  T->S   
    OldLow, OldHigh: integer; //    

begin    { TGraph.CalcMinFlow }

  //    ,  
  if not Assigned(aSource) or not Assigned(aDest) then begin
    Result:=-1;
    Exit;
  end;

  // :
  Sx:= TNode.Create(0, Self); //  
  Tx:= TNode.Create(0, Self); //  
  SumLow:=0;    //      
  //         :
  Node:= NodeFirst;
  while Assigned(Node) do begin
    Node.mFlow:=0; //   
    with Node do begin
      //    :
      Link:= OutLinkFirst;
      while Assigned(Link) do begin
        with Link do begin
          Inc(Node.mFlow, mLow);  //   
          mHigh:= mHigh - mLow;   //  
          mTemp:= mLow;           //   
          mLow:= 0;               //   
          mFlow:=0;               //  
        end;
        Link:= OutLinkNext;
      end;
    end; // with
    Node:= NodeNext;
  end; // while

  //         :
  Node:= NodeFirst;
  while Assigned(Node) do begin
    with Node do begin
      //    :
      Link:= InLinkFirst;
      while Assigned(Link) do begin
        Dec(Node.mFlow, Link.mTemp);   //    
        Link:= InLinkNext;
      end;
    end;
    //      :
    if Node.mFlow <> 0 then begin
      //    :
      if Node.mFlow > 0 then begin
        //       
        Link:= Node.MakeLink(Tx, 0);  //  
        Link.mHigh:= Node.mFlow;      //    
      end else begin
        //       
        Link:= Sx.MakeLink(Node, 0);  //  
        Link.mHigh:= -Node.mFlow;      //    
      end
    end;
    Inc(SumLow, Abs(Node.mFlow)); //     
    Node:= NodeNext;
  end; // while
  SumLow:= SumLow div 2;  //     

  //     = 0,    
  if SumLow=0 then begin
    // :
    Sx.Free;     //  
    Tx.Free;     //  
    //    :
    RestoreFlowData(false);
    Result:=0;
    Exit;
  end;

  //     :
  mNodes.Insert(Sx);   // 
  mNodes.Insert(Tx);   // 

  //      
  // (     ):

  //     :
  OldLow:= 0; OldHigh:=0;

  //    ?
  OldLink:= aDest.GetLink(aSource);
  if Assigned(OldLink) then begin
    // ,        
    Reverse:= OldLink;    //  
    with OldLink do begin
      // ,   
      OldLow := mLow;    //  
      OldHigh:= mHigh;   //  
      //   
      mLow := 0;         //   
      mHigh:= MaxInt;    //   
    end // with
  end else begin
    //    ,    
    Reverse:= aDest.MakeLink(aSource, 0); // 
    Reverse.mHigh:= MaxInt;   //    
  end;

  //     
  //        

  Flow:= CalcMaxFlow0(Sx, Tx);

  //        :
  Result:= Reverse.mFlow;  // 
  Reverse.mFlow:= 0;       //  

  //   ,   ,   

  //   ?
  if Assigned(OldLink) then with OldLink do begin
    //  
    mLow := OldLow;
    mHigh:= OldHigh;
  end else begin
    //      ,
    //       
    aDest.RemoveLink(aSource);
  end;

  //     :

  RemoveNode(Sx);  Sx.Free;     //  
  RemoveNode(Tx);  Tx.Free;     //  

  //     (Sum)
  //     (SumLow)
  //       

  if Flow = SumLow then begin
    //   
    RestoreFlowData(true); //     
  end else begin
    //    
    Result:=-1;             //    
    RestoreFlowData(false); //     
  end;

end;


//   
//      

function TGraph.CalcMaxFlow(aSource, aDest: TNode): integer;
begin
  //    :
  Result:= CalcMinFlow(aSource, aDest);
  //    ,     :
  if Result < 0 then Exit;
  //      :
  Result:= CalcMaxFlow0(aSource, aDest);
end;


//     
// , . 104
// , . 339

function TGraph.CalcMinCostFlow(aSource, aDest: TNode; //   
                                aFlow: integer         //  
                                ): integer;            //  

var  Gray: TSet;      //   () 
     Que: TBuffer;  //     CalcDeltaFlow
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      ,
  //     

  procedure ClearFlowData;
  var Node: TNode;
      Link: TLink;      //  
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mFlow:= MaxInt; //     
      Node.mDist:= 0;      //   
      //    :
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        Link.mFlow:= 0;          //   = 0
        Link.mColor:= CWhite;    //   
        Link:= Node.OutLinkNext;
      end; // while
      Node:= NodeNext;
    end; // while
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   (mColor)   -
  //       

  procedure ClearColors;
  var Node: TNode;      //  
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mColor:= CWhite;
      Node.mFlow:= MaxInt; //     
      Node:= NodeNext;
    end;
    //  -      :
    aSource.mColor:= CGray;
    Gray.Clear;
    Gray.Insert(aSource);
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  //  :
  // -     Gray
  // -    
  // -    
  //  true       

  function Buying: boolean;
  var Node: TNode;      //  
      Link: TLink;      //  
      Delta: integer;   //   
      MinDelta: integer;//   
  begin
    MinDelta:= MaxInt;
    //   :
    Node:= Gray.GetFirst as TNode;
    while Assigned(Node) do begin
      //   :
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        with Link do if mDest.mColor = CWhite then begin
          //    :
          if mFlow < mHigh then begin
            //    :
            // Link.mValue -   
            // mOwner.mDist -      Node
            // mDest.mDist -     mDest
            // Delta - ,       Link
            Delta:= mValue - (mDest.mDist - mOwner.mDist);
            if Delta < MinDelta
              then MinDelta:= Delta;  //   
          end; // if
        end;
        Link:= Node.OutLinkNext;
      end; // while
      Node:= Gray.GetNext as TNode; //    
    end; // while Assigned(Node)
    Result:= MinDelta < MaxInt;     // true,   
    if not Result or (MinDelta=0) then Exit;
    //  mDist -    :
    Node:= NodeFirst;
    while Assigned(Node) do begin
      if Node.mColor = CWhite then Inc(Node.mDist, MinDelta);
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //  ,     
  //  :
  // -    
  // -  
  // -   
  //    .
  //     ( )
  //    

  function CalcDeltaFlow(var aCost: integer): integer;
  var
      Node: TNode;     //  
      Link: TLink;     //    
  begin
    Result:=0; aCost:=0;
    Que.Clear;  //   
    //     
    Node:= Gray.GetFirst as TNode;
    while Assigned(Node) do begin
      Que.Put(Node);
      Node:= Gray.GetNext as TNode;
    end;
    Que.Put(aSource);
    //    :
    while Que.GetCount>0 do begin
      Node:= Que.Get as TNode;
      //     :
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        with Link do if mDest.mColor=CWhite then begin
          //    
          //    ,
          //     
          // Link.mValue --  
          // Link.mDest.mDist - mOwner.mDist --   
          if (mValue = mDest.mDist - mOwner.mDist) //  
             and (mFlow < mHigh) then begin  //    
            mDirect:= +1;  //   
            //       :
            mDest.mColor:= CGray;    //    
            mDest.mPred:= Node;      //  
            mDest.mLink:= Link;      //    
            mDest.mFlow:= Minimum(Node.mFlow, mHigh - mFlow); // 
            Que.Put(mDest);      //    
            Gray.Insert(mDest);  //   
          end;
        end; // if
        Link:= Node.OutLinkNext;
      end;
      //     :
      Link:= Node.InLinkFirst;
      while Assigned(Link) do begin
        with Link do if mOwner.mColor=CWhite then begin
          //    ,
          //        
          if mFlow > 0 then begin  //     
            mDirect:= -1;  //   
            //      :
            with Link.mOwner do begin
              mColor:= CGray;    //    
              mPred:= Node;      //  
              mLink:= Link;      //    
              mFlow:= Minimum(Node.mFlow, Link.mFlow); //  
            end;
            Que.Put(Link.mOwner);     //    
            Gray.Insert(Link.mOwner); //   
          end;
        end;
        Link:= Node.InLinkNext;
      end;  // while Assigned(Link)
      //   ,    :
      if aDest.mColor <>  CWhite then Break;
    end;  // while Que.GetCount>0
    //   ,    
    //        
    if aDest.mColor <> CWhite then begin
      Result:= aDest.mFlow;
      aCost:= aDest.mDist;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      
  // aDelta -  

  procedure RecalcFlow(aDelta: integer);
  var Node: TNode;     //  
  begin
    //           
    Node:= aDest;  // 
    // ,    :
    while Node <> aSource do begin
      // Node.mLink -    
      // Node.mLink.mDirect = +1 / -1
      with Node.mLink do Inc(mFlow, mDirect * aDelta); // 
      Node:= Node.mPred;  //  
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var
  Flow: integer;        //  
  DeltaFlow: integer;   //  
  Cost1: integer;       //   
  OK: boolean;          //   

begin { TGraph.CalcMinCostFlow }

  Result:= 0;       //   
  Flow:= 0;         //  
  Gray:= CreateSet; //   () 
  Que:= TBuffer.Create;  //   
  ClearFlowData;    //     (mFlow=MaxInt)
  ClearColors;      //       

  //      :
  while Flow < aFlow do begin
    //  ""   ()
    //    .
    //      ,   :
    repeat
      //        
      DeltaFlow:= CalcDeltaFlow(Cost1);
      if DeltaFlow > 0  then begin
        OK:= true;
        //  ,   :
        if DeltaFlow > aFlow - Flow then DeltaFlow:= aFlow - Flow;
        //          :
        RecalcFlow(DeltaFlow);        //    
        Inc(Flow, DeltaFlow);         //  
        Inc(Result, Cost1*DeltaFlow); //   
        ClearColors;                  //     
      end else begin
        //    ,      
        // ( Buying   )
        OK:= Buying;
        //    ,     
        if not OK then Break;
      end;
      //     ,    
    until Flow = aFlow;
    //    ,     
    if not OK then begin
      Result:=-1;  //  = -1
      Break;
    end;
  end; // while Flow < aFlow
  //  :
  Gray.Free; //  
  Que.Free;  //  
end;


//########################################################################
//                                                                       #
//                                                          #
//                                                                       #
//########################################################################

{   :

  TPairs = (pMaxN,  //  
            pMaxW   //   ()
            pMinW   //   ()
           );
}

type  // TPair --    

     TPair = class (TCostSet)
       function Compare(arg: TItem): TCompare;  override;
       procedure Print(var aFile: TextFile); override;
     end;

function TPair.Compare(arg: TItem): TCompare;
begin
  if Self = arg then begin Result:= cmpEq; Exit; end;
  if mCost = (arg as TCostSet).mCost then begin
    //     
    Result:= mSet.Compare((arg as TCostSet).mSet);
  end else begin
    //    ,    ( -  )
    if mCost > (arg as TCostSet).mCost
      then Result:= cmpGreate
      else Result:= cmpLess
  end;
end;

//    (    )

procedure TPair.Print(var aFile: TextFile);
var N: TNode;
    k: integer;
begin
  k:= mSet.GetCount;
  N:= mSet.GetFirst as TNode;
  while Assigned(N) do begin
    Write(aFile, N.GetName);
    Dec(k);
    if k > 0 then Write(aFile, '-');
    N:= mSet.GetNext as TNode;
  end;
  Write(aFile, ' = ', mCost);
end;

type  // TPairsBlock -- ,  

      TPairsBlock = class (TSetList)
        mLabel : TNode;    //   ( )
        mCost  : integer;  //      
        function Compare(arg: TItem): TCompare; override;
        procedure Print(var aFile: Text); override;
        destructor Destroy;  override;
      end;

destructor TPairsBlock.Destroy;
begin
  ClrAndDestroy;  //    
  Inherited;
end;

//       

function TPairsBlock.Compare(arg: TItem): TCompare;
begin
  Result:= cmpLess;
  if GetCount > (arg as TPairsBlock).GetCount  then Result:= cmpGreate
end;

procedure TPairsBlock.Print(var aFile: Text);
begin
  Writeln(aFile,mLabel.GetName,' Cost=', mCost:3,'  Count=',GetCount:2);
  inherited Print(aFile);
  Writeln(aFile,'- - - - -');
end;

//########################################################################
//    (/)  :
// pMaxN --   (   )
// pMaxW --  
// pMinW --  
//########################################################################

function TGraph.GenPairs_Dissect(aMode: TPairs): TCostSet;

var Blocks: TSet;        //   ()
    BestCost: integer;   //   () 
    BestCnt: integer;    //   ()  /
    Pairs: TCostSet;     //  
    Nodes: TSet;         //    
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    
  procedure InitNodes;
  var Node: TNode;
      LD, LR : TLink; //    
  begin
    //    
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mPower:= 0;  //     
      Node:= NodeNext;
    end;
    //         
    // (    )
    Node:= NodeFirst;
    while Assigned(Node) do begin
      with Node do begin
        //   
        mColor:= CWhite;
        //    LD
        LD:= OutLinkFirst;
        while Assigned(LD) do begin
          //  LR --  
          LR:= LD.mDest.GetLink(Node);
          if Assigned(LR) then begin
            //   
            if (aMode=pMaxW) and (LR.mValue < LD.mValue) then begin
              //     
              LD.mColor:= CBlack; //  
              LR.mColor:= CWhite; //   
            end else begin
              //     
              LR.mColor:= CBlack; //  
              LD.mColor:= CWhite; //   
            end;
          end else begin
            //    
            LD.mColor:= CBlack; //    
          end;
          //       
          Inc(LD.mOwner.mPower);
          Inc(LD.mDest.mPower);
          LD:= OutLinkNext;
        end;
      end;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //  ,     

  function GetBestNode: TNode;
  var Node: TNode;
      Max : integer;
  begin
    Max:= 0; Result:= nil;
    Node:= NodeFirst;
    while Assigned(Node) do begin
      with Node do if (mColor=CWhite)    //    
                      and (mPower > Max) //     
      then begin
        Result:= Node;  //   
        Max:= mPower;   //     
      end;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      ()
  //     

  function GenBlocks: TSet;

    // C      :
    function MakePair(aLink: TLink): TPair;
    begin
      Result:= TPair.CreateEmpty;
      with aLink do begin
        //        :
        Result.mSet.Insert(mOwner);
        Result.mSet.Insert(mDest);
        //    :
        Dec(mOwner.mPower);
        Dec(mDest.mPower);
      end;
      //  ()       /
      case aMode of
        pMaxN: Result.mCost:=  1;              // 
        pMinW: Result.mCost:=  aLink.mValue;   //  
        pMaxW: Result.mCost:= -aLink.mValue;   //   ***
      end;
    end;

  var Node: TNode;         //  
      Link: TLink;         //   
      Pair: TPair;         //  
      Min: integer;        //     
      Block: TPairsBlock;  //  

  begin  { GenBlocks }
    InitNodes;           //   mPower
    Result:= CreateSet;  //  
    //  :
    Node:= GetBestNode;  //   ""    
    while Assigned(Node) do begin
      Min:= MaxInt;    //        
      Block:=nil;      //     

      //   :
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        //    
        if Link.mColor = CBlack then begin
          Link.mColor:= CWhite;  //  
          Pair:= MakePair(Link); //  
          //      
          if Min > Pair.mCost then Min:= Pair.mCost;
          //     ,  :
          if not Assigned(Block) then Block:= TPairsBlock.Create;
          Block.Insert(Pair);  //    
        end;
        Link:= Node.OutLinkNext;  //  
      end; // while

      //   :
      Link:= Node.InLinkFirst;
      while Assigned(Link) do begin
        //    
        if Link.mColor = CBlack then begin
          Link.mColor:= CWhite;  //  
          Pair:= MakePair(Link); //  
          //      
          if Min > Pair.mCost then Min:= Pair.mCost;
          //     ,  :
          if not Assigned(Block) then Block:= TPairsBlock.Create;
          Block.Insert(Pair);  //    
        end;
        Link:= Node.InLinkNext;  //  
      end; // while

      //   ,    :
      if Assigned(Block) then begin
        Block.mLabel:= Node;   //  
        Block.mCost:= Min;     //     
        Result.Insert(Block);  //  
      end;
      Node.mColor:= CBlack; // ,   
      Node:= GetBestNode;   //   ""    
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    

  procedure BlockHandle(aBlock: integer;  //  
                        aNodes: TSet;     //  
                        aRes: TCostSet    //  
                        );
  var Block: TPairsBlock; //  
      Nodes: TSet;        //   
      Res  : TCostSet;    //    ()
      Pair : TPair;       //    

      //  

      function Prediction(aBlock: TPairsBlock;  //  
                          aNodes: TSet;         //  
                          aRes: TCostSet        //  
                          ): boolean;
      var B: TPairsBlock;         //  
          Cost, Count : integer;  //     /
      begin
        Result:= not Assigned(aRes);
        if Result then Exit;
        //       
        Cost:= aRes.mCost;  Count:= 0;
        if Assigned(aRes.mSet) then Count:= aRes.mSet.GetCount;
        //     
        B:= Blocks.GetFirst as TPairsBlock;
        while B <> aBlock do B:= Blocks.GetNext as TPairsBlock;
        //      
        B:= Blocks.GetNext as TPairsBlock;
        //   
        while Assigned(B) do begin
          if not aNodes.Exist(B.mLabel) then begin
            Inc(Count);         //   
            Inc(Cost, B.mCost); //   
          end;
          B:= Blocks.GetNext as TPairsBlock;
        end;
        //   ,
        //     :
        // -     , 
        // -       
        Result:= (Count > BestCnt) or
                 (Count = BestCnt) and (Cost < BestCost)
      end;
      var OK: boolean;
  begin  { BlockHandle }
    //   :
    Block:= Blocks.GetItem(aBlock) as TPairsBlock;
    //         ,
    //     
    if not aNodes.Exist(Block.mLabel) then begin
      // ,   ,     
      //    :
      Pair:= Block.GetFirst as TPair;
      while Assigned(Pair) do begin
        //     ()
        //  ,   
        //     .  
        if  not aNodes.TestIntersect(Pair.mSet) then begin
          //    
          Nodes:= aNodes.Copy as TSet;     //   
          Res:= aRes.Copy as TCostSet;     //   
          //      
          Nodes.Add(Pair.mSet);   //   
          Res.Insert(Pair);       //   
          if aBlock < Blocks.GetCount then begin
            //    .
            //      
            //     ,
            //     
            if Prediction(Block,Nodes,Res)
              then BlockHandle(aBlock+1,Nodes,Res);
          end else begin
            //    (aBlock = Blocks.GetCount),
            //  ,     :
            // -   Result   
            // -   /   
            // -  / ,    
            OK:= not Assigned(Result) or
               (Res.mSet.GetCount > BestCnt) or
               (Res.mSet.GetCount = BestCnt) and (BestCost > Res.mCost);
            if OK then begin
              BestCost:= Res.mCost;         //  () 
              BestCnt:= Res.mSet.GetCount;  //    /
              Result.Free;                  //   
              Result:= Res.Copy as TCostSet;//   
            end;
          end;
          //        :
          Res.Free;   //  
          Nodes.Free; //  
        end; // if
        Pair:= Block.GetNext as TPair;  //    
      end; // while
    end; // if
    //       ,
    //   ,  
    if (aMode = pMaxN) and (BestCnt = Self.Nodes div 2) then Exit;
    //      ,   
    if aBlock < Blocks.GetCount then begin
      //    .
      //      
      //     ,     
      if Prediction(Block,aNodes,aRes) then BlockHandle(aBlock+1,aNodes,aRes);
    end else begin
      //    (aBlock = Blocks.GetCount),
      //  ,     :
      // -   Result   
      // -   /   
      // -  / ,    
      if not Assigned(Result) or
         (aRes.mSet.GetCount > BestCnt) or
         (aRes.mSet.GetCount = BestCnt) and (BestCost > aRes.mCost)
        then begin
          BestCost:= aRes.mCost;         //  () 
          BestCnt:= aRes.mSet.GetCount;  //    /
          Result.Free;                   //   
          Result:= aRes.Copy as TCostSet;//   
        end;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  // (     )

  procedure RestoreCost;
  var Pair : TPair;
      Buf: TBuffer;
  begin
    Buf:= TBuffer.Create;
    with Result.mSet do begin
      //     Result.mSet
      Pair:= GetFirst as TPair;
      while Assigned(Pair) do begin
        Pair.mCost:= -Pair.mCost;
        Buf.Push(Pair);
        Pair:= GetNext as TPair;
      end;
      //   Result    
      //       
      Result.Clear;
      while Buf.GetCount > 0 do begin
        Result.Insert(Buf.Pop as TPair);
      end;
    end;
    Buf.Free;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //        

  procedure RemoveFromBlocks;
  var  Block: TPairsBlock;
  begin
    with Blocks do begin
      Block:= GetFirst as TPairsBlock;
      while Assigned(Block) do begin
        Block.Sub(Result.mSet);
        Block:= GetNext as TPairsBlock;
      end;
    end;
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

begin { TGraph.GenPairs_Dissect }

  Result:= nil;        //  --  

  Blocks:= GenBlocks;  //  
  if Blocks.GetCount = 0 then begin Blocks.Free; Exit; end;

  //  :
  BestCost:= MaxInt;            //   
  BestCnt:= 0;                  //   ()  /
  Nodes:= CreateSet;            //   
  Pairs:= TCostSet.CreateEmpty; //   

  //   :
  BlockHandle(1, Nodes, Pairs);  //  Result

  //  :
  Pairs.Free;        //   ()
  Nodes.Free;        //    
  //     Result.mSet
  if Assigned(Result) then  RemoveFromBlocks;
  Blocks.ClrAndDestroy;  //   
  Blocks.Free;           //   

  //     
  //   :
  if aMode = pMaxW then RestoreCost;
end;

//########################################################################
//                                                                       #
//                              #
//                                              #
//                                                                       #
//########################################################################

type  //    

  TBlossom = class(TNodeInt)
    mBase: TNode;     //   
    mNewBase: TNode;  //  ()  
    mNext: TNode;     //    
                      //    
              //      
    function IsPresent(aNode: TNode): boolean;
              //    
              //    mNext
    function SetNewBase(aNewBase: TNode): TNode;
              //     
    function  GetBase: TNode;
              //     
    function  GetNewBase: TNode;
              //     
              //    aIn  aOut
    function GetProfit(aIn, aOut : TNode): integer;
    function GetName: string; override;
    procedure Print(var aFile: TextFile); override;
  end;

{ TBlossom }

//    aNode   
//    

function TBlossom.IsPresent(aNode: TNode): boolean;
var N: TNode;
begin
  N:= aNode;
  repeat
    N:= N.mBlossom;     //  mBlossom
                        //    
    Result:= N = Self;
    if Result then Break;
  until N = N.mBlossom; // ,    
end;

//     aNewBase.
//     

function TBlossom.SetNewBase(aNewBase: TNode): TNode;
var Bloss: TBlossom;
begin
  //     
  Bloss:= aNewBase.mBlossom as TBlossom;
  if aNewBase <> mNewBase
    then Bloss.mNewBase:= aNewBase;  //   
  Result:= aNewBase.GetPair;  //    
  // ,     
  if Bloss.mBlossom = Bloss then Exit;
  //      ,
  //     
  Result:= (Bloss.mBlossom as TBlossom).SetNewBase(Bloss);
end;

//     
//     

function  TBlossom.GetBase: TNode;
begin
  //      
  if mBase is TBlossom
    then Result:= (mBase as TBlossom).GetBase
    else Result:= mBase;  //    
end;

//     
//     

function  TBlossom.GetNewBase: TNode;
begin
  //      
  if mNewBase is TBlossom
    then Result:= (mNewBase as TBlossom).GetNewBase
    else Result:= mNewBase;  //    
end;

//    ()  
//   :
// aIn  --   (  )
// aOut --   (   )
//  =    ,   

function TBlossom.GetProfit(aIn, aOut : TNode): integer;

  //      aNode
  //   aBloss    

  function CalcProfit(aBloss: TBlossom; aNode: TNode): integer;
  var Bloss: TBlossom;
  begin
    //    
    Result:= aNode.mProfit;
    // ,    
    if aBloss = Self then Exit;
    //    Bloss --   
    Bloss:= aBloss.mBlossom as TBlossom;
    //     ,
    //    
    Result:= Result + CalcProfit(Bloss, aBloss);
  end;

var Bloss: TBlossom;      //   
    Profit_In: integer;   //   
    Profit_Out: integer;  //   
begin
  //     
  Bloss:= aIn.mBlossom as TBlossom;
  //     
  Profit_In:= CalcProfit(Bloss, aIn);
  //     
  Bloss:= aOut.mBlossom as TBlossom;
  //     
  Profit_Out:= CalcProfit(Bloss, aOut);
  //    
  Result:= Profit_Out - Profit_In;
end;

//     

function TBlossom.GetName: string;

  function Short(const arg: string): string;
  begin
    Result:= System.Copy(arg,1,3)
  end;

var N: TNode;
    Flag: boolean;
begin
  Result:= inherited GetName+': ';
  Result:= Result+' {';
  N:= mBase;  Flag:= false;
  repeat
    if Flag
      then Result:= Result+Short(N.GetName)+'='
      else Result:= Result+Short(N.GetName)+'-';
    N:= N.mRight;  Flag:= not Flag;
  until N = mBase;

  if (mBase <> mNewBase) and Assigned(mNewBase) and (mNewBase <> Self)
    then Result:= Result+' <'+Short(mNewBase.GetName)+'>';

  if Assigned(mPair) then with mPair do
    Result:= Result+' '+mOwner.GetName+'+'+mDest.GetName;

  if mValue <> 0 then if mOut
    then Result:= Result+' ->> '+IntToStr(mValue)
    else Result:= Result+' --> '+IntToStr(mValue);

  Result:= Result+' : '+IntToStr(mProfit);
  Result:= Result+'}';
end;

procedure TBlossom.Print(var aFile: TextFile);
begin
  Writeln(aFile, GetName);
end;

//########################################################################
//  

//     
//   
// (true -- , false -- )

function  GetOut(aNode: TNode): boolean;
begin
  Result:= aNode.GetBlossom.mOut;
end;

//     ,
//      

function  GetRoot(aNode: TNode): TNode;
begin
  Result:= aNode.GetBlossom.mRoot as TNode;
end;

//########################################################################
//                                                                       #
//                              #
//                                                     #
//                                                                       #
//########################################################################
//  aMode: TPairs  --  :
//    pMaxN  --     
//    pMaxW  --   
//    pMinW  --   

function TGraph.GenPairs_Edmonds(aMode: TPairs): TCostSet;

var //   

  gMainQue : TBuffer;     //   
  gQue : TBuffer;         //     
  gBlossomStack : TBuffer;//    
  gBlosNum: integer;      //   
  gAppendix: TSet;        //     
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //  

  //  
  procedure BlossExpand(aBloss: TBlossom); forward;

  //  
  procedure RemoveBlossoms; forward;

  //   aNode <-> aNode.mPair
  procedure Disconnect(aNode, aBloss : TNode); forward;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    
  //       ( mValue)

  procedure InitGraphValues;
  var N: TNode;   //  
      L: TLink;   //  
      MaxVal: integer;  //   
  begin
    MaxVal:=0;
    //        
    N:= NodeFirst;
    while Assigned(N) do begin
      //    
      L:= N.OutLinkFirst;
      while Assigned(L) do with L do begin
        if mValue > MaxVal then MaxVal:= mValue;
        mTemp:= mValue;     //    
        L:= N.OutLinkNext;  //  
      end;
      N:= NodeNext; //  
    end;
    MaxVal:= (MaxVal+1) div 2;
    //     
    N:= NodeFirst;
    while Assigned(N) do begin
      N.mLimit:= N.mValue; //    
      N.mValue:= MaxVal;
      N:= NodeNext;
    end;

    if aMode = pMinW then begin
      //     
      //   
      N:= NodeFirst;
      while Assigned(N) do begin
        //    
        L:= N.OutLinkFirst;
        while Assigned(L) do with L do begin
          //   
          //    
          mValue:= 2*MaxVal - mValue;
          L:= N.OutLinkNext;
        end;
        N:= NodeNext;
      end;
    end;
  end;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //       

  procedure RestoreGraphValues;
  var N: TNode;   //  
      L: TLink;   //  
  begin
    N:= NodeFirst;
    while Assigned(N) do begin
      N.mValue:= N.mLimit;  //   mValue
      L:= N.OutLinkFirst;
      while Assigned(L) do with L do begin
        mValue:= mTemp;   //   mValue
        L:= N.OutLinkNext;
      end;
      N:= NodeNext;
    end;
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    aQue    aNode

  procedure PutInQue(aQue: TBuffer; aNode : TNode);
  var N: TNode;
  begin
    if aNode is TBlossom then with aNode as TBlossom do begin
      //   ,  PutInQue 
      N:= mBase;
      repeat
        PutInQue(aQue, N);
        N:= N.mRight;
      until N = mBase;
    end else begin
      //   ,   
      aQue.Put(aNode);
    end;
  end;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //        aLink

  function BlossomMake(aLink: TLink): TBlossom;

  var Base : TNode; //   

    //      aNode
    //    

    function CalcProfit(aNode: TNode): integer;
    var Direct: boolean;  //   :
                          // true -    
        Sign: boolean;    //   :
                          // true - , false - 
        N: TNode;         //     
        N1, N2: TNode;    //   
        Link: TLink;      //  
        Delta: integer;   //   
    begin
      Result:= 0;
      if aNode = Base then Exit;  //    = 0
      //  aNode   ...
      if aNode is TBlossom then with aNode as TBlossom do begin
        //     
        //   aNode.mNewBase  aNode.mBase
        N1:= GetNewBase;  //    
        N2:= GetBase;     //    
        Result:= GetProfit(N1,N2);
      end;
      //     
      //   aNode     Base
      // (true -    )
      Direct:= aNode.mPair = aNode.mRLnk;
      N1:= aNode;   //      
      Sign:= true;  // true -  , false - 
      repeat
        if Direct then begin
          N2:= N1.mRight;     //  
          Link:= N1.mRLnk;    //   N1 <-> N2
        end else begin
          N2:= N1.mLeft;      //  
          Link:= N2.mRLnk;    //   N1 <-> N2
        end;
        Delta:= Link.mValue;  //    ()
        //    ,    
        if Sign then Delta:= -Delta;
        //  
        Inc(Result, Delta);
        //    
        //    N1 <--> N2
        //     ,
        //       
        if not Sign then begin
          //  Link --   
          if N1 is TBlossom then with N1 as TBlossom do begin
            //  N1   
            //  N --     N1
            if IsPresent(Link.mOwner)
              then N:= Link.mOwner else N:= Link.mDest;
            //   (with N1 as TBlossom)
            Delta:= GetProfit(GetNewBase, N);
            //  
            Inc(Result, Delta);
          end;

          if N2 is TBlossom then with N2 as TBlossom do begin
            //  N2   
            //  N --     N2
            if IsPresent(Link.mOwner)
              then N:= Link.mOwner else N:= Link.mDest;
            //   (with N2 as TBlossom)
            //    ,    N2
            // ,      
            if N2 = Base
              then Delta:= GetProfit(GetBase, N)
              else Delta:= GetProfit(GetNewBase, N);
            //  
            Inc(Result, Delta);
          end;
        end; // if not Sign
        //    
        Sign:= not Sign;
        N1:= N2;
      until N1 = Base;
    end;
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  var Owner, Dest : TNode; //    
      N1, N2  : TNode;     //   
      N  : TNode;

  begin  // function BlossomMake

    //  gBlosNum     
    Inc(gBlosNum);
    //      
    Result:= TBlossom.Create(gBlosNum, 0, nil);
    //     
    with aLink do begin
      Owner:= mOwner.GetBlossom;
      Dest:= mDest.GetBlossom;
    end;

    //   mLink    

    Owner.mLink:= aLink;
    Dest.mLink:= aLink;
    Dest.mPred:= Owner;

    //     
    //     
    //     .
    //      mColor

    N:= Owner; //    ( )
    repeat
      N.mColor:= gBlosNum;  //    
      if GetOut(N) then begin
        //    ,
        //    
        N:= N.GetPair;
        if Assigned(N) then N:= N.GetBlossom;
      end else begin
        //    ,
        //    
        N:= N.mPred;
        if Assigned(N) then N:= N.GetBlossom;
      end;
    until not Assigned(N);

    //     .
    //       
    N:= Dest; //    ( )
    while N.mColor <> gBlosNum do begin
      // ,     
      N:= N.GetPair.GetBlossom; //  
      N:= N.mPred.GetBlossom;   //   
    end;

    //  N    
    //     
    Base:= N;

    //    

    with Result do begin
      mOut:= true;          //   
      mBase:= Base;         //  
      mNewBase:= Base;      //  () 
      mPair:= Base.mPair;   //   
      mOwnDest:= Base.mOwnDest; //    (Owner|Dest)
      mDist:= Base.mDist;   //  
    end;

    //   mRight, mLeft  mRLnk
    //     

    Owner.mRight:= Dest;  //  
    Dest.mLeft:= Owner;   //  
    Owner.mRLnk:= aLink;  //  

    //    
    //   N.mRight    N.mLeft
    //        aLink
    //    .

    //  GetOut     ,
    //        .

    //       
    N2:= Dest;  //  
    while N2 <> Base do begin
      if GetOut(N2) then begin
        // N1 -- , N2 - 
        N1:= N2.GetPair.GetBlossom;
        N2.mRLnk:= N2.mPair;  //   
      end else begin
        // N1 -- , N2 -- 
        N1:= N2.mPred.GetBlossom;
        N2.mRLnk:= N2.mLink;  //     
      end;
      N2.mRight:= N1;   N1.mLeft:= N2;
      N2:= N1;
    end;

    //       
    N2:= Owner;  //  
    while N2 <> Base do begin
      if GetOut(N2) then begin
        // N1 -- , N2 - 
        N1:= N2.GetPair.GetBlossom;
        N1.mRLnk:= N2.mPair;  //   
      end else begin
        // N1 -- , N2 -- 
        N1:= N2.mPred.GetBlossom;
        N1.mRLnk:= N2.mLink;  //     
      end;
      N2.mLeft:= N1;   N1.mRight:= N2;
      N2:= N1;
    end;

    //      gQue
    //   .
    //     
    N:= Base;
    repeat
      if not GetOut(N) then begin
        //  N --   ,
        //      
        PutInQue(gQue, N);
      end;
      //    
      //      
      N.mBlossom:= Result;
      N:= N.mRight;
    until  N= Base;

    //     
    //    mProfit
    // --     
    N:= Base;
    repeat
      N.mProfit:= CalcProfit(N);
      N:= N.mRight;
    until N= Base;

  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     aLink.
  //   ()  Delta = 0.
  //        (aMode = pMaxN),
  // ..    ,   .

  function Delta(aLink : TLink): integer;
  begin
    if aMode = pMaxN
      then Result:=0  //     
      else with aLink do
        //      
        //    
        Result:= mOwner.mValue + mDest.mValue - mValue;
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //       aRoot.
  //  :
  // --   ,    . .
  // --   ,    . .
  // --  nil,    .

  function Alter(aRoot: TNode): TNode;

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //    2 ()
    //      ,
    //   

    procedure Mul_2;
    var Node: TNode;
        L : TLink;
    begin
      PosPush;  //   
      Node:= NodeFirst;
      while Assigned(Node) do begin
        with Node do begin
          mValue:= 2*mValue;
          mDist:= 2*mDist;
          mProfit:= 2*mProfit;
          //   
          L:= OutLinkFirst;
          while Assigned(L) do begin
            L.mValue:= 2*L.mValue;
            L:= OutLinkNext;
          end;
        end;
        Node:= NodeNext;
      end;
      PosPop;  //   

      //  ,    gBlossomStack
      Node:= gBlossomStack.GetFirst as TNode;
      while Assigned(Node) do begin
        with Node as TBlossom do begin
          mValue:= 2*mValue;    //   
          mDist:= 2*mDist;      //  
          mProfit:= 2*mProfit;  //   
        end;
        Node:= gBlossomStack.GetNext as TNode;
      end;
    end;

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //     
    //    ()  

    function GetMinBlossValue(aRoot : TNode): integer;
    var Bloss: TBlossom;
    begin
      Result:= MaxInt;
      //     
      Bloss:= gBlossomStack.GetFirst as TBlossom;
      while Assigned(Bloss) do begin
        with Bloss do begin
          if (GetRoot(Bloss) = aRoot) //     
             and not GetOut(Bloss)    //     
             then begin
            //   ,   
            if Result > mValue then Result:= mValue
          end;
        end;
        Bloss:= gBlossomStack.GetNext as TBlossom;
      end;
    end;
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //   ,
    //     
    // aSuccess --     

    function CalcForExpand(var aSuccess: boolean): integer;
    var Node: TNode;      //  
        Blossom : TNode;  //   
        L : TLink;        //  
        d : integer;      //  
        Res: extended;    //   ()

    begin
      aSuccess:= false; //    
      gAppendix.Clear;  //      

      //     . 
      Result:= GetMinBlossValue(aRoot);
      //  Result <> MaxInt,     
      if Result <> MaxInt then begin
        Result:= Result div 2;
        aSuccess:= true;  //    
        if Result = 0
          then Exit;  //      . 
      end;
      //    .   ,
      //      
      Res:= Result; // Res : extended;

      //     
      Node:= NodeFirst;
      while Assigned(Node) do with Node do begin
        //      
        if (GetRoot(Node)<>aRoot) //      
           or not GetOut(Node)    //    
           then begin             //   
            Node:= NodeNext as TNode;
            Continue;
        end;
        //  ,     
        Blossom:= Node.GetBlossom;
        //      Node (with Node)
        L:= OutLinkFirst;
        while Assigned(L) do with L do begin
          //     ?
          if GetRoot(mDest) = aRoot then begin
            //      
            if (mDest.GetBlossom = Blossom)
              //        ...
              or not GetOut(mDest) //      
              then begin
              //        
              L:= OutLinkNext;
              Continue;
            end;
            //        
            //   ,
            //   L    
            //    (with L)
            d:= mOwner.mValue + mDest.mValue - mValue;
            //    d    Res
            if d/2 < Res {} then begin
              //     
              // Res   !
              Res:= d/2;  //   
              gAppendix.Clear;  //    
              aSuccess:= true;  //    
            end;
            //    
            //    ...
            if Round(2*Res) {} = d {}
              //     
               then gAppendix.Insert(Node);
          end else begin
            //       .
            //  ,  L    ,
            //    
            //    (with L)
            d:= mOwner.mValue + mDest.mValue - mValue;
            //    d    Res
            if d < Res {} then begin
              Res:= d;          //   
              gAppendix.Clear;  //    
              aSuccess:= true;  //    
            end;
            //    
            //    ...
            if Round(Res) = d
              //       
              then gAppendix.Insert(Node);
          end;  // if GetRoot(mDest) = aRoot
          L:= OutLinkNext;  //     Node
        end; // while Assigned(L)
        //      
        Node:= NodeNext;
      end; // while Assigned(Node)

      //       
      if (Round(2*Res) mod 2) <>0 then begin
        //   () 
        Res:= Round(2*Res); //    2
        Mul_2; //      2
      end;
      //   
      //    
      // : Result -- , Res -- Real
      if Result > Round(Res) then Result:= Round(Res);
    end;
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    //         aDelta
    // (  CalcForExpand).
    //  TRUE,       

    function ChangeValues(aDelta: integer): boolean;
    var Node: TNode;
        Count: integer;
    begin
      Result:= false;

      //    
      Node:= NodeFirst;
      while Assigned(Node) do begin
        //     
        if (GetRoot(Node) = aRoot) then begin
          //   
          if GetOut(Node)
            then Dec(Node.mValue, aDelta)   // Node  -
            else Inc(Node.mValue, aDelta);  // Node  +
        end;
        Node:= NodeNext;
      end; // while

      //  ,    gBlossomStack
      Node:= gBlossomStack.GetFirst as TNode;
      while Assigned(Node) do begin
        with (Node as TBlossom) do begin
          if (GetRoot(Node) = aRoot)  //     
            //    
            then if GetOut(Node)
              then Inc(mValue, 2*aDelta)  //  + 2*aDelta
              else Dec(mValue, 2*aDelta); //  - 2*aDelta
        end; // with
        Node:= gBlossomStack.GetNext as TNode;
      end; // while

      //      gBlossomStack
      //      ,
      //    .
      Count:= gBlossomStack.GetCount;
      while Count > 0 do begin
        Dec(Count);
        //     
        Node:= gBlossomStack.Get as TNode;
        with (Node as TBlossom) do begin
          if GetRoot(Node) = aRoot then begin
            //     
            if not GetOut(Node) //   
               and (mValue = 0) { } then begin
              //    ,
              //  
              Result:= true;  //   
              BlossExpand(Node as TBlossom);
              Node.Free;      //   
              Continue;       //    
            end; // if
          end; // if
        end; // with Node as TBlossom
        gBlossomStack.Put(Node);  //    ()
      end; // while
     end; // function

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //   mOut     false ()
    //      .
    //     .

    procedure LocalReset;
    var Node : TNode;
    begin
      PosPush;  //   
      Node:= NodeFirst;
      while Assigned(Node) do begin
        if Node.mBlossom = Node then with Node do begin
          //      
          mOut:= false;
          mRoot:= nil;
        end;
        Node.mDist:= 0;
        Node:= NodeNext;
      end;
      PosPop;  //   

      //  
      Node:= gBlossomStack.GetFirst as TNode;
      while Assigned(Node) do begin
        with Node do begin
          mOut:= false;
          mRoot:= nil;
          mDist:= 0;
        end;
        Node:= gBlossomStack.GetNext as TNode;
      end;
    end;
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //     aNode
    //     
    //     aRoot

    function NodeHandle(aNode: TNode): TNode;
    var Dest : TNode;      //  
        Pair : TNode;      //   Dest
        NodeBloss : TNode; //   aNode
        DestBloss : TNode; //   Dest
        PairBloss : TNode; //   Pair
        Bloss : TNode;     //   
        Base : TNode;      //   NodeBloss
        Node : TNode;      //    
        Link : TLink;      //     
        Profit : integer;  //   
        d : integer;       //   Link
    begin
      Result:= nil; //     
      //   ,  
      //   aNode (  )
      NodeBloss:= aNode.GetBlossom;
      //     ?
      if NodeBloss is TBlossom then with NodeBloss as TBlossom do begin
        //  NodeBloss --  
        //    ?
        if Assigned(mPair) then begin
          //   ,   ()  ,
          //     mPair
          if IsPresent(mPair.mOwner)
            then Base:= mPair.mOwner // mOwner  
            else Base:= mPair.mDest; // mDest  
        end else begin
          //     
          Base:= GetNewBase;   //    
        end;
        //        
        //    Base   aNode
        Profit:= GetProfit(Base, aNode);
        //     
        aNode.mDist:= Base.mDist + Profit;
      end; // if NodeBloss is TBlossom

      //  aNode    
      if (aNode.mDist > 0) and Assigned(NodeBloss.mPair) then begin
        //     
        Result:= aNode; //   
        Exit;           //    
      end;

      //  aNode     ,
      //    aNode,
      //   :
      // a)   () 
      // b)   ,  
      // c)   ,  
      // d)  

      //   
      Link:= aNode.OutLinkFirst;
      while Assigned(Link) do begin
        //    ( )
        d:= Delta(Link);
        if d <> 0 then begin
          //  L ,  
          Link:= aNode.OutLinkNext; Continue;
        end;
        //  d = 0,   L 
        Dest:= Link.mDest;  //    Dest
        //   Link.mDest
        //    DestBloss
        DestBloss:= Dest.GetBlossom;
        if DestBloss.GetPair <> aNode then
          //  DestBloss   
          //    aNode
          // Dest   ?
          if GetRoot(Dest) <> aRoot then begin
            //   Dest    
            //    (mDist)   
            DestBloss.mDist:= aNode.mDist + Link.mValue;
            with DestBloss do begin
              //     (  )
              mRoot:= aRoot;  //    
              mOut:= false;   //   
              mPred:= NodeBloss; //    
              mLink:= Link;   //     
              //     
              Pair:= GetPair; // DestBloss.GetPair;
            end;
            //    DestBloss ,
            //    ,
            //  ,  
            if not Assigned(Pair) then begin
              //      L.mDest,
              //  ,   
              Result:= DestBloss;       //   
              Result.mPred:= NodeBloss; //  
              Result.mLink:= Link;      //   
              Break;                    //   
            end // if not Assigned(Pair)
            else begin
              //    Pair (DestBloss.mPair)
              //    (   )
              PairBloss:= Pair.GetBlossom;
              with PairBloss do begin
                //     
                mRoot:= aRoot;  //    
                mOut:= true;    //   
                Link:= mPair;   //   Dest <=> Pair
                //      
                //    (   )
                Pair.mDist:= DestBloss.mDist - Link.mValue;
                mDist:= Pair.mDist; // PairBloss.mDist
              end; // with PairBloss

              //      ...
              if DestBloss is TBlossom then
                with DestBloss as TBlossom do begin
                //       Node
                //     Pair.mDist  
                // GetProfit(Node, Dest)
                if IsPresent(Link.mOwner)
                  then Node:= Link.mOwner
                  else Node:= Link.mDest;
                Profit:= GetProfit(Node, Dest);
                Inc(Pair.mDist, Profit);
              end; // if DestBloss is TBlossom
              //      
              //    
              PutInQue(gQue, PairBloss);
            end // else Assigned(Pair)
          end else begin
            //  GetRoot(Dest) = aRoot,
            //  ,   L.mDest,
            //    
            if aNode.GetBlossom <> Dest.GetBlossom then begin
              //       
              if GetOut(Dest) then begin
                //  Dest   ,
                //     
                Bloss:= BlossomMake(Link);//  
                Bloss.mRoot:= aRoot;      //   
                Bloss.mOut:= true;        //  
                gBlossomStack.Push(Bloss);//   
              end;
            end;  // if aNode.GetBlossom <> Dest.GetBlossom
          end; // else
        Link:= aNode.OutLinkNext;
      end; // while Assigned(L)
    end;
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  var Node : TNode;       //   
      DeltaMin: integer;  //     
      Success: boolean;   //    

  begin  // function Alter(aRoot: TNode): TNode;

    //        aRoot
    //        

    Result:= nil;
    //       
    LocalReset;

    //    
    with aRoot do begin
      mRoot:= aRoot;  //  
      mOut:= true;    //  
      mPred:= nil;    //    
      mLink:= nil;    //    
    end;

    //     gQue
    // aRoot --   (  )

    PutInQue(gQue, aRoot);

    repeat
      // gAppendix --     .
      //   CalcForExpand
      //     .
      //     .
      //      :
      with gAppendix do begin
        Node:= GetFirst as TNode;
        while Assigned(Node) do begin
          gQue.Put(Node);   //    
          Node:= GetNext as TNode;
        end;
      end;
      //    
      //     
      Node:= gQue.Get as TNode;
      while Assigned(Node) do begin
        Node.mRoot:= aRoot; //   
        //   Node 
        //     .
        //     .
        Result:= NodeHandle(Node);
        //     
        //  
        if Assigned(Result) then Break;
        //   
        Node:= gQue.Get as TNode;
      end; // while Assigned(Node)

      //       
      //   .

      if Assigned(Result) then Break;

      //      .
      //     ,
      //     .
      //  Success --    .

      DeltaMin:= CalcForExpand(Success);

      //     ,  

      if not Success then Break;

      //        DeltaMin.
      //  ChangeValues     TRUE,
      //     
      //       ,
      //   .

      if ChangeValues(DeltaMin) then begin
        //        
        //        
        //     .
        gMainQue.Push(aRoot.GetBlossom);
        Break;  //   
      end;

      //        
      //    .
      //    
      //    .
    until false;
    //    .
    gAppendix.Clear;
    gQue.Clear;
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      (  ).
  //      ,
  //     .
  // aLink -- ,  .

  procedure PairMake(aLink : TLink);
  var Owner, Dest : TNode;
  begin
    //   ,    aLink
    with aLink do begin
      Owner:= mOwner.GetBlossom; //   
      Dest:= mDest.GetBlossom;   //   
    end;
    //    :
    //  :
    Owner.mPair:= aLink; Dest.mPair:= aLink;
    //     
    Owner.mOwnDest:= false; //  
    Dest.mOwnDest:= true;   //  
    //     
    //       ()
    if Owner is TBlossom then with Owner as TBlossom do begin
      SetNewBase(aLink.mOwner);
    end;
    if Dest is TBlossom then with Dest as TBlossom do begin
      SetNewBase(aLink.mDest);
    end;
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   aNode <-> aNode.mPair
  //     ,
  //     .
  //     aBloss,
  //      ,
  //   (nil) --    .

  procedure Disconnect(aNode, aBloss : TNode);

    //      mNewBase
    //    (mBase)

    procedure BasesRestore(aNode : TNode);
    var Node: TNode;
    begin
      //   mNewBase,
      //       
      // ()
      Node:= aNode.mBlossom;  //  
      repeat
        //      
        with Node as TBlossom do mNewBase:= mBase;
        // ,    
        if Node = Node.mBlossom then Break;
        //     
        Node:= Node.mBlossom;
      until false;
    end;

  var Node, NodeBloss : TNode;
  begin
    //    ?
    if Assigned(aBloss) then begin
      //       aBloss
      // (   ),
      //    ,
      //      aBloss
      NodeBloss:= aNode;
      //   :
      while NodeBloss.mBlossom <> aBloss do
        NodeBloss:= NodeBloss.mBlossom;
    end else begin
      //      (  )
      //      Node
      NodeBloss:= aNode.GetBlossom;
    end;

    //     
    Node:= NodeBloss;  //  
    repeat
      Node.mPair:= nil; //    
      // ,     
      if not (Node is TBlossom) then Break;
      //    ()  
      Node:= (Node as TBlossom).mBase;
    until false;
    //   aNode    NodeBloss,
    //      aBloss,
    //     NodeBloss ()
    if (NodeBloss is TBlossom) and not Assigned(aBloss)
      then BasesRestore(aNode);
  end;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   
  //    :
  // a)  --       
  // b)         

  procedure BlossExpand(aBloss: TBlossom);

    //      

    procedure RemoveFromBlossom;
    var N : TNode;
    begin
      //      
      N:= aBloss.mBase;
      repeat
        N.mBlossom:= N; //    
        N.mRoot:= nil;  //    
        N:= N.mRight;
      until N = aBloss.mBase;
    end;

    //        
    // ()
    
    function GetMinNode(aBloss : TBlossom): TNode;
    var Node, N: TNode;
        MinVal: integer;
    begin
      MinVal:= MaxInt;
      with aBloss as TBlossom do begin
        Result:= mBase;
        Node:= mBase;
        //     
        repeat
          // Node -  
          // N --  
          //    .
          if Node is TBlossom
            then N:= GetMinNode(Node as TBlossom)
            else N:= Node;
          if N.mValue < MinVal then begin
            MinVal:= N.mValue;
            Result:= N;
          end;
          Node:= Node.mRight;
        until Node = mBase;
      end;
    end;

  var N1, N2 : TNode; //     
      Next: TNode;    //    
      Base : TNode;   //  ()  
      Pair : TNode;   //   
      Delta: integer; //      
      Link: TLink;    //    
      Direct: boolean;//   

  begin  // procedure BlossExpand

    with aBloss do begin
      // ,      (mPair)
      if Assigned(mPair) then begin
        //    
        //      
        if mOwnDest
          then Base:= mPair.mDest   //   mPair
          else Base:= mPair.mOwner; //   mPair
        Next:= SetNewBase(Base);    //     
        //  Next --   Base ( nil)
      end else begin
        //     ,
        //        ,
        //      .
        //        
        //   .
        //        
        Base:= GetMinNode(aBloss);
        //   ,   
        while Base.mBlossom <> aBloss do Base:= Base.mBlossom;
        //     
        Next:= SetNewBase(Base);
        //  
        Pair:= Base.GetPair;
        if Assigned(Pair) then begin
          Disconnect(Base, aBloss); //   
          Disconnect(Pair, aBloss); //   
        end;
      end;
      //     
      // (   ()  )
      if mValue <> 0 then begin
        Delta:= mValue div 2;  // mValue   
        N1:= mNewBase.mRight;
        repeat
          if N1 is TBlossom
            then Inc(((N1 as TBlossom).GetNewBase).mValue, Delta)
            else Inc(N1.mValue, Delta);
          N1:= N1.mRight;
        until N1 = mNewBase;
      end;

      //    ,
      //     

      if mNewBase=mBase then begin
        //     
        //      
        RemoveFromBlossom;
        //       
        if Assigned(mPair) then PairMake(mPair);
        Exit;
      end;

      //     (Next <> NIL)
      //      
      RemoveFromBlossom;

      //     
      Next:= Next.GetBlossom;

      //     ,  
      //   
      Direct:= Next = mNewBase.mRight;
      //  
      N1:= mNewBase; //    
      repeat
        if Direct then begin
          //    
          N2:= N1.mRight;
          Link:= N2.mRLnk;  //   
          N1:= N2.mRight;
        end else begin
          //    
          N2:= N1.mLeft;
          N1:= N2.mLeft;
          Link:= N1.mRLnk;  //   
        end;
        PairMake(Link);//   N1 <=> N2
      until N1= mBase; //    

      //        
      if Assigned(mPair) then PairMake(mPair);
    end;   // with aBloss
  end;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    .
  //   
  //  ,    .
  // aFin --   ()  

  procedure Reverse(aFin : TNode);
  var N1, N2, Next : TNode;
  begin
    //     
    //      
    Next:= aFin;
    repeat
      N2:= Next.GetBlossom;  // N2 -- 
      N1:= N2.mLink.mOwner.GetBlossom;  // N1 -- 
      Next:= N1.GetPair;  // Next --  
      PairMake(N2.mLink); //   N1 <=> N2
      //     
    until not Assigned(Next);
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      aNode

  procedure Handle(aNode: TNode);
  var FreeNode,      //   
      FreeBloss,     //   
      Start : TNode; //     
  begin
    //       aNode
    //     ,  
    //  .
    // FreeNode --   

    FreeNode:= Alter(aNode);

    //     ,  
    if not Assigned(FreeNode) then Exit;

    //    
    // (  )

    //    
    // (    )
    FreeBloss:= FreeNode.GetBlossom;
    //    ?
    if not Assigned(FreeBloss.mPair) then begin
      //      ,
      //     
      Start:= FreeBloss;  //     
    end else begin
      // ,  , --    .
      //     
      //       
      Start:= FreeBloss.GetPair.GetBlossom;
      //      ?
      if FreeBloss is TBlossom then with FreeBloss as TBlossom do begin
        // ,    
        Disconnect(GetNewBase, nil);  //   
        SetNewBase(FreeNode); //     
      end else begin
        // ,       
        Disconnect(FreeNode, nil);  //    
      end;
      //       
      gMainQue.Push(FreeBloss);
    end;
    //        ,
    //     ,
    //    Start.
    if Start <> aNode.GetBlossom then Reverse(Start);
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //        .
  //    ,   

  procedure RemoveBlossoms;
  var Bloss: TBlossom;
  begin
    while gBlossomStack.GetCount>0 do begin
      Bloss:= gBlossomStack.Pop as TBlossom;
      BlossExpand(Bloss);
      Bloss.Free;
    end;
  end;
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   .
  //        
  //    .

  function PairsMake: TCostSet;
  var   Node: TNode;
        Pair: TPair;
           L: TLink;
           S: TSet;
        Cost: integer;
  begin
    Cost:= 0;
    //   S    
    S:= CreateSet;
    //   
    Node:= NodeFirst;
    while Assigned(Node) do begin
      if Assigned(Node.mPair) then begin
        //   Node  ,
        //  -
        Pair:= TPair.CreateEmpty;
        with Pair do begin
          //   
          mDestroy:= true;
          mSet.Insert(Node);
          mSet.Insert(Node.GetPair);
          L:= Node.GetLink(Node.GetPair);
          //   L.mTemp    L.mValue
          mCost:= L.mTemp;
          Inc(Cost, mCost);
        end;
        //     
        S.Insert(Pair);
        //   mPair
        Node.GetPair.mPair:= nil;
        Node.mPair:= nil;
      end;
      Node:= NodeNext;
    end;
    //     ,
    //      
    if aMode = pMaxN then Cost:= S.GetCount;
    //     Cost
    Result:= TCostSet.Create(Cost, S, true);
  end;

var Node: TNode;  //  

begin

  Self.ResetNodes;  //   ()
  InitGraphValues;  //     

  gMainQue:= TBuffer.Create;//    
  gQue:= TBuffer.Create;    //   
  gBlosNum:= 0;             //   
  gAppendix:= CreateSet;    //     
  gBlossomStack:= TBuffer.Create;  //    

  //      
  Node:= NodeFirst;
  while Assigned(Node) do begin
    gMainQue.Put(Node);
    Node:= NodeNext;
  end;

  //   
  while gMainQue.GetCount>0 do begin
    //    
    Node:= gMainQue.Get as TNode;
    //   ?
    if Assigned(Node.mPair) //     
      //      
      or Assigned(Node.GetBlossom.mPair)
      then Continue;        //   
    //  ()  
    Handle(Node);
  end;

  //  :
  RemoveBlossoms;     //   
  gBlossomStack.Free; //    
  gAppendix.Free;     //     
  gQue.Free;          //   
  gMainQue.Free;      //    
  Result:= PairsMake; //   
  RestoreGraphValues; //     
end;

//########################################################################
//                                                                       #
//                                          #
//                                                                       #
//########################################################################

//    .
//    ʸ,    ( )
// , . 405
//    ,   mHigh=0
//       aPairs

//     :
// * TNode.mColor --  - ,  -  
// * TNode.mDist  --  ,  "  "
// * TNode.mFlow  --    (0/1)
// * TNode.mRoot  -- ,     
// * TNode.mLink  --      
//
// * TLink.mColor --  -    
// * TLink.mValue --  (, ) 
// * TLink.mHigh  --   , 0 - , 1 - 
// * TLink.mLow   --  ()  

function TGraph.MarkMinPairsDicoty(var aPairs: integer): integer;

var    Que: TBuffer;  //  
    NodesL: TSet;     //   
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     :
  //     
  //     
  //     

  function  InitNodes: integer;
  var Node: TNode;
      Link: TLink;
      L, R: integer; //     
  begin
    //   
    L:= 0; R:= 0;
    //   :
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mColor:= CWhite;   //   
      if Node.IsLeft then begin
        Inc(L);               //   
        Node.mDist:= MaxInt;  //     
        NodesL.Insert(Node);  //     
      end else if Node.IsRight then begin
        Inc(R);               //   
        Node.mDist:=0;        //   = 0
      end;
      Node:= NodeNext;
    end; // while
    //       :
    if L > R then Result:= R else Result:= L;

    //   :
    Node:= NodesL.GetFirst as TNode;
    while Assigned(Node) do begin
      //   
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        with Link do if mHigh=0 then begin
          //   
          Link.mColor:= CWhite;
          //    
          if Node.mDist > Link.mValue  then
            Node.mDist:= Link.mValue; //  " "
        end;
        Link:= Node.OutLinkNext;
      end; // while
      Node:= NodesL.GetNext as TNode;
    end; // while

  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   () :
  // aLeft -      ()
  //     (),
  //  ,    ,   NIL.
  //     TNode.mRoot,    .
  //     ,
  //   aDelta    ,
  //        .

  function FindAugmenting(aLeft : TNode; var aDelta: integer): TNode;
  var Node: TNode;
      Link: TLink;
      Cost: integer;
  begin
    Result:= nil;       //    
    Que.Clear;          //   
    Que.Put(aLeft);     //     
    //         
    while (Que.GetCount > 0) and not Assigned(Result) do begin
      Node:= Que.Get as TNode;
      Node.mFlow:= 1;   //   
      if Node.IsLeft then begin
        //    
        //      ,     
        Link:= Node.OutLinkFirst;
        while Assigned(Link) do begin
          //     
          with Link do if (mHigh=0) and    //  
                          (mColor=CWhite)  //   
            then begin
              //   ()  
              Cost:= mValue - (mOwner.mDist - mDest.mDist);
              if Cost = 0 then begin
                //     (Cost=0)
                if  mDest.mColor = CWhite then begin
                  //   ,
                  //       
                  Result:= mDest;      //   
                  Result.mRoot:= Link; //     
                  Break;               //     
                end else if mDest.mFlow = 0 then begin
                  //   ,
                  //    ,      
                  mDest.mRoot:= Link; //     
                  Que.Put(mDest);     //   
                end; // else
              end else begin
                //     (Cost > 0),
                //  aDelta    
                if aDelta > Cost then aDelta:= Cost;
              end;
          end; // with Link
          Link:= Node.OutLinkNext;
        end; // while -   
      end else begin
        //        (Node.mIsLeft = FALSE)
        Link:= Node.mLink;        //     
        Link.mOwner.mRoot:= Link; //      
        //          
        Que.Put(Link.mOwner);
      end;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   () .
  //   ,    
  //   FindAugmenting.
  // aLeft, aRight --     () .
  //   Node.mRoot,   FindAugmenting,
  //       Node.mLink.

  procedure Inverse(aLeft, aRight : TNode);
  var Node: TNode;
      Link: TLink;
      Flag: boolean;
  begin
    //      
    //   
    Flag:= true;     //      ()
    Node:= aRight;
    repeat
      Link:= Node.mRoot as TLink;
      if Flag then begin
        //   ,   
        //     
        with Node do if Assigned(mLink) then mLink.mColor:= CWhite;
        Node.mLink:= Link;      //    
        Link.mColor:= CBlack;   //    
        Node:= Link.mOwner;     //     
        Node.mLink:= Link;      //      
      end else begin
        //   ,   
        Link.mColor:= CWhite;
        Node:= Link.mDest;     //     
      end;
      Flag:= not Flag; //      ()
    until Node = aLeft;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      TNode.mFlow --  .
  //     ()  

  procedure ClearFlags;
  var Node: TNode;
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mFlow:= 0;   //      
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   (, )  
  //      .
  //     (  ).
  //        ,
  //    aDelta   " ",
  //       

  function FindPairs(var aDelta: integer): integer;
  var LNode, RNode : TNode;
  begin
    Result:= 0;       //   
    aDelta:= MaxInt;  //   
    ClearFlags;       //     TNode.mFlow
    //   
    LNode:= NodesL.GetFirst as TNode;
    while Assigned(LNode) do begin
      if LNode.mColor = CWhite then begin
        //   ()  
        //       aDelta
        RNode:= FindAugmenting(LNode, aDelta);
        if Assigned(RNode) then begin
          //   
          //     RNode  LNode
          Inverse(LNode, RNode);
          //      
          //    
          LNode.mColor:= CBlack;
          RNode.mColor:= CBlack;
          Inc(Result);  //    
        end;
      end;
      LNode:= NodesL.GetNext as TNode;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     Node.mDist   aDelta
  // ( )
  //     () 
  //  ,       .

  procedure Correct(aDelta: integer);
  var Node: TNode;
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      //   ,    
      if Node.mFlow <> 0 then begin
        Inc(Node.mDist, aDelta);
      end;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //       
  //     ()  TLink.mLow
  //       TNode.mDist.
  // ( TLink.mLow     )

  function CalcCost: integer;
  var Node: TNode;
      Link: TLink;
  begin
    Result:= 0;
    //    
    Node:= NodesL.GetFirst as TNode;
    while Assigned(Node) do begin
      // Node.mLink -    
      Inc(Result, Node.mLink.mValue);
      //   
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        //    ,   
        with Link do
          if mHigh=0 then mLow:= mValue - (mOwner.mDist - mDest.mDist);
        Link:= Node.OutLinkNext;
      end;
      Node:= NodesL.GetNext as TNode;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var
   P: integer;     //      
   Delta: integer; // ,    
   Count: integer; //     

begin  { TGraph.SetPairsDicoty }

  Que:= TBuffer.Create; //  
  NodesL:= CreateSet;   //   
  Count:= InitNodes;    //     
  aPairs:= Count;

  repeat     //     
    repeat   // ,    
      P:= FindPairs(Delta); // P -    
                            // Delta - .    
      Dec(Count, P);        // Count -   
    until (P=0) or (Count=0);
    //    ,    
    //   Delta ( ),
    // ""      
    if Count > 0 then Correct(Delta);
  until Count=0;
  //    ,
  //       
  Result:= CalcCost;
  //  :
  NodesL.Free;  //   
  Que.Free;     //  
end;


//     ,    
//  MarkMinPairsDicoty

function TGraph.GenMinPairsDicoty: TCostSet;
var Cost: integer;   // 
    Pairs: integer;  //   
    Node: TNode;
    S : TSet;
begin
  Cost:= MarkMinPairsDicoty(Pairs);
  S:= CreateSet;
  Node:= NodeFirst;
  while Assigned(Node) do begin
    if Node.IsLeft
      then S.Insert(Node.mLink);
    Node:= NodeNext;
  end;
  Result:= TCostSet.Create(Cost, S, true);
end;


//     
//     

function TGraph.GenMinPairsDicotyFlow: TCostSet;
var  S, T : TNode;
     Node : TNode;
     Link : TLink;
     Cost : integer;
      Cnt : integer;
      Res : TSet;

begin

  //       
  Node:= NodeFirst;
  while Assigned(Node) do begin
    Link:= Node.OutLinkFirst;
    while Assigned(Link) do begin
      with Link do begin
        mTemp:= mHigh;  //  
        mHigh:= 1;      //     
      end;
      Link:= Node.OutLinkNext;
    end;
    Node:= NodeNext;
  end;

  //    S   R:
  S:= TNode.Create(0, Self);
  T:= TNode.Create(0, Self);

  //  ,     ,    
  //    =0,   =1
  //      
  Cnt:= 0;
  Node:= NodeFirst;
  while Assigned(Node) do begin
    if Node.IsLeft then begin
      //      ,
      Link:= S.MakeLink(Node, 0);
      Link.mHigh:= 1;
      Inc(Cnt);
    end  else begin
      //      ,
      Link:= Node.MakeLink(T, 0);
      Link.mHigh:= 1;
    end;
    Node:= NodeNext;
  end;  // while

  //          :
  if Cnt > mNodes.GetCount div 2 then Cnt:= mNodes.GetCount - Cnt;

  //   :
  InsertNode(S);   // 
  InsertNode(T);   // 

  //    :
  Cost:= CalcMinCostFlow(S, T, Cnt);

  //       :

  RemoveNode(T);
  RemoveNode(S);
  T.Free; S.Free;

  Res:= CreateSet;
  if Cost>=0 then begin
    //   
    //        (Link.mFlow <> 0)
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        with Link do begin
          if mFlow <> 0 then Res.Insert(Link);
          mHigh:= mTemp; //   
        end;
        Link:= Node.OutLinkNext;
      end;
      Node:= NodeNext;
    end;
  end; // if
  //    :
  Result:= TCostSet.Create(Cost, Res, true);
end;


//########################################################################
//                                                                       #
//                                            #
//                                                                       #
//########################################################################

//    ,  ,  ,
//       

function TGraph.GenPostPath(var aCost: integer): TBuffer;
begin
  if mDirect
    then Result:= GenPostPathDir(aCost)
    else Result:= GenPostPathUndir(aCost)
end;

//   ( )   

function TGraph.GenPostPathUndir(var aCost: integer): TBuffer;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    (  TLink.mLimit)

  procedure ModifyLimit(aLink: TLink; aDelta: integer);
  var L: TLink; //  
  begin
    Inc(aLink.mLimit, aDelta);              //  
    L:= aLink.mDest.GetLink(aLink.mOwner);  //  
    Inc(L.mLimit, aDelta);
  end;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      
  //     TNode.mLimit

  function GenOddNodes: TSet;
  var Node: TNode;
  begin
    Result:= nil;
    //   :
    Node:= NodeFirst;
    while Assigned(Node) do begin
      // :  mColor= 0;  mPred= mLink= nil;  mDist= mFlow= MaxInt;
      Node.ResetNode;
      //   mLimit     ():
      Node.mLimit:= Node.OutGetCnt;  //Node.mLnkOut.GetCount;
      //    ,    :
      if Odd(Node.mLimit) then begin
        if not Assigned(Result) then Result:= CreateSet;
        Result.Insert(Node);
      end;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      TLink.mLimit
  //      TNode.mLimit

  procedure LimitsPrepare;
  var OddNodes: TSet;     //     
        Pairs : TCostSet; //   
         Pair : TPair;    //  
           Gr : TGraph;   //  
      Ni, Nj  : TNode;    //   
      N2i,N2j : TNode;    //      
         NFin : TNode;    //     
         Dist : integer;  //    
         Link : TLink;
         i, j : integer;
  begin
    //      :
    SetLinksLimit(1);
    //      :
    OddNodes:= GenOddNodes;
    //    ,  :
    if not Assigned(OddNodes) then Exit;

    //   :
    Gr:= TGraph.Create('Odd Nodes:', //   
                        false,       //   
                        false,       //   
                        true         //  
                       );
    //         :
    Ni:= OddNodes.GetFirst as TNode;
    while Assigned(Ni) do begin
      Gr.InsertNode(Ni.Copy as TNode);
      Ni:= OddNodes.GetNext as TNode;
    end;
    //        
    //   :
    Self.InitMap_Floyd;
    //       
    // (  ).
    //       
    //    
    for i:= 1 to OddNodes.GetCount-1 do begin
      Ni:= OddNodes.GetItem(i) as TNode;     //    
      N2i:= Gr.mNodes.GetItem(i) as TNode;   //    
      //    Ni :
      for j:= i+1 to OddNodes.GetCount do begin
        Nj:= OddNodes.GetItem(j) as TNode;   //    
        N2j:= Gr.mNodes.GetItem(j) as TNode; //    . 
        Dist:= Ni.GetFarLink(Nj).mDist; //    . 
        Gr.SetLink(N2i, N2j, Dist);     //     .
      end;
    end;
    //   
    //     :
    Pairs:= Gr.GenPairs_Edmonds(pMinW);

    //        mLimit
    //    :
    Pair:= Pairs.mSet.GetFirst as TPair;
    while Assigned(Pair) do begin
      // Pair.mSet     .
      //      :
      Ni:= mNodes.GetObject(Pair.mSet.GetFirst as TNode) as TNode;
      //      :
      NFin:= mNodes.GetObject(Pair.mSet.GetNext as TNode) as TNode;
      repeat
        Inc(Ni.mLimit);         // +1   (.) 
        Nj:= Ni.GetNear(NFin);  //     
        Link:= Ni.GetLink(Nj);  //    
        ModifyLimit(Link, +1);  // +1 
        Ni:= Nj;                //    
      until Ni = NFin;          //    
      Inc(Ni.mLimit);           // +1   
      Pair:= Pairs.mSet.GetNext as TPair;  //  
    end;
    //  :
    Pairs.ClrAndDestroy;    // 
    Pairs.Free;             //  
    DoneMap;                //  
    Gr.Free;                //  
    OddNodes.Free;          //     
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      aNode

  procedure AddPath(aNode: TNode);
  var  Link: TLink;
  begin
    while true do begin
      //     
      // (  Link.mLimit > 0)
      Link:= aNode.OutLinkFirst;
      while Assigned(Link) and (Link.mLimit <= 0)
        do Link:= aNode.OutLinkNext;
      //    ,     :
      if not Assigned(Link) then Break;
      //    (  ):
      Result.Put(Link);        //     
      Dec(aNode.mLimit);       //    
      aNode:= Link.mDest;      //    
      Dec(aNode.mLimit);       //    
      Inc(aCost, Link.mValue); //   ( )
      ModifyLimit(Link, -1);   //   
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var Node: TNode;
    Link: TLink;
    Count: integer; //     Result

begin { TGraph.GenPostPathUndir }

  Result:= TBuffer.Create;  //    
  aCost:= 0;                //  
  LimitsPrepare;            //   mLimit ( .)
  Node:= NodeFirst;         //     

  //   ,
  //        
  repeat
    //      Node    :
    AddPath(Node);

    //    ()   
    //    (Node.mLimit > 0)
    Node:= nil;
    Count:= Result.GetCount;        //     Result
    while Count>0 do begin
      Link:= Result.Get as TLink;   //     
      Result.Put(Link);             //    
      Dec(Count);
      //   mLimit    ,  .
      //       Link.mDest
      if Link.mDest.mLimit > 0 then begin
        Node:= Link.mDest;
        Break;
      end;
    end;
    //   ,       :
  until not Assigned(Node);

  //   ,      :
  Node:= NodeFirst;
  Count:= Result.GetCount;
  while Count>0 do begin
    Link:= Result.Get as TLink; //     
    //     ?
    if Link.mOwner = Node then begin
      Result.Push(Link);        // ,    
      Break;                    //   
    end;
    Result.Put(Link);           //     
    Dec(Count);
  end;
end;

//########################################################################
//
//        
//
//########################################################################

function TGraph.GenPostPathDir(var aCost: integer): TBuffer;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   mLimit     .
  //    

  function LimitsPrepare: integer;
  var Node : TNode;    //  
      Link : TLink;    //  
      Flag : boolean;  //    
      S, T : TNode;    //    
      Flow: integer;   //   
  begin
    Result:= 0;
    //       :
    SetLinksLimit(1);
    //         ,
    //     TNode.mLimit:
    Flag:= false;        //   
    //  :
    Node:= NodeFirst;
    while Assigned(Node) do begin
      with Node do begin
        mLimit:= InGetCnt + OutGetCnt;  //  
        mDist:= InGetCnt - OutGetCnt;   // 
        Flag:= Flag or (mDist<>0); //    
      end;
      Node:= NodeNext;
    end;
    //     (  ),
    //    :

    if not Flag then Exit;

    //       
    //  ,   
    // :
    S:= TNode.Create(0, Self); //  
    T:= TNode.Create(0, Self); //  

    Flow:= 0; //    
    //   ,  S  T:
    Node:= NodeFirst;
    while Assigned(Node) do begin
      //        :
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        Link.mLow:= 0;            //   = 0
        Link.mHigh:= MaxInt;      //   
        Link:= Node.OutLinkNext;
      end;
      //    (mDist <> 0),
      //       :
      if Node.mDist < 0 then begin
        //       :
        Link:= Node.MakeLink(T, 0); //  
        Link.mHigh:= -Node.mDist;   //   
      end else if Node.mDist > 0 then begin
        //   :
        Inc(Flow, Node.mDist);      //  
        //       :
        Link:= S.MakeLink(Node, 0); //  
        Link.mHigh:= Node.mDist;    //   
      end;
      Node:= NodeNext;
    end; // while
    //     :
    mNodes.Insert(S);         // 
    mNodes.Insert(T);         // 

    //   Flow      :
    Result:= CalcMinCostFlow(S, T, Flow);

    //     :
    RemoveNode(S);  S.Free;     //  
    RemoveNode(T);  T.Free;     //  

    //          
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        Flow:= Link.mFlow;  //   
        //   
        //         :
        if Flow > 0 then begin
          Inc(Link.mLimit, Flow);         // +  
          Inc(Link.mOwner.mLimit, Flow);  // +   
          Inc(Link.mDest.mLimit, Flow);   // +   
        end;
        Link:= Node.OutLinkNext;
      end;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      aNode

  procedure AddPath(aNode: TNode);
  var  Link: TLink;
  begin
    while true do begin
      //      (Link.mLimit > 0)
      Link:= aNode.OutLinkFirst;
      while Assigned(Link) and (Link.mLimit <= 0)
        do Link:= aNode.OutLinkNext;
      //    ,   :
      if not Assigned(Link) then Break;
      //    (  ):
      Result.Put(Link);        //     
      Dec(aNode.mLimit);       //    
      aNode:= Link.mDest;      //    
      Dec(aNode.mLimit);       //    
      Inc(aCost, Link.mValue); //   ()
      Dec(Link.mLimit);        //   
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var Node: TNode;     //  
    Link: TLink;     //  
    Count: integer;  //     Result

begin { TGraph.GenPostPathDir }

  Result:= TBuffer.Create;  //    
  aCost:= -1;    //      

  //     ,   :

  if not IsLinked then Exit;

  aCost:= 0;          //  
  LimitsPrepare;      //   mLimit   
  Node:= NodeFirst;   //     

  //      Result:
  repeat
    //      Node
    //     Result:
    AddPath(Node);

    //    ()   
    //    (Node.mLimit > 0)
    Node:= nil;
    Count:= Result.GetCount;        //    
    while Count>0 do begin
      Link:= Result.Get as TLink;   //     
      Result.Put(Link);             //    
      Dec(Count);
      //   mLimit    , 
      //       Link.mDest
      if Link.mDest.mLimit > 0 then begin
        Node:= Link.mDest;
        Break;
      end;
    end;
    // ,     
  until not Assigned(Node);

  //   ,      :
  Node:= NodeFirst;
  Count:= Result.GetCount;
  while Count>0 do begin
    Link:= Result.Get as TLink; //     
    //     ?
    if Link.mOwner = Node then begin
      Result.Push(Link);        // ,    
      Break;                    //   
    end;
    Result.Put(Link);           //     
    Dec(Count);
  end;
end;

//########################################################################
//                                                                       #
//                                    #
//                   , . 7, . 251                            #
//                                                                       #
//########################################################################

//        

function TGraph.GenHamilton_Full(var aCost: integer): TBuffer;
var
    Cost: integer;   //   
    Count: integer;  //   
    First: TNode;    //  
    Node: TNode;     //  
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    .
  //     .
  // aNode -   
  procedure SaveResult(aNode: TNode);
  begin
    aCost:= Cost;         //   
    First.mPred:= aNode;  //  
    Result.Clear;         //   
    Node:= aNode;         //     
    repeat
      Result.Push(Node);  //      
      Node:= Node.mPred;  //  
    until Node=First;     //     
    Result.Push(Node);    //   
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     

  procedure Local(aNode: TNode);
  var L : TLink;  //   
  begin
    aNode.mColor:= CBlack;  //   
    Inc(Count);             //    
    //   ?
    if Count < mNodes.GetCount then begin
      // ,      
      L:= aNode.OutLinkFirst;
      while Assigned(L) do begin
        if L.mDest.mColor = CWhite then begin
          //  :
          Inc(Cost, L.mValue);        //  
          if Cost < aCost then begin  //    
            L.mDest.mPred:= aNode;    //    
            Local(L.mDest);           //      .
          end;
          Dec(Cost, L.mValue);        //  
        end;
        L:= aNode.OutLinkNext;        //   
      end;
    end else begin
      //  Count = mNodes.GetCount --   
      //       First
      L:= aNode.OutLinkFirst;
      while Assigned(L) and (L.mDest<>First)
        do L:= aNode.OutLinkNext;
      if Assigned(L) then begin
        //   :
        Inc(Cost, L.mValue);      //  
        if Cost < aCost           //    
          then SaveResult(aNode); //     
        Dec(Cost, L.mValue);      //  
      end;
    end;
    //       
    Dec(Count);
    aNode.mColor:= CWhite;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var Temp: TBuffer; //      

begin { TGraph.GenHamilton_Full }

  Result:= TBuffer.Create;  //   
  aCost:= -1;                 //    
  //     ,   :
  if not IsLinked then Exit;

  //    aCost  
  Temp:= GenHamilton_Greed(aCost, true);
  Temp.Free;

  ResetNodes;           //   
  Cost:= 0;             //   = 0
  Count:= 0;            //    = 0
  First:= NodeFirst;    //  
  Local(First);         //   
  //    ,   1
  if aCost = MaxInt then aCost:=-1;
end;

//########################################################################
//    
//     
//########################################################################

function TGraph.GenHamilton_Greed(var aCost: integer;
                              aGreed: boolean
                              ): TBuffer;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  function GetAnyLink(aNode: TNode): TLink;
  begin
    Result:= aNode.OutLinkFirst;
    while Assigned(Result) and
         (Result.mDest.mColor <> CWhite)
      do Result:= aNode.OutLinkNext;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    
  function GetBestLink(aNode: TNode): TLink;
  var  Link: TLink;
       BestCost: integer;
  begin
    Result:= nil;
    BestCost:= MaxInt;
    Link:= aNode.OutLinkFirst;
    while Assigned(Link) do begin
      if (Link.mDest.mColor = CWhite) and
         (Link.mValue < BestCost) then begin
        BestCost:= Link.mValue;
        Result:= Link;
      end;
      Link:= aNode.OutLinkNext;
    end; // while
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var
    Cost: integer;   //   
    Count: integer;  //   
    Start: TNode;    //  
    Node: TNode;     //  
    L : TLink;       //  

begin { GenHamilton_Greed }

  Result:= TBuffer.Create;  //   
  aCost:= -1;                 //    
  //     ,   :
  if not IsLinked then Exit;
  ResetNodes;           //   
  Start:= NodeFirst;    //    
  Node:= Start;
  Cost:= 0;             //   = 0
  Count:= mNodes.GetCount-1; //   
  while Count > 0 do begin
    Node.mColor:= CBlack;    //   
    Result.Put(Node);
    //    
    if aGreed
      then L:= GetBestLink(Node) //   
      else L:= GetAnyLink(Node); //   
    //   ,  
    if not Assigned(L) then break;
    //  :
    Inc(Cost, L.mValue);    //  
    Dec(Count);             //  - 1
    Node:= L.mDest;         //  
  end; // while

  if Count=0 then begin
    //    
    Result.Put(Node);        //    
    L:= Node.GetLink(Start); //     
    if Assigned(L) then begin
      Inc(Cost, L.mValue);  //  
      aCost:= Cost;         //  
    end else begin
      Result.Clear;
    end;
  end;
end;

//########################################################################
//           
//      
//########################################################################

type  //      

      THamNode = class (TNode)
        mLeft : boolean;    //    (TRUE)
        mNode : TNode;      //  
        mBack : THamNode;   //      
        constructor Create(aOwner: TGraph; aNode: TNode; aLeft: boolean);
        function  Compare(arg: TItem): TCompare; override;
        function  GetName: string;  override;
        function  NextCircNode: THamNode;
        function  GenFactorsLinks: TSet;
      end;

constructor THamNode.Create(aOwner: TGraph; aNode: TNode; aLeft: boolean);
begin
  inherited Create(0, aOwner);
  mNode:= aNode;    //  
  mLeft:= aLeft;    // : TRUE - Left, FALSE - Right
end;

function  THamNode.GetName: string;
begin
  //  +   
  if mLeft then Result:= 'L.' else Result:= 'R.';
  Result:= Result + mNode.GetName;
end;

//    ,
//     ()

function THamNode.GenFactorsLinks: TSet;
var Link: TLink;
begin
  Result:= CreateSet;
  Link:= OutLinkFirst;
  while Assigned(Link) do begin
    //       ()
    //     (mHigh = 0),    
    with Link do
    if ((mDest as THamNode).mBack.mValue = Self.mValue) and (mHigh = 0)
      then Result.Insert(Link);  //    
    Link:= OutLinkNext;
  end;
end;

//      
//     

function THamNode.NextCircNode: THamNode;
begin
  Result:= (mLink.mDest as THamNode).mBack;
end;

//      

function  THamNode.Compare(arg: TItem): TCompare;
begin
  Result:= cmpEq;
  if arg = Self then Exit;
  if GetName < (arg as THamNode).GetName
    then Result:= cmpLess
  else if GetName > (arg as THamNode).GetName
    then Result:= cmpGreate
end;

//       
// Kristofides - 404

type
  THamGraph = class (TGraph)
    mGraph: TGraph;   //  () 
    constructor Create(aGraph : TGraph);
  protected
    function Check: boolean;
    function CalcFactors(var aStart: THamNode): integer;
    procedure SaveBest(aBuf: TBuffer);
    function CalcCostAndFactors(var aFactors: integer;//- 
                                var aStart: THamNode  //  
                                ): integer;           //  
    function Shrink(aFact: integer): THugeGraph;
  end;


constructor THamGraph.Create(aGraph : TGraph);
var Node: TNode;
    HamL, HamR: THamNode;
    Link : TLink;
begin
  inherited Create('Hamilton',  // 
                    true,       //  
                    false,      //   
                    true);      //  

  mGraph:= aGraph;            //    

  //         :
  Node:= aGraph.NodeFirst;
  while Assigned(Node) do begin
    //   
    HamL:= THamNode.Create(Self, Node, True);  // 
    HamR:= THamNode.Create(Self, Node, False); // 
    //      :
    HamL.mBack:= HamR;
    HamR.mBack:= HamL;
    //      
    mNodes.Insert(HamL);
    mNodes.Insert(HamR);
    Node:= aGraph.NodeNext;
  end;

  //  ,      :
  HamL:= NodeFirst as THamNode;
  while Assigned(HamL) do begin
    //   ?
    if HamL.mLeft then  begin
      // ,  
      PosPush;  //    
      //       
      Link:= HamL.mNode.OutLinkFirst;
      while Assigned(Link) do begin
        Node:= Link.mDest;  //    
        //       :
        HamR:= NodeFirst as THamNode;
        while Assigned(HamR) do begin
          with HamR do if not mLeft and (mNode = Node) then Break;
          HamR:= NodeNext  as THamNode;
        end;
        //  HamR     
        //        (  )
        HamL.MakeLink(HamR, Link.mValue);
        //      
        Link:= HamL.mNode.OutLinkNext;
      end; // while
      PosPop;  //    
    end; // if
    HamL:= NodeNext as THamNode;
  end; // while
end;

//      mValue   ()
//   .
//       mValue

function THamGraph.CalcFactors(var aStart: THamNode): integer;
var NodesCnt: integer;  //    
    Start, Left : THamNode; //    
    CircLen: integer;   //   
    MinLen : integer;   //   
    Link : TLink;
begin
  Result:= 0;         //  
  MinLen:= MaxInt;    //   
  //    :
  NodesCnt:= mGraph.mNodes.GetCount;
  //   mValue  :
  Start:= NodeFirst as THamNode;
  while Assigned(Start) do begin
    Start.mValue:= 0;
    Start:= NodeNext as THamNode;
  end;
  //       ()
  while NodesCnt>0 do begin
    Inc(Result);  // Result= 1, 2, .. -   
    CircLen:=0;   //    ()
    //       :
    Start:= NodeFirst as THamNode;
    while Assigned(Start) do begin
      with Start do if mLeft and (mValue = 0) then Break;
      Start:= NodeNext as THamNode;
    end;
    //    Start   
    //   Result   :
    // =>  -    ;
    // <=   -    mBack
    //      
    Left:= Start;
    repeat
      Dec(NodesCnt);        //   
      Inc(CircLen);         //    ()
      Left.mValue:= Result; //    
      // Left.mLink --    Left   Right
      //      (  )
      Link:= Left.mLink;
      //      :
      Left:= (Link.mDest as THamNode).mBack;
    until Left = Start;
    //     :
    if CircLen < MinLen then begin
      MinLen:= CircLen;   //   ()
      aStart:= Start;     //    ( )
    end;
  end; // while NodesCnt >0
end;

//      
//  () 
// :
// FALSE --            
// TRUE  --         
// (Link.mHigh <> 0) --    ()

function THamGraph.Check: boolean;
var Node: THamNode;
    Link: TLink;
begin
  Result:= false;
  //  
  Node:= NodeFirst as THamNode;
  while Assigned(Node) do begin
    if Node.mLeft then begin
      //      
      Link:= Node.OutLinkFirst;
      while Assigned(Link) and (Link.mHigh<>0)
        do Link:= Node.OutLinkNext;
      if not Assigned(Link) then Exit;
    end else begin
      //      
      Link:= Node.InLinkFirst;
      while Assigned(Link) and (Link.mHigh<>0)
        do Link:= Node.InLinkNext;
      if not Assigned(Link) then Exit;
    end;
    Node:= NodeNext as THamNode;
  end;
  Result:= true;  //   
end;

//      
//      

procedure THamGraph.SaveBest(aBuf: TBuffer);
var Start, Node : THamNode;
begin
  aBuf.Clear; //   
  //       
  Start:= NodeFirst as THamNode;
  while not Start.mLeft do Start:= NodeNext as THamNode;
  //    Start  Start
  Node:= Start;
  repeat
    aBuf.Put(Node.mNode);
    Node:= Node.NextCircNode;
  until Node = Start;
end;

//         
// aFact -  
//        :
// MarkMinPairsDicoty -> CalcFactors -> Shrink

function THamGraph.Shrink(aFact: integer): THugeGraph;
var i: integer;
    Link: TLink;
    N1, N2: TNodeInt;   //     
    NH1, NH2: THamNode; //    
    Cost: integer;      //   
begin
  //   :
  Result:= THugeGraph.Create('',  // 
                         true,    //  
                         false,   //   
                         true     //  
                        );
  //    ,    1  aFact

  for i:= 1 to aFact do Result.InsertNode(TNodeInt.Create(i, 0, Result));

  //        

  with Result do begin
    N1:= NodeFirst as TNodeInt;
    while Assigned(N1) do begin
      PosPush;
      N2:= NodeFirst as TNodeInt;
      while Assigned(N2) do begin
        if N1 <> N2 then SetLink(N1, N2, MaxInt);
        N2:= NodeNext as TNodeInt;
      end;
      PosPop;
      N1:= NodeNext as TNodeInt;
    end;
  end;

  //       
  //      TLink.mLow
  //      ()
  //  TGraph.MarkMinPairsDicoty

  NH1:= NodeFirst as THamNode;
  while Assigned(NH1) do begin
    //    :
    if NH1.mLeft then begin
      //   :
      Link:= NH1.OutLinkFirst;
      while Assigned(Link) do begin
        NH2:= (Link.mDest as THamNode).mBack;
        //       (),  
        if NH1.mValue <> NH2.mValue then begin
          //     
          // (  TGraph.MarkMinPairsDicoty)
          Cost:= Link.mLow;
          //     ()   
          //   NH1.mValue, NH2.mValue  
          //    
          Link:= Result.GetLinkByNumber(NH1.mValue, NH2.mValue);
          //     () 
          if Link.mValue > Cost then Link.mValue:= Cost;
        end;
        Link:= NH1.OutLinkNext;
      end;
    end;
    NH1:= NodeNext as THamNode;
  end;
  //   ,     
  Result.Compress;
end;


//       
//     (   )
// :
// Result   --        mHigh
// aFactors --  
// aStart   --    

function THamGraph.CalcCostAndFactors(
                            var aFactors: integer; //- 
                            var aStart: THamNode   //  
                            ): integer;            //  

var Level: integer; //  

  //    

  procedure Local(aGraph: THamGraph);
  var Shrinked: TGraph; //  ,   
      HamGr: THamGraph; //     
      Cost: integer;    //  
      Factors: integer; //  
      Dummy: integer;   //    ( )
      Node: THamNode;   //    
  begin

    Inc(Level); //   

    if Level=1 then begin
      //        
      if not aGraph.Check then begin
        //   
        Result:= -1;   Exit;
      end;
    end;

    //       
    Cost:= aGraph.MarkMinPairsDicoty(Dummy);

    //   ( )
    Inc(Result, Cost);

    //   ()
    Factors:= aGraph.CalcFactors(Node);

    //       
    //     

    if Level=1 then begin
      aFactors:= Factors;
      aStart:= Node;
    end;

    //   ,     
    if Factors = 1 then Exit;

    //     (  )
    //    ,    
    //    
    Shrinked:= aGraph.Shrink(Factors);

    //     THamGraph
    //     

    HamGr:= THamGraph.Create(Shrinked);
    Local(HamGr);

    //   
    HamGr.Free;
    Shrinked.Free;
    Dec(Level); //   -1
  end;

begin   { THamGraph.CalcCost }
  Result:=0;   aFactors:= 0; aStart:= nil;
  Level:= 0;
  Local(Self);
end;

// TClosed -  , :
// --  ,
// --  ,
// --    (),
// --      

type  TClosed = class (TItem)
        mCost: integer;  //  
        mId: integer;    //  
        mClosed: TSet;   //    ()
        mFactors: TSet;  //     
        constructor Create(aCost: Integer; aClosed: TSet);
        destructor Destroy;  override;
        function Compare(arg: TItem): TCompare; override;
        procedure InsertLinksSet(aSet: TSet);
        procedure GatesOn;
        procedure GatesOff;
        procedure Print(var aFile: TextFile); override;
      end;

var ClosedId : integer; //    
                        //    

constructor TClosed.Create(aCost: Integer; aClosed: TSet);
begin
  inherited Create;
  Inc(ClosedId);    //   
  mId:= ClosedId;   //     
  mCost:= aCost;    //      
  //        
  mClosed:= CreateSet;
  if Assigned(aClosed) then mClosed.CopyItems(aClosed);
  //     
  // (      ).
  //        
  //  TClosed.InsertLinksSet
  mFactors:= CreateSet;
end;

destructor TClosed.Destroy;
begin
  mFactors.ClrAndDestroy;
  mFactors.Free;
  mClosed.Free;
  inherited;
end;

//      aSet

procedure TClosed.InsertLinksSet(aSet: TSet);
begin
  mFactors.Insert(aSet)
end;

//       
//     ,
//     -   c
//    

function TClosed.Compare(arg: TItem): TCompare;
begin
  if mCost < (arg as TClosed).mCost
    then Result:= cmpLess
  else if mCost > (arg as TClosed).mCost
    then Result:= cmpGreate
  else {       }
    if mId < (arg as TClosed).mId
      then Result:= cmpLess
    else if mId > (arg as TClosed).mId
      then Result:= cmpGreate
    else Result:= cmpEq
end;

//  ,   mClosed

procedure TClosed.GatesOff;
var Link: TLink;
begin
  with mClosed do begin
    Link:= GetFirst as TLink;
    while Assigned(Link) do begin
      Link.mHigh:= 1; //   
      Link:= GetNext as TLink;
    end;
  end;
end;

//  ,   mClosed

procedure TClosed.GatesOn;
var Link: TLink;
begin
  with mClosed do begin
    Link:= GetFirst as TLink;
    while Assigned(Link) do begin
      Link.mHigh:= 0; //   
      Link:= GetNext as TLink;
    end;
  end;
end;

//    

procedure TClosed.Print(var aFile: TextFile);
var L : TLink;
begin
  Writeln(aFile);  Writeln(aFile, '=== C L O S E D === ', mCost);
  L:= mClosed.GetFirst as TLink;;
  while Assigned(L) do begin
    Write(aFile, (L.mOwner as THamNode).mNode.GetName);
    Write(aFile, ' -> ');
    Writeln(aFile, (L.mDest as THamNode).mNode.GetName);
    L:= mClosed.GetNext as TLink;;
  end;
end;

//########################################################################
//             TGraph.GenHamilton
//      
//########################################################################

function TGraph.GenHamilton(var aCost: integer): TBuffer;

var   HamGraph: THamGraph; //    (Dicotyledonous)
      Que : TSet;          //    ( )
      BestCost: integer;   //     
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //  ,    aLinks

  procedure LinksOff(aLinks: TSet);
  var Link: TLink;
  begin
    Link:= aLinks.GetFirst as TLink;
    while Assigned(Link) do begin
      Link.mHigh:= 1; //  
      Link:= aLinks.GetNext as TLink;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //  ,    aLinks

  procedure LinksOn(aLinks: TSet);
  var Link: TLink;
  begin
    Link:= aLinks.GetFirst as TLink;
    while Assigned(Link) do begin
      Link.mHigh:= 0;  //  
      Link:= aLinks.GetNext as TLink;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     aClosed
  // (     )

  procedure Handle(aClosed: TClosed);
  var Cost: integer;    //     
      Factors: integer; //     
      Start: THamNode;  //    
      Closed: TClosed;  //    
      Links: TSet;      //      
      Node: THamNode;
  begin
    //    
    aClosed.GatesOff;

    // :   ,   ,
    //    ,
    //       TClosed
    //   

    //    :
    Links:= aClosed.mFactors.GetFirst as TSet;
    while Assigned(Links) do begin
      //      ,    
      LinksOff(Links);
      //       
      Cost:= HamGraph.CalcCostAndFactors(Factors, Start);
      //    ,   
      LinksOn(Links);
      //   ,    ...
      if (Cost>=0) and (Cost< BestCost) then begin
        //         .
        //   =1,    .
        if Factors=1 then begin
          BestCost:= Cost;            //  
          aCost:= Cost;               //    
          //       
          HamGraph.SaveBest(Result);
        end else begin
          //    Factors > 1 ( ),
          //         
          Links.Add(aClosed.mClosed); //  Links  mClosed
          Closed:= TClosed.Create(Cost, Links);
          //   
          //     Closed  
          Node:= Start; //  ()  
          repeat
            //  THamNode.GenFactorsLinks   
            //    
            Closed.InsertLinksSet(Node.GenFactorsLinks);
            Node:= Node.NextCircNode; //   
          until Node = Start;
          Que.Insert(Closed); //     
        end;
      end;
      //    :
      Links:= aClosed.mFactors.GetNext as TSet
    end;
    //    
    aClosed.GatesOn;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var  Closed: TClosed; //    

begin  { TGraph.GenHamilton }

  Result:= TBuffer.Create;  //      
  aCost:= -1;               //     

  //     HamGraph (Dicotyledonous)
  //       ,
  //     Que (  TGate)

  HamGraph:= THamGraph.Create(Self);
  Que:= CreateSet;  //      
  BestCost:= MaxInt;//     

  //        ,
  //        
  Closed:= TClosed.Create(0, nil);
  Closed.InsertLinksSet(CreateSet);
  Que.Insert(Closed);

  //    :
  while Que.GetCount > 0 do begin
    //    - 
    Closed:= Que.GetFirst as TClosed;
    //         ,
    //     
    if Closed.mCost < BestCost then Handle(Closed);
    //  - 
    Que.Delete(Closed);
    Closed.Free;
  end;

  //  :
  Que.Free;       //      
  HamGraph.Free;  //    (Dicotyledonous)
end;


//########################################################################
//      
//      
//   
//########################################################################

function TGraph.GenHamilton_Quick(var aCost: integer): TBuffer;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   :
  //     (aY, aZ)
  //   

  function FindX(aX : TNode; var aY, aZ : TNode): integer;
  var L: TLink;
      YVal, ZVal : integer; //   
  begin
    aY:= nil; aZ:= nil;
    YVal:= MaxInt; ZVal:= MaxInt;
    //   :
    L:= aX.OutLinkFirst;
    //  YVal <= ZVal
    while Assigned(L) do begin
      if L.mValue <= YVal then begin
        ZVal:= YVal;
        YVal:= L.mValue;
        aZ:= aY; aY:= L.mDest;
      end else if L.mValue < ZVal then begin
        ZVal:= L.mValue;
        aZ:= L.mDest;
      end;
      L:= aX.OutLinkNext;
    end;
    //      
    Result:= YVal + ZVal;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      
  //    

  function FindNode(var aX, aY, aZ : TNode): integer;
  var X, Y, Z : TNode;
      Cost, BestCost: integer;
  begin
    aX:= nil; aY:= nil; aZ:= nil;
    BestCost:= MaxInt;
    X:= NodeFirst;
    while Assigned(X) do begin
      Cost:= FindX(X, Y, Z);
      if Cost < BestCost then begin
        BestCost:= Cost;
        aX:= X; aY:= Y; aZ:= Z;
      end;
      X:= NodeNext;
    end;
    Result:= BestCost;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var NodeX, NodeY, NodeZ : TNode;
    Cost, CostX : integer;

begin  { GenHamilton_Quick }

  //     :
  if mDirect then begin
    Result:= GenHamilton(aCost);
    Exit;
  end;

  //   X      Y, Z ,
  //     X   Y  Z 

  CostX:= FindNode(NodeX, NodeY, NodeZ);

  //      Y   Z
  //    X

  Result:= GenHamPathStrip(NodeY, NodeZ, NodeX, Cost);
  if Cost < 0 then Exit;

  aCost:= Cost + CostX;  //   

  //      Z  Y,    
  Result.Push(NodeX);

  //  ,         
  NodeX:= NodeFirst;
  repeat
    NodeY:= Result.Get as TNode;
    if NodeY <> NodeX then Result.Put(NodeY);
  until NodeY = NodeX;
  Result.Push(NodeY);
end;


//########################################################################
//                                                                       #
//                                                         #
// ,  273,                                                 #
//                                  #
//                                                                       #
//########################################################################

//      
//     

function CalcValue(aLink: TLink): integer;
begin
  with aLink do Result:= mValue + mOwner.mDist + mDest.mDist;
end;


type  // THamLink --     

  THamLink = class (TItem)
    mValue: integer; //    
    mLink: TLink;    //    
    constructor Create(aLink: TLink);
    function  Compare(arg: TItem): TCompare; override;
    procedure Print(var aFile: TextFile); override;
  end;


constructor THamLink.Create(aLink: TLink);
begin
  inherited Create;
  mLink:= aLink;
  //      
  mValue:= CalcValue(aLink);
end;


function THamLink.Compare(arg: TItem): TCompare;
begin
  Result:= inherited Compare(arg);
  if Result = cmpEq then Exit;
  with mLink do begin
    //        
    //    
    if not mOwner.mOwner.mDirect and  //   
      (mOwner=(arg as THamLink).mLink.mDest) and
      (mDest=(arg as THamLink).mLink.mOwner)
      then begin
        //    ,  :
        Result:= cmpEq;    Exit;
      end;
  end;
  //    ,     
  Result:= cmpLess;
  if mValue > (arg as THamLink).mValue then Result:= cmpGreate;
end;

//      ,
//    -  

procedure THamLink.Print(var aFile: TextFile);
var N1, N2 : TNode;
begin
  with mLink do begin
    N1:=mOwner;  N2:=mDest;
    if not mOwner.mOwner.mDirect and
       (mOwner.Compare(mDest) = cmpGreate) then begin
      N1:=mDest;   N2:=mOwner;
    end;
    Write(aFile, ' ' + N1.GetName + N2.GetName);
    Write(aFile, '=', mValue:3, '':1);
  end;
end;


//########################################################################
//      aStart   aFin
//########################################################################

function TGraph.GenHamPath(aStart,       //  
                           aFin: TNode;  //  
                           var aCost: integer  //  
                           ): TBuffer;
begin
  if mDirect
    //        
    then Result:= GenHamPathDir(aStart, aFin, aCost)
    //    -   
    // ,  273,
    //    
    else Result:= GenHamPathStrip(aStart, aFin, nil, aCost);
end;


//########################################################################
//      aStart   aFin
//   
//########################################################################

function TGraph.GenHamPathDir(aStart, aFin: TNode; var aCost: integer): TBuffer;
var Closer,      //  
    Node: TNode;
begin
  Closer:= TNode.Create(0, Self);  //  .  
  InsertNode(Closer);              //    
  Closer.MakeLink(aStart, 0);      //   Closer -> aStart
  aFin.MakeLink(Closer, 0);        //   aFin -> Closer
  Result:= GenHamilton(aCost);     //   
  //      
  repeat
    Node:= Result.Get as TNode;
    if Node = Closer then Break;
    Result.Put(Node);
  until false;
  aFin.RemoveLink(Closer);    //   aFin -> Closer
  Closer.RemoveLink(aStart);  //   Closer -> aStart
  RemoveNode(Closer);         //     
  Closer.Free;                //  
end;


//########################################################################
//      aStart   aFin
//    aStrip
// ( aStrip=nil,   )
// ,  273,
//    
//########################################################################

function TGraph.GenHamPathStrip(aStart,        //  
                                aFin,          //  
                                aStrip: TNode; //  
                                var aCost: integer  //  
                                ): TBuffer;

//   TNode  :
// mDist    --  
// mPower   --      
// mColor   --  
// mRoot    --  
// mFlow    --     
// mLimit   --    ( )

var   Que : TSet; //        
     Tree : TSet; //     
   Marked : TBuffer;  //     
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    mDist ( )

  procedure ClearFines;
  var Node: TNode;      //  
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      with Node do begin
        mDist:= 0;    // 
        mLimit:= 1;   //     
        mRoot:= nil;  //  
      end;
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    Node.mLimit
  //      

  procedure SetLimits(aLimit: integer);
  var Node: TNode;
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mLimit:= aLimit; //     ()
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     ,    
  // (      )
  // mColor = CWhite   --  
  // mPower = 0        --    

  procedure ResetNodesAndLinks;
  var Node : TNode;
      Link : TLink;
  begin
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mColor:= CWhite; // 
      Node.mPower:= 0;      //     
      //   
      Link:= Node.OutLinkFirst;
      while Assigned(Link) do begin
        Link.mColor:= CWhite;
        Link:= Node.OutLinkNext;
      end;
      Node:= NodeNext;
    end;
    //    :
    if Assigned(aStrip) then aStrip.mColor:= CRed;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      
  //          Tree

  function GenCover: boolean;

  var Node : TNode;     //  
      Link : TLink;     //    
        HL : THamLink;  //   
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //    Que     .
    //  HL   Que    

    procedure AddTreeLinks(aNode: TNode);
    var  Link : TLink;
           HL : THamLink;
    begin
      //   
      Link:= aNode.OutLinkFirst;
      while Assigned(Link) do begin
        //   ,    
        if Link.mDest.mColor = CWhite then begin
          //   :
          HL:= THamLink.Create(Link);
          //    :
          if not Que.Insert(HL) then HL.Free;
        end;
        Link:= aNode.OutLinkNext;
      end;
    end;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  var N : integer;  //   

  begin  { GenCover }
    Result:= true;
    //  :
    Tree.ClrAndDestroy;   //  
    ResetNodesAndLinks;   //    : mColor = CWhite
    Node:= aStart;        //    
    Node.mColor:= CBlack; //  
    //      
    // (    )
    N:= mNodes.GetCount-1;
    if Assigned(aStrip) then Dec(N); //   
    while Assigned(Node) and (Tree.GetCount < N) do  begin
      //    Que     :
      AddTreeLinks(Node);
      //   Que     
      Node:= nil;                     //     
      HL:= Que.GetFirst as THamLink;  //   -- 
      while Assigned(HL) do begin
        Que.Delete(HL);               //    
        //      ,   
        with HL.mLink do if mDest.mColor = CWhite
            then Node:= mDest;   //  
        if Assigned(Node) then begin
          //     :
          Node.mColor:= CBlack;     //     
          Tree.Insert(HL);          //     ()
          HL.mLink.mColor:= CBlack; //   
          //     :
          Link:= HL.mLink.GetReverse;
          if Assigned(Link) then Link.mColor:= CBlack;
          Break; //   while
        end else begin
          //       ,  
          HL.Free;
          HL:= Que.GetNext as THamLink; //   
        end;
      end; // while
    end; // while
    //    ,    ,    
    if Tree.GetCount < N then begin
      Tree.ClrAndDestroy;
      Result:= false;
    end; // if
    Que.ClrAndDestroy;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    ,     TNode.mPower
  //      

  function CalcPowerNodes: integer;
  var HL : THamLink;   //    Tree
      Node: TNode;
  begin
    //    mPower
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mPower:= 0;  //     
      Node:= NodeNext;
    end;
    //     Tree:
    HL:= Tree.GetFirst as THamLink;
    while Assigned(HL) do begin
      //     :
      Inc(HL.mLink.mOwner.mPower);    //   
      Inc(HL.mLink.mDest.mPower);     //   
      HL:= Tree.GetNext as THamLink;
    end;
    //       2
    Result:=-2;
    Node:= NodeFirst;
    while Assigned(Node) do begin
      if Node.mPower=1 then Inc(Result);
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //      (  mRoot).
  //       .
  //    .

  procedure Mark(aNode: TNode);
  var   Link : TLink;
        Node : TNode;
  begin
    Marked.Clear;         //     
    Marked.Put(aNode);    //     
    aNode.mColor:= CGray; //   
    //    :
    while Marked.GetCount > 0 do begin
      Node:= Marked.Get as TNode;   //    
      Node.mRoot:= aNode;           //      mRoot
      Link:= Node.OutLinkFirst;
      //     ,   :
      while Assigned(Link) do begin
        //   
        if (Link.mColor = CBlack) and
           //      
           (Node.mColor <> Link.mDest.mColor) then begin
          //   :
          Marked.Put(Link.mDest);     //  
          Link.mDest.mColor:= CGray;  //   
        end;
        Link:= Node.OutLinkNext;
      end;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //        aLink
  //   CalcNodeFine
  //      aStatus

  function CalcLinkFine(aLink: TLink; var aStatus: boolean): integer;
  var Link : TLink;
      Node : TNode;
      Fine : integer;
  begin
    Result:= MaxInt;  aStatus:= false;
    //       ,   :
    //   - ,  - ,  - 
    //     ( mRoot)
    Mark(aLink.mOwner);  //  mOwner
    Mark(aLink.mDest);   //  mDest
    //    ,   
    Node:= NodeFirst;
    while Assigned(Node) do begin
      //     (  ):
      if Node <> aStrip then Node.mColor:= CBlack;
      //     mOwner    
      if (Node.mRoot = aLink.mOwner) and (Node <> aLink.mOwner) then begin
        //   ,   
        Link:= Node.OutLinkFirst;
        while Assigned(Link) do begin
          //      ,
          //        :
          // Link.mDest.mRoot --  
          // aLink.mDest --   
          if Link.mDest.mRoot = aLink.mDest then begin
            Fine:= CalcValue(Link);
            if Result > Fine then begin
              Result:= Fine;  //  
              aStatus:= true; //   
            end;
          end;
          Link:= Node.OutLinkNext;
        end;
      end;
      Node:= NodeNext;
    end;
    //   ( )
    //   
    if aStatus then Result:= Result - CalcValue(aLink);
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //       
  //   CalcFines
  //      aStatus

  function CalcNodeFine(aNode: TNode; var aStatus: boolean): integer;
  var Link, Reverse: TLink;  //    
      Fine: integer;         //  
      OK: boolean;           //  
  begin
    Result:= MaxInt;  aStatus:= false;
    //  ""  ,    
    // ( )    ,
    //    
    Link:= aNode.OutLinkFirst;
    while Assigned(Link) and (Result > 0) do begin
      if Link.mColor = CBlack then begin
        //    
        //  ""     ( )
        Link.mColor:= CGray;
        Reverse:= Link.GetReverse;
        if Assigned(Reverse) then Reverse.mColor:= CGray;
        //  :
        aNode.OutPosPush;   //   
        Fine:= CalcLinkFine(Link, OK); //  
        aNode.OutPosPop;    //   
        //       
        Link.mColor:= CBlack;
        if Assigned(Reverse) then Reverse.mColor:= CBlack;
        //    
        if OK and (Result > Fine) then begin
          Result:= Fine;
          aStatus:= true;
        end;
      end;
      Link:= aNode.OutLinkNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //       
  //    

  function CalcFines: integer;
  var Node: TNode;
      Fine: integer;
      OK: boolean;
  begin
    Result:=0; //    
    //  ,    2,
    //         mFlow
    // Node.mPower --    
    // Node.mFlow --     
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Node.mFlow:= 0;         // mFlow --  
      if (Node.mPower > 2)    //   
         and (Node.mLimit > 0)//     
        then begin
        Dec(Node.mLimit); //  
        Inc(Result);      //   
        //   ,   :
        PosPush; //    
        Fine:= CalcNodeFine(Node, OK);
        PosPop;  //    
        if OK then begin
          if Fine = 0 then Fine:=1;  //    
          Node.mFlow:= Fine;         // mFlow --  
        end;
      end; // if
      Node:= NodeNext;
    end; // while
    //    mDist:= mDist + Node.mFlow
    Node:= NodeFirst;
    while Assigned(Node) do begin
      Inc(Node.mDist, Node.mFlow);
      Node:= NodeNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    
  //     aLink
  //   CalcDeltaNode
  //    ,  

  function CalcDeltaLink(aLink: TLink; var aRes: TLink): integer;
  var Link : TLink;
      Node : TNode;
  begin
    Result:= MaxInt;
    aRes:= nil;
    //       ,
    //   - ,   - ,  - 
    //     ( mRoot)
    Mark(aLink.mOwner);  //  mOwner
    Mark(aLink.mDest);   //  mDest
    //    ,   
    //     1 ()
    Node:= NodeFirst;
    while Assigned(Node) do begin
      //     (  )
      if Node <> aStrip then Node.mColor:= CBlack;
      //      mOwner
      if (Node.mRoot = aLink.mOwner) and (Node.mPower < 2)
        then begin
        //   
        Link:= Node.OutLinkFirst;
        while Assigned(Link) do begin
          //       
          if (Link.mDest.mRoot = aLink.mDest)  //   
            and (Link.mDest.mPower < 2)        //    2-
            and (Result > CalcValue(Link))     //    
          then begin
              Result:= CalcValue(Link);  //  
              aRes:= Link;               //  
          end; // if
          Link:= Node.OutLinkNext;
        end; // while
      end; // if
      Node:= NodeNext;
    end; // while
    //     
    if Assigned(aRes) then Result:= Result - CalcValue(aLink);
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //         aNode
  //   ForceCoverTree
  //   : aBlack -- , aWhite -- 

  procedure FindBlackAndWhite(aNode: TNode; var aBlack, aWhite: TLink);
  var Link : TLink;   //  
      White: TLink;   //  
      Delta: integer; //  
      Minim: integer; //  
  begin
    aBlack:= nil;    aWhite:= nil;
    Minim:= MaxInt;
    //  ""  ,    
    // (   )
    //    ,
    //      
    Link:= aNode.OutLinkFirst;
    while Assigned(Link) do begin
      if Link.mColor = CBlack then begin
        //    ,   ( )
        Link.mColor:= CGray;
        //      :
        with Link do begin
          Dec(mOwner.mPower);
          Dec(mDest.mPower);
        end;
        aNode.OutPosPush; // .  
        //   
        Delta:= CalcDeltaLink(Link, White);
        aNode.OutPosPop;  // .  
        //   
        Link.mColor:= CBlack;
        //    
        // .   :
        with Link do begin
          Inc(mOwner.mPower);
          Inc(mDest.mPower);
        end;
        //   
        if Assigned(White) and (Delta < Minim) then begin
          Minim:= Delta;   // 
          aBlack:= Link;   //   ( )
          aWhite:= White;  //   ( )
        end;
      end;
      Link:= aNode.OutLinkNext;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     

  procedure Tree_Delete(aLink: TLink);
  var HL : THamLink; //    Tree
      Rev: TLink;    //  
  begin
    //       aLink:
    Dec(aLink.mOwner.mPower);
    Dec(aLink.mDest.mPower);
    //      
    aLink.mColor:= CWhite;
    Rev:= aLink.GetReverse;  //  
    if Assigned(Rev) then Rev.mColor:= CWhite;
    //      
    HL:= Tree.GetFirst as THamLink;
    while Assigned(HL) do begin
      if (HL.mLink = aLink) or (HL.mLink = Rev) then Break;
      HL:= Tree.GetNext as THamLink;
    end;
    //      
    if Assigned(HL) then begin
      Tree.Delete(HL);
      HL.Free;
    end;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     

  procedure Tree_Insert(aLink: TLink);
  var HL : THamLink;
      Rev: TLink;
  begin
    //    :
    Inc(aLink.mOwner.mPower);
    Inc(aLink.mDest.mPower);
    //      
    aLink.mColor:= CBlack;
    Rev:= aLink.GetReverse;
    if Assigned(Rev) then Rev.mColor:= CBlack;
    //   :
    HL:= THamLink.Create(aLink);
    //     :
    if not Tree.Insert(HL) then HL.Free;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     

  procedure ForceCoverTree;
  var Node: TNode;
      Black, White : TLink;
  begin
    //   (    2)
    // Node.mPower --    
    Node:= NodeFirst;
    while Assigned(Node) do begin
      if Node.mPower > 2 then begin
        //   ,   
        PosPush; //    
        FindBlackAndWhite(Node, Black, White);
        PosPop;  //    
        //    
        if Assigned(Black) and Assigned(White) then begin
          //  
          Tree_Delete(Black);   //     
          Tree_Insert(White);   //     
        end;
      end;
      Node:= NodeNext;
    end; // while
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  //    

  function Force(var aCost: integer): TBuffer;

    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //     
    //   aBuf   Result
    //     

    function MoveResult(aRes: TBuffer): integer;
    var HL : THamLink; //     Tree
    begin
      Result:= 0;
      aRes.Clear;
      //     :
      HL:= Tree.GetFirst as THamLink;
      while Assigned(HL) do begin
        aRes.Put(HL.mLink);
        Inc(Result, HL.mLink.mValue); //  
        HL:= Tree.GetNext as THamLink;
      end;
    end;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  var  Delta : integer;  // ,    

  begin { Force }
    // ,    
    repeat
      //    ""
      ForceCoverTree;
      //        
      Delta:= CalcPowerNodes;
    until Delta=0;
    //     Buf   Result
    //     aCost
    Result:= TBuffer.Create;
    aCost:= MoveResult(Result);  //
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //   :
  //      Start  Fin

  function ConvertLinksToNodes(aLinks: TBuffer): TBuffer;

    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    //    aLinks   aNode 

    function FindNext(aNode: TNode): TNode;
    var Link: TLink;
        i: integer;
    begin
      Result:= nil;
      //       
      for i:= 1 to aLinks.GetCount do begin
        Link:= aLinks.Get as TLink;
        //   ,   
        with Link do begin
          if aNode = mOwner then begin
            Result:= mDest;  Break;
          end;
          if aNode = mDest then begin
            Result:= mOwner; Break;
          end;
        end;
        //    ,     
        aLinks.Put(Link);
      end;
    end;
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  var  Node: TNode;

  begin { ConvertLinksToNodes }

    Result:= TBuffer.Create;
    //     aStart
    Node:= aStart;
    repeat
      Result.Put(Node);       //   
      Node:= FindNext(Node);  //   
    until (Node = aFin) or not Assigned(Node);
    //    aFin
    if Assigned(Node) then Result.Put(Node);
    aLinks.ClrAndDestroy;
    aLinks.Free;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var
    OK : boolean;       //   
    Changed: boolean;   //     
    Delta : integer;    // ,    
    MinDelta : integer; //   
    Nodes: integer;     //   
    Cost: integer;      //  
    BestCost: integer;  //    
    Res: TBuffer;       //   ()
    FineLimit: integer; //    ()

begin  { TGraph.GenHamPath }

  aCost:= -1; //     

  //      :
  Result:= TBuffer.Create;
  //   ,  :
  if mDirect then Exit;

  //       
  FineLimit:= mNodes.GetCount;

  // :
  Que:= CreateSet;          //    
  Tree:= CreateSet;         //    
  Marked:= TBuffer.Create;  //    

  BestCost:= MaxInt;  //   

  //    mFines:
  ClearFines;

  //    (mDist)   :
  aStart.mDist:= MaxInt div 4;
  aFin.mDist:= aStart.mDist;

  MinDelta:= MaxInt;       //  
  Nodes:= 0;               //   

  // ,    
  repeat
    //       :
    OK:= GenCover;
    // ,    :
    if not OK then Break;
    //   Delta    - 2
    Delta:= CalcPowerNodes;
    //   
    //      ( mPower)
    //     Nodes
    Nodes:= CalcFines;
    Changed:= false;   //     
    //   
    Res:= Force(Cost);
    //    ,  ,   
    if Cost < BestCost then begin
      Changed:= true;   //     
      BestCost:= Cost;  //   
      aCost:= Cost;     //   
      Result.Free;      //   
      Result:= Res;     //   
    end else Res.Free;  //     
    //   
    if Delta < MinDelta then begin
      Changed:= true;       //     
      MinDelta:= Delta;     //   
    end;
    //     
    if Changed then begin
      SetLimits(FineLimit); //    
      FineLimit:= 2 + 3*FineLimit div 4; //   
    end;
  until Nodes = 0;  // ,    
  //     :
  //  -->    Start ...  Fin
  if OK then Result:= ConvertLinksToNodes(Result);
  //  :
  Marked.Free;        //    
  Tree.ClrAndDestroy; //  
  Tree.Free;
  Que.Free;           //  
end;

//########################################################################
// GenHamPath_Full --     
//########################################################################

function TGraph.GenHamPath_Full(aStart, aFin: TNode;
                             var aCost: integer): TBuffer;
var
    Cost: integer;   //   
    Count: integer;  //   
    Node: TNode;     //  
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    .
  //     .
  // aFin -   
  procedure SaveResult;
  begin
    aCost:= Cost;         //   
    Result.Clear;         //   
    Node:= aFin;          //     
    repeat
      Result.Push(Node);  //      
      Node:= Node.mPred;  //  
    until Node=aStart;    //     
    Result.Push(Node);    //   
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     

  procedure Local(aNode: TNode);

  var L : TLink;  //   
  begin { Local }
    aNode.mColor:= CBlack;  //   
    Inc(Count);             //    
    //   ( ) ?
    if Count < mNodes.GetCount-1 then begin
      // ,      
      L:= aNode.OutLinkFirst;
      while Assigned(L) do begin
        if (L.mDest <> aFin) and (L.mDest.mColor = CWhite) then begin
          //     :
          Inc(Cost, L.mValue);        //  
          if Cost < aCost then begin  //    
            L.mDest.mPred:= aNode;    //    
            Local(L.mDest);           //      .
          end;
          Dec(Cost, L.mValue);        //  
        end;
        L:= aNode.OutLinkNext;        //   
      end;
    end else begin
      //  Count = mNodes.GetCount-1 --  ,  aFin, 
      //      aFin
      L:= aNode.OutLinkFirst;
      while Assigned(L) and (L.mDest<>aFin)
        do L:= aNode.OutLinkNext;
      if Assigned(L) then begin
        //   :
        Inc(Cost, L.mValue);      //  
        L.mDest.mPred:= aNode;    //   
        if Cost < aCost           //    
          then SaveResult;        //     
        Dec(Cost, L.mValue);      //  
      end;
    end;
    //       
    Dec(Count);
    aNode.mColor:= CWhite;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

var Temp: TBuffer;  //      

begin { TGraph.GenHamPath_Full }

  Result:= TBuffer.Create;  //   
  aCost:= -1;               //    
  //     ,   :
  if not IsLinked then Exit;

  //      aCost
  //   
  Temp:= GenHamPath_Greed(aStart, aFin, aCost, true);
  Temp.Free;

  ResetNodes;           //   
  Cost:= 0;             //   = 0
  Count:= 0;            //    = 0
  Local(aStart);        //   
  //    ,   1
  if aCost = MaxInt then aCost:=-1;
end;


//########################################################################
// GenHamPath_Greed --   
//    (aGreed=true)
//    (aGreed=false);
//########################################################################

function TGraph.GenHamPath_Greed(aStart, aFin: TNode;
                                var aCost: integer;
                                aGreed: boolean): TBuffer;

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //     
  function GetAnyLink(aNode: TNode): TLink;
  begin
    Result:= aNode.OutLinkFirst;
    while Assigned(Result) and
         (Result.mDest.mColor <> CWhite)
      do Result:= aNode.OutLinkNext;
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  //    
  function GetBestLink(aNode: TNode): TLink;
  var  Link: TLink;
       BestCost: integer;
  begin
    Result:= nil;
    BestCost:= MaxInt;
    Link:= aNode.OutLinkFirst;
    while Assigned(Link) do begin
      if (Link.mDest.mColor = CWhite) and
         (Link.mValue < BestCost) then begin
        BestCost:= Link.mValue;
        Result:= Link;
      end;
      Link:= aNode.OutLinkNext;
    end; // while
  end;
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var
    Cost: integer;   //   
    Count: integer;  //   
    Node: TNode;     //  
    L : TLink;       //  

begin { GenHamPath_Greed }
  Result:= TBuffer.Create;  //   
  aCost:= -1;                 //    
  //     ,   :
  if not IsLinked then Exit;
  ResetNodes;           //   
  aFin.mColor:= CBlack; //   
  Cost:= 0;             //   = 0

  Count:= mNodes.GetCount-2; //   
  Node:= aStart;
  while Count > 0 do begin
    Node.mColor:= CBlack;
    Result.Put(Node);
    //    
    if aGreed
      then L:= GetBestLink(Node) //   
      else L:= GetAnyLink(Node); //   
    //   ,  
    if not Assigned(L) then break;
    //  :
    Inc(Cost, L.mValue);    //  
    Dec(Count);             //  - 1
    Node:= L.mDest;         //  
  end; // while

  if Count=0 then begin
    //    ,  
    Result.Put(Node);       //    
    L:= Node.GetLink(aFin); //     
    if Assigned(L) then begin
      Result.Put(aFin);
      Inc(Cost, L.mValue);  //  
      aCost:= Cost;         //  
    end else Result.Clear;  //    
  end;
end;

//########################################################################
//                                                                       #
//       (, )    #
// TGraph.GenTravelCycle(var aCost: integer): TBuffer;                   #
//                                                                       #
//       ()          #
// TGraph.GenTravelPath(aStart, aFin: TNode; var aCost: integer): TBuffer#
//                                                                       #
//########################################################################

// ,    .
//      aValue ()

procedure TGraph.MakeFull(aValue: integer);
var Node: TNode;
    N : TNode;
    L: TLink;
begin
  //     
  Node:= NodeFirst as TNode;
  //  
  while Assigned(Node) do begin
    Self.PosPush; // .  
    N:= NodeFirst as TNode;
    while Assigned(N) do begin
      if N <> Node then begin
        //     Node --> N
        L:= Node.GetLink(N);
        //   ,   
        if not Assigned(L) then SetLink(Node, N, aValue);
      end;
      N:= NodeNext as TNode;
    end;
    Self.PosPop;  // .  
    Node:= NodeNext as TNode;
  end;
end;

//    
//     

procedure Decompress(aBuf: TBuffer);
var i: integer;
    Node, Next : TNode;
    Link: TFarLink;
begin
  Node:= aBuf.Get as TNode;
  for i:=1 to aBuf.GetCount do begin
    Next:= aBuf.Get as TNode;
    aBuf.Put(Node);
    repeat
      Link:= Node.GetFarLink(Next);
      if Link.mNodeNear = Next then Break;
      Node:= Link.mNodeNear;
      aBuf.Put(Node);
    until false;
    Node:= Next;
  end;
  aBuf.Put(Next);
end;

//   .
//  -,   aBuf,
//     aGr

procedure ReplaceNodes(aGr: TGraph; aBuf: TBuffer);
var Node, N: TNode;
    i: integer;
begin
  for i:=1 to aBuf.GetCount do begin
    //    Node
    Node:= aBuf.Get as TNode;
    //   N
    N:= aGr.GetNode(Node);
    //      
    aBuf.Put(N);
  end;
end;

//       (, )

function TGraph.GenTravelCycle(var aCost: integer): TBuffer;
var FullGraph : TGraph;
begin
  //    
  FullGraph:= Self.Copy as TGraph;
  //    ,
  //      
  FullGraph.MakeFull(MaxInt div 2);
  //      
  //     
  FullGraph.Compress;
  //      
  Result:= FullGraph.GenHamilton(aCost);
  //        
  Decompress(Result);
  //       Self
  ReplaceNodes(Self, Result);
  //   
  FullGraph.Free;
end;

//       ()

function TGraph.GenTravelPath(aStart, aFin: TNode; var aCost: integer): TBuffer;
var FullGraph : TGraph;
    Start, Fin : TNode;
begin
  //    
  FullGraph:= Self.Copy as TGraph;
  //    ,
  //      
  FullGraph.MakeFull(MaxInt div 2);
  //      
  //     
  FullGraph.Compress;
  //         
  Start:= FullGraph.GetNode(aStart);
  Fin:= FullGraph.GetNode(aFin);
  //      
  Result:= FullGraph.GenHamPath(Start, Fin, aCost);
  //        
  Decompress(Result);
  //       Self
  ReplaceNodes(Self, Result);
  //   
  FullGraph.Free;
end;

//########################################################################
//                                                                       #
// THugeGraph --   (    Integer)        #
//            #
//                                                                       #
//########################################################################


//   

constructor THugeGraph.GenRandom(aDir: boolean; // 
                      aLoadNodes, aLoadLinks,   // 
                      aNodes,                   //  
                      aLinks                    //  ,%
                      : integer);
const CArr : array[boolean] of char = ('F','T');

var    nm: string;
        i: integer;
    Links: integer;
    nS,nD: integer;
     Link: TLink;

begin
  //   :
  nm:='Huge: '+ CArr[aDir] +' : '+
         IntToStr(aLoadNodes)+ ' : '+
         IntToStr(aLoadLinks)+ ' : '+
         IntToStr(aNodes)+' : '+
         IntToStr(aLinks);

  //   :
  inherited Create (nm, aDir, aLoadNodes > 0, aLoadLinks > 0);

  //  
  for i:= 1 to aNodes do begin
    InsertNode(TNodeInt.Create(i, 1+Random(aLoadNodes), Self));
  end;

  //  :
  Links:= aNodes*(aNodes-1);  //   
  if aLinks < 100 {%}
    then Links:= Round(Links * aLinks / 100);

  while Links > 0 do begin
    nS:= 1+Random(aNodes);
    nD:= 1+Random(aNodes);
    if nS = nD then Continue;
    Link:= GetLinkByNumber(nS, nD);
    if not Assigned(Link) then begin
      MakeLink(nS, nD, 1+Random(aLoadLinks));
      Dec(Links);
      if not mDirect then Dec(Links);
    end;
  end;
end;


//    

constructor THugeGraph.GenDicoty(aLoadLinks, aPairs: integer);
var nm : string; //  
     n : integer;
    N1, N2 : TNode;

begin
  //   :
  nm:='Huge : Dicoty '+ IntToStr(aLoadLinks)+ ' : '+ IntToStr(2*aPairs);
  //   :
  inherited Create (nm,     // 
                    true,   // 
                    true,   //  
                    true);  //  
  //         n
  for n:= 1 to 2*aPairs do begin
    mNodes.Insert(TNodeInt.Create(n, n, Self));
  end;
  //      ()     ()
  N1:= NodeFirst;
  while Assigned(N1) do begin
    if (N1.mValue mod 2) = 1 then begin
      PosPush;
      N2:= NodeFirst;
      while Assigned(N2) do begin
        if (N2.mValue mod 2) = 0
          then SetLink(N1, N2, 1+Random(aLoadLinks));
        N2:= NodeNext;
      end;
      PosPop;
    end;
    N1:= NodeNext;
  end;
end;


//   

constructor THugeGraph.GenFull(aDir: boolean; // 
                    aLoadNodes, aLoadLinks,   // 
                    aNodes                    //  
                    : integer);

const CArr : array[boolean] of char = ('F','T');
var i : integer;
    N1, N2 : TNode;
    nm : string;
begin
  //   :
  nm:='Huge Full: '+ CArr[aDir] +' : '+
         IntToStr(aLoadNodes)+ ' : '+
         IntToStr(aLoadLinks)+ ' : '+
         IntToStr(aNodes);

  //   :
  inherited Create (nm, aDir, aLoadNodes > 0, aLoadLinks > 0);

  //  
  for i:= 1 to aNodes do begin
    mNodes.Insert(TNodeInt.Create(i, 1+Random(aLoadNodes), Self));
  end;

  //      
  N1:= NodeFirst;
  while Assigned(N1) do begin
    PosPush;
    N2:= NodeFirst;
    while Assigned(N2) do begin
      if N1 <> N2 then SetLink(N1, N2, 1+Random(aLoadLinks));
      N2:= NodeNext;
    end;
    PosPop;
    N1:= NodeNext;
  end;
end;


//          

function THugeGraph.GetLinkByNumber(aSource, aDest: integer): TLink;
var S: TNodeInt;
begin
  Result:= nil;
  S:= GetNode(aSource);
  if not Assigned(S) then Exit;
  Result:= S.GetLinkByNumber(aDest);
end;


//       

function THugeGraph.GetNode(aNumber: integer): TNodeInt;
begin
  mNodes.PositionPush;
  Result:= TNodeInt(mNodes.GetFirst);
  while Assigned(Result) do begin
    if Result.mNumber = aNumber then Break;
    Result:= mNodes.GetNext as TNodeInt;
  end;
  mNodes.PositionPop;
end;


//        

procedure THugeGraph.MakeLink(aSource, aDest, aVal: integer);
var S, D: TNode;
begin
  if GetLinkByNumber(aSource, aDest) <> nil then Exit;
  S:= GetNode(aSource);
  D:= GetNode(aDest);
  if Assigned(S) and Assigned(D)
    then SetLink(S, D, aVal)
end;


//        

function GenHugeGraph (aDir: boolean;          // 
                       aLoadNodes, aLoadLinks, // 
                       aNodes,                 //  
                       aLinks                  //  ,%
                       : integer): TGraph;
begin
  if aLinks >= 100 {%}
    then Result:= THugeGraph.GenFull(aDir,          // 
                            aLoadNodes, aLoadLinks, // 
                            aNodes                  //  
                            )
    else Result:= THugeGraph.GenRandom(aDir,        // 
                            aLoadNodes, aLoadLinks, // 
                            aNodes,                 //  
                            aLinks                  //  ,%
                            );
end;


end.


